/*****************************************************************
 * FILE: aa_globals.h
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: @ALIGNALIGN_RELEASE_DATE@
 * Version:  @ALIGNALIGN_RELEASE_VERSION@
 * 
 ****************************************************************/

#ifndef AA_GLOBALS_H_
#define AA_GLOBALS_H_



#define   AA_PROGNAME  "alignalign"
#define   AA_VERSION   "@ALIGNALIGN_RELEASE_VERSION@"
#define   AA_DATE      "@ALIGNALIGN_RELEASE_DATE@"
#define   AA_TRUE      1
#define   AA_FALSE     0
#define   AA_MAXLINE   1024  /* @@@ maximum size of a line of input    */
#define   AA_MAXFNAME  1024  /* @@@ maximum length of a filename       */
#define   AA_MAXALPHA  128   /* @@@ maximum size of the aa_alphabet       */
#define   AA_MAXNAME   128   /* @@@ maximum size of a sequence name    */
#define   AA_EPS       '-'   // TW - gap symbol
//#define   AA_PROB      1.0   // TW - how much of a "gap cost" should a possible gap count for in optimistc/pessimistic calcs
#define   AA_GAMMA     11    // TW - gap open
#define   AA_LAMBDA    11    // TW - gap extension
#define   AA_COST_MATRIX_FILE  "BLOSUM62"
#define   AA_MALLOCBLOCKSIZE 4080 /* 4K - 16 bytes  (TW size of a page - header) */ 
#define   AA_SCRIPTALLOCSIZE 128  // TW - used to create an array of "breadcrumbs" filled during recovery  
#define   AA_MAXSEQNUM 8192
#define   AA_INF       2147483640
#define   AA_OUTPUTWIDTH 50


enum operation {AA_INS, AA_DEL, AA_SUB};
enum direction {AA_HORIZ, AA_VERT, AA_DIAG};

typedef unsigned char aa_uchar;
typedef unsigned short aa_ushort;
typedef unsigned int aa_uint;
typedef unsigned long aa_ulong;


typedef struct shape { /* list element represention of a (TW quadratic space) shape and its cost */
  struct shape  *next; /* next entry in shape list                          */
  struct shape  *r;    /* for recovery, shape which propagated this shape   */
  int   c;    /* cost on creation, ref (aka children) count after propagation     */
  aa_ushort  s[1]; /* the shape (overlayed with array of length k+l)   */
} Shape;

typedef struct shape_s { /* shape and cost for linear space version       */
  struct shape  *next;   /* next entry in shape list                      */
  int   c;      /* cost                                          */
  aa_ushort  s[1]; /* the shape (overlayed with array of length k+l) */
} Shape_s;

/* 2 structs for an attempt to determine aligment boundary */
//TW - involved with trick to determine how machine is allocating space  (how much buffer does machine allocate)
//TW - if s1 and s0 are different, then padding is occuring
typedef struct s0 {
  int  *p;
  int   x;
} S0;

typedef struct s1 {
  int  *p;
  int   x;
  char  c;
} S1;

/* member access macros */
#define RECOVER(p) ((p)->r)
#define NEXT(p)    ((p)->next)
#define C(p)       ((p)->c)
#define S(p)       ((p)->s)


/* terminal gap cost macros */
#define GET_GAMMA_A(p,i) GetGamma(aa_firstCharsA, aa_lastCharsA, p, i);
#define GET_GAMMA_B(p,i) GetGamma(aa_firstCharsB, aa_lastCharsB, p, i);
#define GET_LAMBDA_A(p,i) GetLambda(aa_firstCharsA, aa_lastCharsA, p, i);
#define GET_LAMBDA_B(p,i) GetLambda(aa_firstCharsB, aa_lastCharsB, p, i);

/* global variables */
extern char  **aa_A;                      /* storage for input alignment aa_A     */
extern char  **aa_B;                      /* storage for input alignment aa_B     */
extern int **aa_cost;/* cost[a,b] = substitution cost     */
//extern int     aa_prob;   /* gap open penalty                  */
extern int   aa_gamma;  /* gap open penalty                  */
extern int   aa_lambda; /* gap extension penalty             */
extern int   aa_gamma_terminal ;
extern int   aa_lambda_terminal ;
extern char   *aa_weight_file;
extern int **aa_weights;/* weights[a,b] = weight of sequence pair (a,b)     */    
extern int     aa_weighted;

extern int     aa_printseqs;  /* */
extern int     aa_printcost;  /* */
extern int     aa_printinput;  /* */
extern int     aa_M, aa_N, aa_K, aa_L;    /* aa_M=|aa_A|,aa_N=|aa_B|,aa_K=# aa_A seqs,aa_L=# aa_B seqs */
extern int    *aa_firstCharsA;
extern int    *aa_firstCharsB;
extern int    *aa_lastCharsA;
extern int    *aa_lastCharsB;
extern int     aa_nodesize;
extern int     aa_nodecount;
extern void   *aa_nodeptr;
extern Shape  *aa_nodepool;
extern void   *aa_nodepoolroot;
extern int     aa_nodepoolcount;
extern int     aa_nodemalloccount;
extern int     aa_nodemallocsize;
extern aa_ulong aa_ec1;
extern aa_uchar **aa_fa; // For profile of columns - for linear-time 
extern aa_uchar **aa_fb;
extern aa_uchar **aa_ca;
extern aa_uchar **aa_cb;
extern int  *aa_alphaconv;
extern char *aa_alphabet;
extern int  aa_alphasize;
/* upperbound on the cost of optimal alignment of aa_A and aa_B */
extern int  aa_cost_upper_bound; 
/* lowerbound lookup table - from optimistic run on reverse of input */
extern int **aa_optH, **aa_optV, **aa_optD; //TW - tables for bound pruning
extern char **aa_namesA;
extern char **aa_namesB;

extern unsigned char *aa_script;

#endif /*AA_GLOBALS_H_*/
