/*****************************************************************
 * FILE: driver.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: @ALIGNALIGN_RELEASE_DATE@
 * Version:  @ALIGNALIGN_RELEASE_VERSION@
 * 
 ****************************************************************/

#include <stdlib.h>		/* NULL */
#include <stdio.h>		/* fprintf */
#include <string.h>		/* strlen */
#include "alignalign.h"
#include "driver.h"
#include <limits.h>


int main(int argc, char *argv[])
{
    char          **A = NULL;	/* storage for input alignment aa_A */
    char          **B = NULL;	/* storage for input alignment aa_B */
    char           *A_file = NULL;
    char           *B_file = NULL;
    char          **namesA = NULL;
    char          **namesB = NULL;

    int             N,
                    M,
                    K = 0,
                    L = 0 ;


    int             i,
                    j;
    int             space_optimized = AA_FALSE;
    int             printsolution = AA_TRUE;
	int 			print_format = CLUSTAL_FORMAT;
	
    int             max_sub = 0;
    FILE           *out = stdout;
    char           *out_file = NULL;

    char           *cost_file = NULL;	/* input cost file name */
	int     	  **cost;/* cost[a,b] = substitution cost     */
    char           *weight_file = NULL;	/* input weightfile name */
	int     	  **weights;/* weights[a,b] = weight of sequence pair (a,b)     */    
	int 		    verbose = AA_TRUE;   
	int     		gamma = AA_GAMMA;  /* gap open penalty                  */
	int     		lambda = AA_LAMBDA; /* gap extension penalty             */	
	int     		gamma_terminal = -1;  /* gap open penalty                  */
	int     		lambda_terminal = -1; /* gap extension penalty             */	

	int 			outputwidth = AA_OUTPUTWIDTH;
		
	char		 	*alphabet = NULL;	
	int			 	*alphaconv = NULL;
	char 			**alignment = NULL; 	
	int 			just_get_score = AA_FALSE;
	int 			convert_f2c = AA_FALSE;	
    /*
     * parse the command line 
     */
    for (i = 1; i < argc; i++) {
		if ('-' == argv[i][0]) {
		    switch (argv[i][1]) {
			    case 'a':
					A_file = argv[i + 1];
					break;
			    case 'b':
					B_file = argv[i + 1];
					break;
			    case 'c':
					cost_file = argv[i + 1];
					break;
			    case 'd':
					gamma = atoi(argv[i + 1]);
					break;
			    case 'e':
					lambda = atoi(argv[i + 1]);
					break;
			    case 'm':  /* yuck-o argument name */
					gamma_terminal = atof(argv[i + 1]);
					break;
			    case 'n': /* yuck-o argument name */
					lambda_terminal = atof(argv[i + 1]);
					break;
				case 'f':
					A_file = argv[i + 1];				
					convert_f2c = AA_TRUE;
					break;
			    case 'g':
					if (i + 2 >= argc) {
					    Usage(argv[0]);
					}
					gamma = atoi(argv[i + 1]);
					lambda = atoi(argv[i + 2]);
					break;
			    case 'h':
					printhelp(argv[0],verbose); //verbose is just to support automated posting of the manpage on the website
					exit(0);
					break;
			    case 'j':
					A_file = argv[i + 1];
					just_get_score = AA_TRUE;
					break;

/*			    case 'p':
					aa_prob = atof(argv[i + 1]);
					break;*/
			    case 'o':
					if (!(out = fopen(argv[i + 1], "w"))) {
					    fprintf(stderr, "%s: error opening file: %s\n",
						    argv[0], argv[i + 1]);
					    exit(1);
					} else {
					    out_file =
						aa_malloc((strlen(argv[i + 1]) + 1) * sizeof(char));
					    strcpy(out_file, argv[i + 1]);
					}
					break;
			    case 'p':
					printsolution = AA_FALSE;
					break;					
			    case 's':
					space_optimized = AA_TRUE;
					break;
			    case 't':
					space_optimized = AA_FALSE;
					break;
			    case 'w':
					outputwidth = atoi(argv[i + 1]);
					break;
			    case 'v':
					verbose = AA_TRUE;
					break;
			    case 'q':
					verbose = AA_FALSE;
					break;
				case 'x':
					print_format = FASTA_FORMAT;
					break;					
				case 'y':
					print_format = CLUSTAL_FORMAT;
					break;
			    case 'z':
					weight_file = argv[i + 1];
					break;
			    default:
					Usage(argv[0]);
					break;
		    };
		}
    }



    if (NULL == A_file || (!convert_f2c && !just_get_score && NULL == B_file)) {
		Usage(argv[0]);
    }

    /*
     * read in the aa_alphabet and cost matrix 
     */
    if (NULL == cost_file) {
		cost_file = AA_COST_MATRIX_FILE;
    }
    int alpha_size = 0;
    alpha_size = ReadCost(cost_file, &cost, &alphabet, &alphaconv);


    /*in
     * set gamma and lambda to 1/2 max sub cost, if not specified by user 
     */
    if (-1 == gamma || -1 == lambda) {
		for (i = 0; i < AA_MAXALPHA; i++) {
		    for (j = 0; j < AA_MAXALPHA; j++) {
				if (max_sub < cost[i][j]) {
				    max_sub = cost[i][j];
				}
		    }
		}
		if (-1 == gamma) {
		    gamma = max_sub / 2;
		    if (0 != max_sub % 2) {
				gamma++;
		    }
		}
		if (-1 == lambda) {
		    lambda = max_sub / 2;
		    if (0 != max_sub % 2) {
				lambda++;
		    }
		}
    }
	if (gamma_terminal < 0) {
		gamma_terminal = gamma ;
	}
	if (lambda_terminal < 0) {
		lambda_terminal = lambda ;
	}    
	aa_setCosts(&cost, alpha_size, alphabet, alphaconv, gamma, lambda, gamma_terminal, lambda_terminal);

    K = ReadSequences(A_file, &A, &M, &namesA);
    if (!just_get_score  && !convert_f2c) {
   		L = ReadSequences(B_file, &B, &N, &namesB);
    }
    
	if ( ReadWeights(weight_file, K+L, &weights) ) {
		aa_setWeights(K+L, weights);
	}
	
	if ( just_get_score ) {
		justScore(K, M, A);
		exit(0);
	}

	if ( convert_f2c) {
		convert_fasta2clustal(K, M, A, namesA, out, outputwidth);
		exit(0);
	}
		
	unsigned char *script;
    int  alignlen = alignalign( space_optimized, &A, K, M, &B, L, N, &alignment, &script);

    // print the alignment
    if (printsolution) {
    	if (print_format == FASTA_FORMAT) {
    		char **names = aa_malloc(( K+L) * sizeof(char*));
    		for (i=0; i<K; i++)
    			names[i] = namesA[i];
    		for (j=0; j<L; j++)	
    			names[i+j] = namesB[j];
			printAlignment_fasta(K+L, alignment, alignlen, out, outputwidth, names);
    	} else if (print_format == CLUSTAL_FORMAT) {
			printAlignment_clustal(K, L, alignment, alignlen, out, outputwidth, namesA, namesB, script);
    	}
    }
    // print solution alignment cost
    if (verbose) {
		int aligncost = sp(alignment, alignlen, K + L);
		int a_cost = sp(alignment, alignlen, K);	
		int b_cost = sp(alignment+K, alignlen, L);
		int inter_cost = alignmentCost(alignment,alignment+K,K,L,alignlen,alignlen);
		fprintf(stderr, "--------------------------------------------\n");
		fprintf(stderr, "%s-%s ", AA_PROGNAME, AA_VERSION);
		if (space_optimized) {
		    fprintf(stderr, "(space optimized)\n");
		} else {
		    fprintf(stderr, "(time optimized)\n");
		}
		fprintf(stderr, "Gap open cost: %d\n", gamma);
		fprintf(stderr, "Gap extension cost: %d\n", lambda);
		if (gamma != gamma_terminal  || lambda != lambda_terminal) {
			fprintf(stderr, "Terminal gap open cost: %d\n", gamma_terminal);
			fprintf(stderr, "Terminal gap extension cost: %d\n", lambda_terminal);
		}
		fprintf(stderr, "Cost matrix: %s\n", cost_file);
		fprintf(stderr, "Input alignments from:\n");
		fprintf(stderr, "\t%s, %d sequences\n", A_file, K);
		fprintf(stderr, "\t%s, %d sequences\n", B_file, L);
		if (NULL != out_file) {
		    fprintf(stderr, "Solution alignment written to: %s\n",  out_file);
		} else {
		    fprintf(stderr, "Solution alignment written to: stdout\n");
		}
		fprintf(stderr, "Solution alignment length: %d\n", alignlen);
		fprintf(stderr, "Solution alignment cost: %d\n", aligncost);
		fprintf(stderr, "Solution component costs: %d + %d + %d\n", a_cost, b_cost, inter_cost);
		fprintf(stderr, "--------------------------------------------\n");
    }

    for (i = 0; i < K; i++) {
	    if (A[i] != NULL) free(A[i]);
	    if (namesA[i] != NULL) free(namesA[i]);
    }
    if (A != NULL) free(A);
    if (namesA != NULL) free(namesA);    

    for (i = 0; i < L; i++) {
	    if (B[i] != NULL) free(B[i]);
	    if (namesB[i] != NULL) free(namesB[i]);
    }
    if (B != NULL) free(B);
    if (namesB != NULL) free(namesB);    
    

    for (i = 0; i < K+L; i++) {
		free(alignment[i]);
    }
	free(alignment);

	for (i=0; i<alpha_size; i++) 
		free (cost[i]);
	free(cost);
     
	free(alphabet);
	free(alphaconv);

	aa_freeWeights();
	
    return 0;
}


void justScore(int K, int M, char **A) {
    unsigned int aligncost = sp(A, M, K);
    printf("the given alignment has score %d\n", aligncost);
    exit(0);
}

void convert_fasta2clustal(int K, int M, char **A, char **namesA, FILE *out, int outputwidth) {
	printAlignment_clustal(K, 0, A, M, out, outputwidth, namesA, NULL, NULL);
    exit(0);
}
