/*****************************************************************
 * FILE: spcost.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: @ALIGNALIGN_RELEASE_DATE@
 * Version:  @ALIGNALIGN_RELEASE_VERSION@
 * 
 ****************************************************************/

#include "spcost.h"
#include <stdio.h>		/* fprintf */

/* this one sums over ALL pairs */
int 
sp (char **A, int m, int n) { /* n := number seqs,  m := length */
  int     i, j, k;
  int     total=0, c = 0;

  for (i = 0; i < n-1; i++) {
    for (j = i+1; j < n; j++) {
	  int first_char_seen_i = AA_FALSE;
	  int first_char_seen_j = AA_FALSE;	  
      c = 0;
      int lastGap = 0;
      int currentGapLength = 0;
      for (k = 0; k < m; k++) {
		if (A[i][k] != AA_EPS && A[j][k] != AA_EPS) {
			c += aa_cost[aa_alphaconv[(int)A[i][k]]][aa_alphaconv[(int)A[j][k]]];
			lastGap = 0;
			currentGapLength = 0;
		  	first_char_seen_i = first_char_seen_j = AA_TRUE;
		} else if (A[i][k] == AA_EPS && A[j][k] != AA_EPS) {
		  c += first_char_seen_i ? aa_lambda : aa_lambda_terminal ;
		  if (lastGap != 1) {
		  	c += first_char_seen_i ? aa_gamma : aa_gamma_terminal;
			currentGapLength=0;
		  }
		  currentGapLength++;
		  lastGap = 1;
	 	  first_char_seen_j = AA_TRUE;	  
		} else if (A[i][k] != AA_EPS && A[j][k] == AA_EPS) {
		  c += first_char_seen_j ? aa_lambda : aa_lambda_terminal ;
		  if (lastGap != 2) {
		  	c += first_char_seen_j ? aa_gamma : aa_gamma_terminal;
		  	currentGapLength=0;
	  	  }
	  	  currentGapLength++;
		  lastGap = 2;
		  first_char_seen_i = AA_TRUE;
		}	
      }
      
      /* if the pairwise alignment ends in a gap, the cost of the alignment should be adjusted to 
       * reflect the terminal gap cost (it was initially calculated with the internal gap cost on the right side
       */
	  if (  (A[i][m-1] != AA_EPS && A[j][m-1] == AA_EPS) ||
	 		(A[i][m-1] == AA_EPS && A[j][m-1] != AA_EPS) ) {
      	c += (aa_gamma_terminal - aa_gamma) + (aa_lambda_terminal - aa_lambda) * currentGapLength;
	  }

	  if (aa_weighted && NULL != aa_weights) 
		  c *= aa_weights[i][j];

      total += c;
    }
  }
  return total;
}


int 
min_pair (char **A, int m, int n1, int n2) {
  int     i, j, k;
  int     c;
  int min_cost = AA_INF;
  
  for (i = 0; i < n1-1; i++) {
    for (j = n1; j < n1+n2; j++) {
      int lastGap = 0;
      int currentGapLength = 0;
	  int first_char_seen_i = AA_FALSE;
	  int first_char_seen_j = AA_FALSE;	  
      c = 0;
      for (k = 0; k < m; k++) {
		if (A[i][k] != AA_EPS && A[j][k] != AA_EPS) {
			c += aa_cost[aa_alphaconv[(int)A[i][k]]][aa_alphaconv[(int)A[j][k]]];
			lastGap = 0;
			currentGapLength = 0;
		  	first_char_seen_i = first_char_seen_j = AA_TRUE;
		} else if (A[i][k] == AA_EPS && A[j][k] != AA_EPS) {
		  c += first_char_seen_i ? aa_lambda : aa_lambda_terminal ;
		  if (lastGap != 1) {
		  	c += first_char_seen_i ? aa_gamma : aa_gamma_terminal;
			currentGapLength=0;
		  }
		  currentGapLength++;
		  lastGap = 1;
	 	  first_char_seen_j = AA_TRUE;	  
		} else if (A[i][k] != AA_EPS && A[j][k] == AA_EPS) {
		  c += first_char_seen_j ? aa_lambda : aa_lambda_terminal ;
		  if (lastGap != 2) {
		  	c += first_char_seen_j ? aa_gamma : aa_gamma_terminal;
		  	currentGapLength=0;
	  	  }
	  	  currentGapLength++;
		  lastGap = 2;
		  first_char_seen_i = AA_TRUE;
		}	
      }
      //shouldn't be weighted
      if (c < min_cost)
      	min_cost=c;
    }
  }
  return min_cost;
}

/* this one sums over pairs, one from aa_A and one from aa_B */
int 
alignmentCost(char **A, char **B, int k, int l, int m, int n ) {
  int i, j, h;
  int left, right;
  int c=0, total = 0;
  int length, Alonger = AA_FALSE;

  if (m < n) { /* aa_B is longer */
    length = m;
  }
  else {       /* aa_A might be longer */
    length = n;
    Alonger = AA_TRUE;
  }


  for (i = 0; i < k; i++) {
    for (j = 0; j < l; j++) {      /* pair of rows (i,j) */
//if (verbose) { printf ("%d,%d: ", i, j);}
		c=0;
/*		int ii;
		for (ii=0;ii<m;ii++) {
			printf("%c", A[i][ii]);	
		}
		printf("\n");
		for (ii=0;ii<n;ii++) {
			printf("%c", B[j][ii]);	
		}
		printf("\n");
*/
		/*
		 * read from LEFT to identify terminal gap
		 */
  		left = 0;
  		while (A[i][left] == AA_EPS && B[j][left] == AA_EPS) {
//if (verbose) { printf ("0.0,"); }
  			left++;	
  		}
		if (A[i][left] != AA_EPS && B[j][left] == AA_EPS) {
			c += aa_gamma_terminal;
	      	while (left < length) {     	
				if (A[i][left] != AA_EPS && B[j][left] == AA_EPS) {
			  		c += aa_lambda_terminal;
//if (verbose) {printf ("%.1f,",c);}			  		
			  		left++;
			  		while (A[i][left] == AA_EPS && B[j][left] == AA_EPS) {
//if (verbose) {printf ("%.1f,",c);}			  		
			  			left++;	
			  		}			  		
				} else {
					break;	
				}
	      	}
		} else if (A[i][left] == AA_EPS && B[j][left] != AA_EPS) {
			c += aa_gamma_terminal;
	      	while (left < length) {     	
				if (A[i][left] == AA_EPS && B[j][left] != AA_EPS) {
			  		c += aa_lambda_terminal;
//if (verbose) {printf ("%.1f,",c);}	
			  		left++;
			  		while (A[i][left] == AA_EPS && B[j][left] == AA_EPS) {
//if (verbose) {printf ("%.1f,",c);}	
			  			left++;	
			  		}			  		
				} else {
					break;	
				}
	      	}
		}
    	
		/*
		 * read from RIGHT to identify terminal gap
		 */
	  	int curr_shape = AA_SUB;    	
	    if (Alonger) {
		  for (h=length ; h < m; h++) {
		    if (A[i][h] != AA_EPS) {
		      c += aa_lambda_terminal;
		      curr_shape = AA_DEL;
		    }
		  }
		} else if (m != n) { // Blonger
		  for (h=length ; h < n; h++) {
		    if (B[j][h] != AA_EPS) {
		      c += aa_lambda_terminal;
  			  curr_shape = AA_INS;
		    }
		  }
		}
  		right = length-1;
  		while (right >= left && A[i][right] == AA_EPS && B[j][right] == AA_EPS) {
				right--;
  		}
		if (curr_shape == AA_SUB) {
			if ( A[i][right] == AA_EPS ) {
				curr_shape = AA_INS;
			} else if ( B[j][right] == AA_EPS ) {
				curr_shape = AA_DEL;
			} 
		}
		if (curr_shape == AA_DEL) {
			if ( A[i][right] != AA_EPS && B[j][right] == AA_EPS) {
	      		while (right >= left) {     	
					if (A[i][right] != AA_EPS && B[j][right] == AA_EPS) {
				  		c += aa_lambda_terminal;
				  		right--;
				  		while (right >= left && A[i][right] == AA_EPS && B[j][right] == AA_EPS) {
				  			right--;	
				  		}			  	
					} else {
						break;	
					}
	      		}
			}
		} else if (curr_shape == AA_INS) { 
			if (A[i][right] == AA_EPS && B[j][right] != AA_EPS) {
		      	while (right >= left) {
					if (A[i][right] == AA_EPS && B[j][right] != AA_EPS) {
				  		c += aa_lambda_terminal;
				  		right--;
				  		while (right >= left && A[i][right] == AA_EPS && B[j][right] == AA_EPS) {
				  			right--;	
				  		}			
					} else {
						break;	
					}
		      	}
			}
		}
		if (curr_shape != AA_SUB) {
			c += aa_gamma_terminal;
		}
    	
    	
    	//add up substitution and extension costs between 
    	// the terminal gaps; also count gaps.
    	
      	for (h = left; h <= right; h++) { 
	 	    /* sum substitution costs for pair (i.j) */
			if (A[i][h] != AA_EPS && B[j][h] != AA_EPS) {
				c += aa_cost[aa_alphaconv[(int)A[i][h]]][aa_alphaconv[(int)B[j][h]]];
//if (verbose) {printf ("%.1f,",c);}
			} else if (A[i][h] != AA_EPS || B[j][h] != AA_EPS) { //columns with one gap char
				c += aa_gamma + aa_lambda ;
//if (verbose) {printf ("%.1f,",c);}
		  		if (A[i][h] != AA_EPS) { // so B[j][h] == EPS
		  			while (h+1 <= right && B[j][h+1] == AA_EPS) { 
				  		if (A[i][h+1] != AA_EPS) {
							c += aa_lambda ;
				  		}
//if (verbose) {printf ("%.1f,",c);}
			  			h++; 
		  			}
		  		} else { // B[j][h] != EPS,  so A[i][h] == EPS
		  			while (h+1 <= right && A[i][h+1] == AA_EPS) { 
				  		if (B[j][h+1] != AA_EPS) {
							c += aa_lambda ;
				  		}
//if (verbose) {printf ("%.1f,",c);}
			  			h++; 
		  			}		  			
		  		}
			} else {
//if (verbose) {printf ("%.1f,",c);}
			}
      	}
     
	    if (aa_weighted && NULL != aa_weights) 
			c *= aa_weights[i][k+j];      
      	total += c;
//if (verbose) {printf("\n");}
    }
  }
  return total;
}

