/*****************************************************************
 * FILE: Opal.c
 *****************************************************************
 * Copyright 2006 by Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: @OPAL_RELEASE_DATE@
 * Version:  @OPAL_RELEASE_VERSION@
 * 
 ****************************************************************/

#include <unistd.h>
#include <stdlib.h>		/* NULL */
#include <stdio.h>		/* fprintf */
#include <string.h>		/* strlen */
#include <math.h>  	 	/*log*/
#include <time.h>       /* for random numbers*/
#include <getopt.h>    /* see http://gnu.teleglobe.net/software/gengetopt/gengetopt.html  */
#include "alignalign/alignalign.h"
#include "Opal.h"



char         ***alignments = NULL;	/* storage for input sequences */
ClusterList    *clusters = NULL;
int			   *cluster_sizes = NULL;
int			   *alignment_roots = NULL;
int				tree_node_count ;
Node 		   *tree_nodes = NULL;

char          **names = NULL;
//int		   *alignment_costs = NULL;
int 			cluster_count = 0;
int        	   *lengths = NULL;
int        	   *orig_lengths = NULL;
char 		  **new_alignment;
HeapNode	  ***distance_nodes = NULL;  //distance_nodes[i][j] is the node for alignment j on heap for i
Heap 		  *distance_heap = NULL;
int   	      *alignment2cluster_map = NULL;
int 		***internal_external_distances;
int		 **tree_cut_optimal_distances;
int 		 **nj_optimal_distances;
int 		  *nj_total_distances;
int 		 **internal_external_total_distances;
int 		 **internal_external_numpairs;
char 		*compressed_alph_map;
int 		max_len;
float	    rand_cut_group1_pct = 2.0/3;
float	    rand_cut_group2_pct = 5.0/6;

int 	guide_tree_method = MST; //AVG_INTERNAL;
int verbose = AA_FALSE;   
int space_optimized = AA_FALSE;
int print_format = CLUSTAL_FORMAT;
int do_local_search = AA_FALSE;
int local_search_method = ANY_CUT; 
int allow_both_2way_and_3way_cut = AA_TRUE;
int align_optimally = AA_TRUE; 
int random_cut_method = TREE_CUT; //ANY_CUT;
int polish_internal = AA_FALSE;
int use_weights_in_merge = AA_TRUE;
int alignment_cost_type = ALIGN_COST;  //or  ALL_PAIRS 
int three_tree_cut_type = CUT_PAIR; // or CUT_MULTI
int two_cut_type = NONE; // or ALL or FEW
int two_cut_reorder_threshold = 5;
int distance_type = NORM_EDIT; // or COMPR_PCT_ID, EDIT, PCT_ID  - only valid for nj,tj,upgma,min-internal
int overall_type = THREE_CUT;

time_t start,now;
int timelimit = -1; //in seconds
int order_counter;
int **induced_ij;
int tree_root; 
int     	  **weights = NULL;/* weight[a,b] = pairwise weight     */

int phases = 1;

void buildAlignment (int K, int space_optimized, char ***initial_alignment, int *alignlen, int phase);
void mergeAlignments (int K, int best_i, int best_j, int space_optimized, int new_node_id, int polishing);

int calcBestAlignCost(int space_optimized, int use_weights, int get_avg, char ***A, int K, int M, int *A_ids, char ***B, int L, int N, int *B_ids, char ***result_alignment, 
																unsigned int *result_alignlen, unsigned char **script, int cost_type /*ALL_PAIRS,ALIGN_COST,MIN_COST*/);
/*int calcAlignMergeCost ( int space_optimized, int use_weights, char ***A, int K, int M, int aligncostA, int *A_ids, char ***B, int L, int N, int aligncostB, int *B_ids) ;

int calcAlignMergeCostHelper ( int space_optimized, int use_weights, char ***A, int K, int M, int aligncostA, int *A_ids, char ***B, int L, int N, int aligncostB, int *B_ids, int allPairs) ;
int calcBestAlignCost(int space_optimized, int use_weights, char ***A, int K, int M, int *A_ids, char ***B, int L, int N, int *B_ids, char ***result_alignment, 
																unsigned int *result_alignlen, unsigned char **script);
int calcBestAlignCostHelper(int space_optimized, int use_weights, char ***A, int K, int M, int *A_ids, char ***B, int L, int N, int *B_ids, char ***result_alignment, 
																unsigned int *result_alignlen, unsigned char **script, int allPairs);
*/
int calcBestAlignCost_preservingOrig (int space_optimized, int use_weights, int get_avg, char **A, int K, int M, int *A_ids, 
 																	char **B, int L, int N, int *B_ids,
 																	char ***result_alignment, unsigned int *result_alignlen,
 																	unsigned char **script, int cost_type /*ALL_PAIRS,ALIGN_COST,MIN_COST*/);
 																	
void nWaySplit (int n, int K, int space_optimized, char** alignment, int alignlen, 
								char ****splits, int **split_sizes, int ***maps) ;
int getInducedScript(char **splitA, int splitA_size, char **splitB, int splitB_size, 
								int alignlen, unsigned char **script);
unsigned int polishAlignment (int num_cuts, int K, int space_optimized, char** alignment, 
								int alignlen, char ***polished_alignment, int randomized_polish_reps) ;
void randomlySplit(int num_groups, int K, char ***alignment, int alignlen, char ****splits, 
						int **split_sizes, int ***maps) ;
void randomlySplit_tree(int K, char ***alignment, int alignlen, char ****splits, int **split_sizes, int ***maps) ;
void randomlySplit_pure(int num_groups, int K, char ***alignment, int alignlen, char ****splits, 
						int **split_sizes, int ***maps) ;
unsigned int localSearch (int num_cuts, int K, int space_optimized, char** alignment, 
								int alignlen, char ***polished_alignment, int *cost) ;
unsigned int localSearch_treecut (int K, int space_optimized, char** alignment, 
								int alignlen, char ***polished_alignment, int *cost) ;
unsigned int localSearch_purecut (int num_cuts, int K, int space_optimized, char** alignment, 
								int alignlen, char ***polished_alignment, int *cost); 
int polishInternal(int cluster_id);

void calculatePairwiseDistances(int K, char **alignment, int alignlen, 
					int ***pairwise_optimal, int ***induced_pairwise, int space_optimized);
					
void bestThreeWayMerge(int K, int space_optimized, char ***splits, int **maps, int *split_sizes, 
					int alignlen, unsigned int *new_alignlen1, unsigned int *new_alignlen2, 
					char ***return_alignment2, 
					unsigned char **return_script1, unsigned char **return_script2, 
					int *cost, int *which_merge);
void print_tree (FILE *out, int id) ;
void assign_order_vals ( int id ) ;
void get_leaves (int id, int **leaf_array, int *counter, int subtree_excluded);
void prepToBuildAlignment(int K);
void cleanAfterAlignmentBuild (int K);
void get_single_edge_treecut_list (int K, int **edges, int num_edges ) ;
void pick_twoedge_treecut (int K, int *edge_i, int *edge_j ) ;
unsigned int treeCutMerge_simple(int K, int space_optimized, char **alignment, int alignlen, 
					int i, int j, char ***merged_alignment, int *cost);
unsigned int treeCutMerge_complex(int K, int space_optimized, char **alignment, int alignlen, 
					int i, int j, char ***merged_alignment, int *cost);
int alterDistanceTable(int id, int K, int ***distances, int **active); 
void fill_compressed_alph_map() ;		
int calc_dist(int aligncost, int i, int j, char **alignment, int len);
void set_weight_matrix_for_cluster (int id) ;		
		
struct option options[] = {
	{"fasta_file", 				required_argument, 		NULL, 					'f'},
	{"cost_file", 				required_argument, 		NULL, 					'c'},
	{"weight_file",				required_argument, 		NULL, 					'w'},
	{"alignment_out_file", 		required_argument, 		NULL, 					'a'},
	{"merge_tree_out_file",		required_argument, 		NULL, 					'm'},	
	{"gamma", 					required_argument, 		NULL, 					'g'},
	{"lambda", 					required_argument, 		NULL, 					'l'},
	{"gamma_term", 				required_argument, 		NULL, 					'd'},
	{"lambda_term", 			required_argument, 		NULL, 					'e'},
	{"output_width",			required_argument, 		NULL, 					'o'}, 
	{"random_repetitions", 		required_argument, 		NULL, 					'r'},
	{"polish_cut_count", 		required_argument, 		NULL, 					'p'},
	{"time_limit", 				required_argument, 		NULL, 					't'},
	{"rand_group1_pct",     	required_argument, 		NULL, 					'x'},
	{"rand_group2_pct",     	required_argument, 		NULL, 					'y'},
	{"help", 					no_argument, 			NULL, 					'h'},		
	{"phases", 					no_argument, 			NULL, 					's'},	
	{"polish_internal",			no_argument,			&polish_internal,		AA_TRUE},
	{"no_polish_internal",		no_argument,			&polish_internal,		AA_FALSE},
	{"local_search", 			no_argument, 			&do_local_search, 		AA_TRUE},		
	{"no_local_search", 		no_argument, 			&do_local_search, 		AA_FALSE},			
	{"verbose", 				no_argument, 			&verbose, 				AA_TRUE},		
	{"quiet", 					no_argument, 			&verbose, 				AA_FALSE},
	{"space", 					no_argument, 			&space_optimized, 		AA_TRUE},		
	{"time", 					no_argument, 			&space_optimized, 		AA_FALSE},
	{"output_fasta", 			no_argument, 			&print_format, 			FASTA_FORMAT},		
	{"output_clustal",			no_argument, 			&print_format, 			CLUSTAL_FORMAT},		
	{"avg_internal", 			no_argument, 			&guide_tree_method, 	AVG_INTERNAL},		
	{"avg_internal_external",	no_argument, 			&guide_tree_method, 	AVG_INTERNAL_EXTERNAL},
	{"total_internal", 			no_argument, 			&guide_tree_method, 	TOTAL_INTERNAL},			
	{"min_internal",			no_argument, 			&guide_tree_method, 	MIN_INTERNAL},
	{"mst",						no_argument, 			&guide_tree_method, 	MST},	
	{"tj",						no_argument, 			&guide_tree_method, 	TJ},
	{"nj",						no_argument, 			&guide_tree_method, 	NJ},	
	{"mst_upgma",				no_argument, 			&guide_tree_method, 	MST_UPGMA},		
	{"upgma",					no_argument, 			&guide_tree_method, 	UPGMA},		
	{"align_optimally",			no_argument,			&align_optimally,		AA_TRUE},
	{"use_pessimistic",			no_argument,			&align_optimally,		AA_FALSE},	
	{"tree_cut", 				no_argument, 			&local_search_method,	TREE_CUT},
	{"3_cut_tree", 				no_argument, 			&local_search_method,	THREE_CUT_TREE},		
	{"any_cut",					no_argument, 			&local_search_method,	ANY_CUT},
	{"2tree_cut_best",			no_argument, 			&two_cut_type,			BEST},
	{"2tree_cut_all",			no_argument, 			&two_cut_type, 			ALL},	 
	{"random_tree_cut", 		no_argument, 			&random_cut_method,		TREE_CUT},	
	{"random_any_cut", 			no_argument, 			&random_cut_method,		ANY_CUT},	
	{"weights_in_merge", 		no_argument, 			&use_weights_in_merge,  AA_TRUE},
	{"no_weights_in_merge", 	no_argument, 			&use_weights_in_merge,  AA_FALSE},	
	{"align_cost_across", 		no_argument, 			&alignment_cost_type, 	ALIGN_COST}, 
	{"align_cost_all", 			no_argument, 			&alignment_cost_type, 	ALL_PAIRS}, 
	{"three_tree_cut_pair", 	no_argument, 			&three_tree_cut_type, 	CUT_PAIR},
	{"three_tree_cut_multi", 	no_argument, 			&three_tree_cut_type, 	CUT_MULTI},	
	{"dist_cost",	 			no_argument, 			&distance_type, 		EDIT}, 
	{"dist_norm_cost",			no_argument, 			&distance_type, 		NORM_EDIT}, 
	{"dist_pct_id",	 			no_argument, 			&distance_type, 		PCT_ID}, 
	{"dist_compr_pct_id",		no_argument, 			&distance_type, 		COMPR_PCT_ID}, 
	{"polish_three_cut",		no_argument, 			&overall_type, 			THREE_CUT}, 	
	{"polish_exhaustive",		no_argument, 			&overall_type, 			EXH_SEARCH}, 	

};

static const char *optstring = "f:c:a:m:g:l:d:e:w:r:p:t:h:x:y:s:";

										
int Compare (Pointer x, Pointer y) // for heap
{
	
   return x - y;
}

int main(int argc, char *argv[])
{

	char 		   *seq_file;
	int				K;

    int             i,
                    j;
    int             printsolution = AA_TRUE;

    int             max_sub = 0;
	char 		   *weight_file = NULL;
    FILE           *align_out = stdout;
    char           *align_outfile = NULL;
    FILE           *mergetree_out = NULL;
    char           *mergetree_outfile = NULL;
	int     	  **cost;/* cost[a,b] = substitution cost     */    

	int     		gamma = OPAL_GAMMA;  /* gap open penalty                  */
	int     		lambda = OPAL_LAMBDA; /* gap extension penalty             */	
	int     		gamma_terminal = OPAL_GAMMA_T;  /* gap open penalty                  */
	int     		lambda_terminal = OPAL_LAMBDA_T; /* gap extension penalty             */	
	int 			outputwidth = AA_OUTPUTWIDTH;
		
	char		 	*alphabet = NULL;	
	int			 	*alphaconv = NULL;
	int 			randomized_polish_reps = -1;
	int				polish_splits_num = 3; 
	
	int c; 
	int tmp;
	extern char *optarg; 
	char *progname = argv[0];

    char *cost_file = aa_malloc(sizeof(char)*(strlen(progname)+20));	/* input cost file name */
    

	char *final = strrchr(progname, '/');  
 	int final_index = final - progname ;
	sprintf (cost_file, "%s", progname); 
    strcpy(cost_file+final_index+1, "BLOSUM62_base100");
		
 	srand( (unsigned)time( NULL ) );     /* Set seed */

	while ((c = getopt_long(argc, argv, optstring, options, NULL)) != EOF) {	
//    while ((c = getopt (argc, argv, "f:c:a:m:g:l:w:r:p:t:xvsou1234567h")) != -1) {

		switch (c) {  
		    case 'f':
				seq_file = optarg;
				break;
		    case 'c':
				cost_file = optarg;
				break;
		    case 'w':
				weight_file = optarg;
				break;
		    case 'h':
				printhelp(progname);
				exit(0);
				break;
		    case 'a':
				if (!(align_out = fopen((char*)optarg, "w"))) {
				    fprintf(stderr, "%s: error opening file: %s\n",
					    progname, (char*)optarg);
				    exit(1);
				} else {
				    align_outfile = 	aa_malloc((strlen((char*)optarg) + 1) * sizeof(char));
				    strcpy(align_outfile, (char*)optarg);
				}
				break;
		    case 'm':
				if (!(mergetree_out = fopen((char*)optarg, "w"))) {
				    fprintf(stderr, "%s: error opening file: %s\n",
					    progname, (char*)optarg);
				    exit(1);
				} else {
				    mergetree_outfile = 	aa_malloc((strlen((char*)optarg) + 1) * sizeof(char));
				    strcpy(mergetree_outfile, (char*)optarg);
				}
				break;				
		    case 'g':
				gamma = atoi((char*)optarg);
				break;
		    case 'l':
				lambda = atoi((char*)optarg);
				break;
		    case 'd':
				gamma_terminal = atoi((char*)optarg);
				break;
		    case 'e':
				lambda_terminal = atoi((char*)optarg);
				break;

		    case 'o':
				outputwidth = atoi((char*)optarg);
				break;
		    case 'r':
				randomized_polish_reps = atoi((char*)optarg);
				break;
		    case 'p':
				tmp = atoi((char*)optarg);
				if (tmp == 2 || tmp == 3) {
					polish_splits_num = tmp;
					allow_both_2way_and_3way_cut = AA_FALSE;
					//otherwise, leave polish_splits_num as the default of 3,
					// and leave "allow..." == TRUE
				}
				break;
		    case 's':
				phases = atoi((char*)optarg);
				break;				
		    case 't':
				timelimit = atoi((char*)optarg);
				break;				
		    case 'x':
				rand_cut_group1_pct = atof((char*)optarg);
				break;								
		    case 'y':
				rand_cut_group2_pct = atof((char*)optarg);
				break;										

			/* the following cases are handled by default with getopts_long, but need to be here if getopts is in use
				-x = no_local_search (default = local_search)
				-v = verbose (default = quiet)
				-s = space efficient (default = time)
				-o = fasta (default = clustal)
				-u = use pessimistic (default = optimal)
				
				-1 avg_internal
				-2 avg_internal_external
				-3 total_internal
				-4 min_internal
				-5 mst
				-6 tj
				-7 upgma

		    case 'x':
				do_local_search = AA_FALSE;
				break;				
		    case 'v':
				verbose = AA_TRUE;
				break;				
		    case 's':
				space_optimized = AA_TRUE;
				break;				
		    case 'o':
				print_format = FASTA_FORMAT;
				break;				
		    case 'u':
				align_optimally = AA_FALSE;
				break;				
		    case '1':
				guide_tree_method =	AVG_INTERNAL;
				break;				
		    case '2':
				guide_tree_method =	AVG_INTERNAL_EXTERNAL;
				break;				
		    case '3':
				guide_tree_method =	TOTAL_INTERNAL;
				break;				
		    case '4':
				guide_tree_method =	MIN_INTERNAL;
				break;				
		    case '5':
				guide_tree_method =	MST;
				break;				
		    case '6':
				guide_tree_method =	TJ;
				break;				
		    case '7':
				guide_tree_method =	UPGMA;
				break;				
			*/
				
			/* End of the cases required if getopt_long is unavailable */				
				
				
		    case 0:
                //printf ("?? getopt returned character code 0 %o ??\n", c);
                break;
		    default:
				Usage(argv[0]);
				break;
		} 
	}

	if ( overall_type == EXH_SEARCH) {
		do_local_search = AA_TRUE;
		local_search_method = TREE_CUT;
		two_cut_type = ALL;
		randomized_polish_reps = 0;
	}
	
	
	if (gamma_terminal < 0 ) {
		gamma_terminal = gamma;
	}
	if (lambda_terminal < 0) {
		lambda_terminal = lambda;	
	}


    if (NULL == seq_file) {
		Usage(argv[0]);
    }


	if ( COMPR_PCT_ID == distance_type) {
		fill_compressed_alph_map();
	}

    /*
     * read in the aa_alphabet and cost matrix 
     */
     if ( verbose)
        printf ("reading cost matrix\n");
         
    if (NULL == cost_file) {
		cost_file = AA_COST_MATRIX_FILE;
    }
    int alpha_size = 0;
    alpha_size = ReadCost(cost_file, &cost, &alphabet, &alphaconv);

    /*in
     * set gamma and lambda to 1/2 max sub cost, if not specified by user 
     */
    if (-1 == gamma || -1 == lambda) {
		for (i = 0; i < AA_MAXALPHA; i++) {
		    for (j = 0; j < AA_MAXALPHA; j++) {
				if (max_sub < cost[i][j]) {
				    max_sub = cost[i][j];
				}
		    }
		}
		if (-1 == gamma) {
		    gamma = max_sub / 2;
		    if (0 != max_sub % 2) {
				gamma++;
		    }
		}
		if (-1 == lambda) {
		    lambda = max_sub / 2;
		    if (0 != max_sub % 2) {
				lambda++;
		    }
		}
    }


	aa_setCosts(&cost, alpha_size, alphabet, alphaconv, gamma, lambda, gamma_terminal, lambda_terminal);

     if ( verbose)
        printf ("reading sequences\n");

    cluster_count = K = ReadSequences(seq_file, &alignments, &lengths, &names);



	//default number of random repititions, if it wasn't set explicitly
	if ( randomized_polish_reps < 0 ) {
		if ( K<3 ) 
			randomized_polish_reps = 0;
		else
			randomized_polish_reps = 3*K;
	
		if ( randomized_polish_reps > 60)
			randomized_polish_reps = 60;
	}


	max_len = 0;
	orig_lengths = aa_malloc(K * sizeof(int));
	for (i=0; i<K; i++){
		orig_lengths[i] = lengths[i];
		max_len = (max_len>lengths[i]) ? max_len : lengths[i];
	}

     if ( verbose)
        printf ("reading weights\n");

	getWeights(weight_file, K, &weights);
//	ReadWeights(weight_file, K, &weights);
	
	char **initial_alignment;
	int phz_cnt = 0;
	int alignlen;
	
	while (phz_cnt < phases) {
		phz_cnt++;
		if (phz_cnt>1) {
			//slow, but easy-to-code, way of resetting sequences
	    	cluster_count = ReadSequences(seq_file, &alignments, &lengths, &names);		
		}

		prepToBuildAlignment(K);
		if ( verbose)
        	printf ("building alignment\n");
			 
		buildAlignment(K, space_optimized, &initial_alignment, &alignlen, phz_cnt);
		cleanAfterAlignmentBuild (K) ; 
	}


    if (verbose) {
		printf("done with initial alignment, begin polishing\n");
    }
	char **polished_alignment;
 	alignlen = polishAlignment(polish_splits_num, K, space_optimized, initial_alignment, alignlen, &polished_alignment, randomized_polish_reps);


    fflush (stdout);
    fflush (stderr);
    // print the alignment
    if (printsolution) {
    	if (print_format == FASTA_FORMAT) {
			printAlignment_fasta(K, polished_alignment, alignlen, align_out, outputwidth, names);
    	} else if (print_format == CLUSTAL_FORMAT) {
			printAlignment_clustal(K, polished_alignment, alignlen, align_out, outputwidth, names);    		
    	}
	    fflush(align_out);
    }

    	
    // print solution alignment cost
    if (verbose) {
 		printcost(space_optimized, gamma, lambda, cost_file, seq_file, K, polished_alignment, alignlen, align_outfile, weights);
    }
    
	if ( mergetree_outfile != NULL ) {
		print_tree(mergetree_out, tree_root);
	    fprintf(stderr, "Solution merge tree written to: %s\n", mergetree_outfile);
	    fprintf(stderr, "--------------------------------------------\n");
	}    
 

    fflush(stdout);
    fflush(stderr);     
 

	/*   free memory   */

	for (i=0; i<K; i++) 
		free(initial_alignment[i]);
	free(initial_alignment);


    for (i = 0; i < K; i++) {
	    if (names[i] != NULL) free(names[i]);
    }
    if (names != NULL) free(names);    
    
	for (i=0; i<alpha_size; i++) 
		free (cost[i]);
	free(cost);
    
	free(orig_lengths);

	free(alphabet);
	free(alphaconv);

	if (NULL != compressed_alph_map) {
		free(compressed_alph_map);	
	}
    

    free(alignment_roots);
    free(tree_nodes);

//    free(alignment_costs);

//	aa_weighted = AA_TRUE; // just to get the next line to run right
	aa_freeWeights();

	for(i=0; i<K; i++) {
		free(polished_alignment[i]);//also frees new_alignment[x] for some other index, since they point to the same string
	}
	free(polished_alignment);

	//free(new_alignment);   - this was seg faulting on twi_31.BLOSUM62_base100.none.min_internal.norm_edit.polish.fasta
		
	free(alignments);
	
	freeHeapPool();
	freeNodePool();	
	
	free(lengths);
	    
    return 0;
}


// only used to assign 1.0 weights for testing, in case no weight file was found
void getWeights(char *weight_file, int K, int ***weights){
 	//placeholder - will look for weights file. If none, do NJ, then one method or another
 	//for now, just set all to 1.0
 	int i,j;

	if (!ReadWeights(weight_file, K, weights) && 
		(aa_gamma_terminal != aa_gamma || aa_lambda_terminal != aa_lambda ) ) {
		*weights = aa_malloc(K * sizeof(int*));
	 	for (i=0; i<K; i++){
	 		(*weights)[i] = aa_malloc(K * sizeof(int));
	 		for (j=0; j<K; j++){
	 			(*weights)[i][j]=1.0;
	 		}
	 	}
	}
}

void print_tree ( FILE *out, int id) {
	if (tree_nodes[id].left_child == -1) { //or, id < K
		fprintf(out, "%s", names[id]);
	} else {
		fprintf (out, "(");		
		print_tree (out, tree_nodes[id].left_child);
		fprintf (out,",");				
		print_tree (out, tree_nodes[id].right_child);		
		fprintf (out,")");
	}
}

void assign_order_vals ( int id) {
	tree_nodes[id].preorder = order_counter++;
	if (tree_nodes[id].left_child >= 0) { //or, id < K
		assign_order_vals(tree_nodes[id].left_child);
		assign_order_vals(tree_nodes[id].right_child);		
	}
	tree_nodes[id].postorder = order_counter++;
}

int setWeights_with_id_lists (int *A_ids, int *B_ids, int K, int L, int ***tmp_weights) {

	if (NULL == weights) {
		aa_setUnweighted();	
		return AA_FALSE;
	}

	int i,j;
	*tmp_weights = aa_malloc( (K+L)*sizeof(int*));
	for (i=0; i<K+L; i++) {
		(*tmp_weights)[i] = aa_malloc((K+L)*sizeof(int));
	}
		
	for (i=0; i<K; i++) {
		for (j=0; j<K; j++) {		
			(*tmp_weights)[i][j] = weights[A_ids[i]][A_ids[j]]; 
		}
		for (j=0; j<L; j++) {
			(*tmp_weights)[i][K+j] = weights[A_ids[i]][B_ids[j]]; 
		}
	}
	for (i=0; i<L; i++) {
		for (j=0; j<K; j++) {		
			(*tmp_weights)[K+i][j] = weights[B_ids[i]][A_ids[j]]; 
		}
		for (j=0; j<L; j++) {
			(*tmp_weights)[K+i][K+j] = weights[B_ids[i]][B_ids[j]]; 
		}
	}
	aa_setWeights(K+L, *tmp_weights);
	return AA_TRUE;	
}

void printweight_info (char **A, char **B, int K, int L, int M, int N, int *A_ids, int *B_ids, int **tmp_weights, int use_weights) {
	int i, j;
	
	printf ("do alignment:\n");
	printf ("aligning seqs (");
	for (i=0; i<K-1; i++) {
		printf ("%d, ", A_ids[i] );
	}
	printf ("%d) and (", A_ids[K-1] );
	for (i=0; i<L-1; i++) {
		printf ("%d, ", B_ids[i] );
	}
	printf ("%d)\n", B_ids[L-1] );
	
	for (i=0; i<K; i++) {
		for (j=0; j<M; j++) {
			printf ("%c", A[i][j]);
		}
		printf ("\n");
	}
	printf ("vs\n");
	for (i=0; i<L; i++) {
		for (j=0; j<N; j++) {
			printf ("%c", B[i][j]);
		}
		printf ("\n");
	}
	
	if (use_weights) {
		printf ("weights:\n");
		for (i=0; i<K+L; i++) {
			for (j=0; j<K+L; j++) {		
				printf("%d\t", tmp_weights[i][j]); 
			}
			printf ("\n");
		}
	}

}

int do_alignment (char ***A, int K, int M, int *A_ids, char ***B, int L, int N, int *B_ids, 
					char ***alignment_ptr, unsigned char **script_ptr, int use_weights ) {
	int **tmp_weights;
	if (use_weights) {
		use_weights = setWeights_with_id_lists (A_ids, B_ids, K, L, &tmp_weights);
	} else if ( aa_gamma != aa_gamma_terminal  || aa_lambda != aa_lambda_terminal ) {
		int i,j;
		tmp_weights = aa_malloc( (K+L) * sizeof(int*));
		for (i=0; i<K+L; i++) {
			tmp_weights[i] = aa_malloc( (K+L) * sizeof(int));
			for (j=0; j<K+L; j++) {		
				tmp_weights[i][j] = i==j?0:1; 
			}
		}		
		aa_setWeights(K+L, tmp_weights);
	} else {
		aa_setUnweighted();	
	}

	//printweight_info (*A, *B, K, L, M, N, A_ids, B_ids, tmp_weights, use_weights);

	int len;

	if (align_optimally) {
		len = alignalign(	space_optimized, A, K, M, B, L, N, alignment_ptr, script_ptr);
	} else {
		
		int m = RemoveNullColumns(*A, K, M);		  
		int n = RemoveNullColumns(*B, L, N);
		
	  	FindFirstAndLastChars (*A, &aa_firstCharsA, &aa_lastCharsA, K, M ) ;
	  	FindFirstAndLastChars (*B, &aa_firstCharsB, &aa_lastCharsB, L, N ) ;    
		
		pessimistic_optimistic_incl_alignment (*A, *B, K, L, m, n, 1.0,  
					alignment_ptr, &len, script_ptr ); 
	}
	
	
	if (use_weights || aa_gamma != aa_gamma_terminal  || aa_lambda != aa_lambda_terminal) {
		int i;
		for (i=0; i<(K+L); i++)		
			free(tmp_weights[i]);
		free(tmp_weights);
	}
	return len;
}


// total_weight = sum of all pairwise weights
// align_weight = sum of the weights across the alignment of A and B
int do_alignment_get_weights (char ***A, int K, int M, int *A_ids, char ***B, int L, int N, int *B_ids, 
					char ***alignment_ptr, unsigned char **script_ptr, int use_weights,
					int *total_weight, int *align_weight ) {

	*total_weight = *align_weight = 0;	
	
	int len = do_alignment (A, K, M, A_ids, B, L, N, B_ids, alignment_ptr, script_ptr, use_weights );

	
	if (use_weights && NULL != weights) {
		int **tmp_weights;
		setWeights_with_id_lists (A_ids, B_ids, K, L, &tmp_weights);

		int i,j;
		for (i=0; i<K; i++)	{
			for (j=0; j<L; j++)	{		
				*align_weight += tmp_weights[i][K+j];
			}
		}	
		*total_weight = *align_weight;
		for (i=0; i<K-1; i++)
			for (j=i+1; j<K; j++)	
				*total_weight += tmp_weights[i][j];
		for (i=0; i<L-1; i++)
			for (j=i+1; j<L; j++)	
				*total_weight += tmp_weights[K+i][K+j];

		for (i=0; i<K+L; i++) {
		  free (tmp_weights[i]);
		}
		free (tmp_weights);


	} else {
		// count of pairs
		*align_weight = K*L;
		*total_weight = (K+L)*(K+L-1)/2;
	}
	
	return len;
}

void prepToBuildAlignment(int K) {
	int i;
    clusters = aa_malloc(K * sizeof (ClusterList));
    cluster_sizes = aa_malloc(K * sizeof(int));

//	alignment_costs = aa_malloc(K * sizeof(int));
	alignment2cluster_map = aa_malloc(K * sizeof(int));	
    ClusterEntry *ce = NULL;
	for (i=0; i<K; i++) {
//		alignment_costs[i] = 0;	
		cluster_sizes[i] = 1;
		
		ce = aa_malloc(sizeof(ClusterEntry));		
		ce->seqid = i;
		ce->next = NULL;  
		clusters[i].first = clusters[i].last = ce;

		alignment2cluster_map[i]=i;		

	}	
	alignment_roots = aa_malloc(K * sizeof(int));
	for (i=0; i<K; i++)
			alignment_roots[i] = i;
	tree_node_count = K;
	tree_nodes = aa_malloc(2 * K * sizeof(Node));
	for (i=0; i<2*K; i++) {
		tree_nodes[i].parent = -1;	
		tree_nodes[i].left_child = -1;
		tree_nodes[i].right_child = -1;		
	}
}

void cleanAfterAlignmentBuild (int K) {

	int i, j;
	DestroyHeap(distance_heap);

	for (i=0; i<K; i++){
		free(distance_nodes[i]);
	}
	free(distance_nodes);


	if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) {
		for (i=0; i<K; i++) {
			for (j=0; j<K; j++) {
				free(internal_external_distances[i][j]);
			}
			free(internal_external_total_distances[i]);
			free(internal_external_numpairs[i]);
			free(internal_external_distances[i]);
		}
		free(internal_external_total_distances);
		free(internal_external_numpairs );
		free(internal_external_distances );
	}


	ClusterEntry *ce = clusters[0].first;
	ClusterEntry *next_ce;
	while (ce != NULL) {
		next_ce = ce->next;	
		free(ce);
		ce = next_ce;
	}    

	//free(clusters);	
	
/*	free(lengths);
    free(cluster_sizes);
	free(alignment2cluster_map);
*/	
}

void buildAlignment(int K, int space_optimized, char ***initial_alignment, int *alignlen, int phase) {
  
  if (phase>1) {
	printf("\n\nBegining phase %d\n----------------\n", phase);
  }
	int i,j,k;
	int aligncost = 0;	

	distance_heap = CreateHeap(Compare);
	
	distance_nodes = aa_malloc(K * sizeof(HeapNode**));
	for (i=0; i<K; i++){
		distance_nodes[i] = aa_malloc(K * sizeof(HeapNode*));
	}
		
	if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) {
		internal_external_total_distances 	= aa_malloc(K * sizeof(int*));
		internal_external_numpairs 			= aa_malloc(K * sizeof(int*));
		internal_external_distances 		= aa_malloc(K * sizeof(int**));
		for (i=0; i<K; i++) {
			internal_external_total_distances[i]= aa_malloc(K * sizeof(int));
			internal_external_numpairs[i]		= aa_malloc(K * sizeof(int));
			internal_external_distances[i] 		= aa_malloc(K * sizeof(int*));
			for (j=0; j<K; j++) {
				internal_external_distances[i][j] = aa_malloc( K * sizeof(int)); // the last element holds the sum
			}
		}
	}
	if (TREE_CUT == local_search_method || THREE_CUT_TREE == local_search_method) {
		tree_cut_optimal_distances	= aa_malloc(K * sizeof(int*));
		for (i=0; i<K; i++) 
			tree_cut_optimal_distances[i] = aa_malloc(K * sizeof(int));
	}
	if ( NJ == guide_tree_method) {
		nj_optimal_distances	= aa_malloc(K * sizeof(int*));
		nj_total_distances      = aa_malloc(K * sizeof(int));
		for (i=0; i<K; i++) {
			nj_optimal_distances[i] = aa_malloc(K * sizeof(int));
			nj_total_distances[i] = 0 ;
		}
	}				


	IDPair *pair;

	unsigned int tmp_len;
	int dist;
	char **res_align;			
	int tmp_cost1, tmp_cost2; // only used for AVG_INTERNAL_EXTERNAL case
	int min_cost = AA_INF, min_i, min_j;//, number_pairs;// only used for AVG_INTERNAL_EXTERNAL case


	for (i=0; i<K; i++){
				
		if ( NJ == guide_tree_method) {
			for (j=0; j<i; j++)
				nj_total_distances[i] += nj_optimal_distances[i][j];
		}

		for (j=i+1; j<K; j++){

			if ( AVG_INTERNAL_EXTERNAL != guide_tree_method ) {

				if (phase == 1) {
					aligncost = calcBestAlignCost(space_optimized, AA_FALSE, AA_FALSE, 
												&alignments[i], 1, lengths[i], &i,
												&alignments[j], 1, lengths[j], &j, 
												&res_align, &tmp_len, NULL, ALL_PAIRS);


				} else {
					//Make a 2-sequence matrix, then use sp to calc score
					char **tmp = aa_malloc(2*sizeof(char*));
					tmp[0] = (*initial_alignment)[i];
					tmp[1] = (*initial_alignment)[j];	
					aa_setUnweighted();		

					aligncost =sp (tmp, *alignlen, 2);

					int ii;
					tmp_len = 0;
					for (ii=0; ii<*alignlen; ii++) {
					  if (tmp[0][ii] != AA_EPS || tmp[1][ii] != AA_EPS ) {
						tmp_len++;
					  }
					}
					free(tmp);
				}

				//dist is used only for the initial merge, and might be based on pct_id
				dist = calc_dist(aligncost, i, j, res_align, tmp_len);


				//				printf ("%d,%d: %d  (%d/%d)\n", i,j,dist, aligncost, tmp_len);

				if (phase == 1) {
					free(res_align[0]);
					free(res_align[1]);			
					free(res_align);			
				}

//				printf("initial distance: %d:%d = %d\n", i, j, dist);
			}


			if (TREE_CUT == local_search_method || THREE_CUT_TREE == local_search_method) 
				tree_cut_optimal_distances[i][j] = tree_cut_optimal_distances[j][i] = aligncost;


			if ( NJ == guide_tree_method) {
				nj_optimal_distances[i][j] = nj_optimal_distances[j][i] = dist;
				nj_total_distances[i] += dist;
			}

			if ( AVG_INTERNAL == guide_tree_method || 
				TOTAL_INTERNAL == guide_tree_method ||
				MIN_INTERNAL == guide_tree_method ||
				MST == guide_tree_method ||
				MST_UPGMA == guide_tree_method ||				
				UPGMA == guide_tree_method ||
				TJ == guide_tree_method ) {
				
				pair = aa_malloc(sizeof(IDPair));
				pair->i = i;
				pair->j = j;

				distance_nodes[j][i] = distance_nodes[i][j] = HeapInsert((Pointer)dist, (Pointer) pair, distance_heap);
				
			} else if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) {
				
				// optimally align i and j (calcBestAlignCost_preservingOrig) ... tmp_align			
				//use initial_alignment temporarily here since it's available.

				tmp_cost1 = calcBestAlignCost_preservingOrig(space_optimized, AA_FALSE, AA_FALSE,
												 alignments[i], 1, lengths[i], &i,
												alignments[j], 1, lengths[j], &j,
												initial_alignment, &tmp_len, NULL, ALL_PAIRS);
				//unweighted because of previous line	
				//tmp_cost1 = sp(*initial_alignment, tmp_len, 2);

				//start with the cost of aligning A & B
				aligncost = 0;
//				number_pairs = 1;				
				
				int id_pair[2];
				id_pair[0] = i;
				id_pair[1] = j;
				for (k=0; k<K; k++) {
					if (k != i && k != j){
						//calc alignmerge cost of tmp_align and k  
						//want to weight this alignment, it's normalized in the called function
/*						tmp_cost2 = calcAlignMergeCost(	space_optimized, AA_TRUE, 
												initial_alignment, 2, tmp_len, tmp_cost1, id_pair,
																&alignments[k], 1, lengths[k], 0, &k );
*/	
						tmp_cost2 = calcBestAlignCost(space_optimized, AA_FALSE, AA_FALSE, 
												initial_alignment, 2, tmp_len, id_pair,
												&alignments[k], 1, lengths[k], &k, 
												NULL, NULL, NULL, ALIGN_COST);
	
						internal_external_distances[i][j][k] = 
								internal_external_distances[j][i][k] = tmp_cost2;				
						aligncost += tmp_cost2 ;
//						number_pairs += 2;
					}
				}
				
				aligncost = (int)(.5+(aligncost/K)); //avg of all the weighted costs
				aligncost += tmp_cost1 ;
				
				internal_external_total_distances[j][i] = 
						internal_external_total_distances[i][j] = aligncost;				
//				internal_external_numpairs[j][i]		= 
//						internal_external_numpairs[i][j]		= number_pairs;				

				//track lowest cost - add just the lowest cost entry to the heap, at the end of the outer loop
				if (aligncost /*/number_pairs*/ < min_cost) {
					min_cost = aligncost; //number_pairs;
					min_i = i;
					min_j = j;
				}				
			}
		}	
	}
	
			
	if ( NJ == guide_tree_method ) {
		// need to pick the min over all i,j of  dist[i][j] - (tot[i]+tot[j])/(K-2)
		min_cost = AA_INF;
		for (i=0; i<K-1; i++){
			for (j=i+1; j<K; j++){
				int reduction = (nj_total_distances[i] + nj_total_distances[j])/(K-2);
				int M = nj_optimal_distances[i][j] - reduction;
				if (M < min_cost) {
					min_cost = M;
					min_i = i;
					min_j = j;
				}					
			}
		}		
		pair = aa_malloc(sizeof(IDPair));
		pair->i = min_i;
		pair->j = min_j;

		HeapInsert((Pointer)min_cost, (Pointer) pair, distance_heap);

	} else if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) {
		//add just the lowest cost entry to the heap.
		// This is the only entry in the heap if the method id internal_external ... just using the infrastructure in place for other methods
		pair = aa_malloc(sizeof(IDPair));
		pair->i = min_i;
		pair->j = min_j;
		HeapInsert((Pointer)min_cost, (Pointer) pair, distance_heap);
	}

	// begin clustering sequences
	int best_i = -1;	
	int best_j = -1;
		 
    while (cluster_count > 1) {  	
    	
    	pair = (IDPair*)HeapDeleteMin(distance_heap);
		best_i = pair->i;
		best_j = pair->j;
		
   		free(pair);
   		
    	//these are the ids of the alignments. 
    	//What I really need are the cluster ids, so use map:
    	int my_i = alignment2cluster_map[best_i];
    	int my_j = alignment2cluster_map[best_j];

		if (my_j < my_i) {
			int tmp = my_i;
			my_i = my_j;
			my_j = tmp;	 
		}
		
		mergeAlignments (K, my_i, my_j, space_optimized, tree_node_count, AA_FALSE);
		tree_root = tree_node_count; // when loop exits, this will be true.
				
		tree_node_count++;
    }
 
 	
 
	//	need to shuffle the alignment around to match the original input of strings
	ClusterEntry *ce = clusters[0].first;
	i = 0;
	*initial_alignment = aa_malloc(K * sizeof(char *));
	*alignlen = lengths[best_i];
	while (ce != NULL) {
		(*initial_alignment)[ce->seqid] = aa_malloc(*alignlen * sizeof(char));
		memcpy((*initial_alignment)[ce->seqid], alignments[best_i][i], *alignlen);
		free(alignments[best_i][i]);
		ce = ce->next;	
		i++;
	}
	
	return 	;	
} 

int calc_dist(int aligncost, int i, int j, char **alignment, int len) {
	int dist;
	
	if (EDIT == distance_type) {
		dist = aligncost;
	} else if (NORM_EDIT == distance_type) {
		int len_sum=0, len_cnt = 0;
		
		ClusterEntry *ce = clusters[i].first;
		while (ce != NULL) {
			len_sum += orig_lengths[ce->seqid];
			len_cnt ++;
			ce = ce->next;	
		}
		ce = clusters[j].first;
		while (ce != NULL) {
			len_sum += orig_lengths[ce->seqid];
			len_cnt++;
			ce = ce->next;	
		}
		
		float avg_len = (float)len_sum/len_cnt;
		dist = (int)(.5+ (max_len*aligncost/avg_len)) ; //multiply by max_len just to scale the ints up to something distinguishable
	} else  { //COMPR_PCT_ID or PCT_ID
	 	//need to check pct_id, possibly over a compressed alphabet
		int x;
		int total_cols = 0;
		int ident_cols = 0;						
		for (x=0; x<len; x++) {
			char a = alignment[0][x];
			char b = alignment[1][x];							
			if (a != AA_EPS && b != AA_EPS) {
				total_cols++;
				if (COMPR_PCT_ID == distance_type) {
					a = compressed_alph_map[(int)a];
					b = compressed_alph_map[(int)b];								
				}
				if (a == b) {
					ident_cols++;
				}
			}
		}
		
		float D = 1.0 - ((float)ident_cols/total_cols);
		//kimura correction
		float tmp = 1 - D - ((D*D)/5);
		if (tmp<0) {
			tmp = 0.00000001;
		}
		dist = (int)(.5 + 1000*(-log(tmp))); // scaling by 1000 to make the ints distinguishable
											  // dist will be 1 if pct_id = 99.9%,   
											  // dist will be ~3000 if pct_id = 18%,   
	}
	
	return dist;
}



void mergeAlignments (int K, int best_i, int best_j, int space_optimized, int new_node_id, int polishing) {
	//merge the closest alignments.
	// If the merge-method calls for it, use alignalign, 
	// In any case, replace the alignment at i_first with the merged alignment
    // Update entries for other alignments as necessary  
    

	int i_first = clusters[best_i].first->seqid;
	int j_first = clusters[best_j].first->seqid;

//printf ("\n *****merging %d and %d  (with dist %d)****\n", i_first, j_first, (int)(distance_nodes[i_first][j_first]->Key) );
	

	int *A_ids = aa_malloc( cluster_sizes[best_i] * sizeof(int));
	int *B_ids = aa_malloc( cluster_sizes[best_j] * sizeof(int));	
	int *AB_ids = aa_malloc( (cluster_sizes[best_i]+cluster_sizes[best_j]) * sizeof(int));	
	int cntr1 = 0, cntr2 = 0;

	if (NULL != weights) {
		ClusterEntry *ce = clusters[best_i].first;
		while (ce != NULL) {
			AB_ids[cntr1] = ce->seqid;
			A_ids[cntr1++] = ce->seqid;
			ce = ce->next;
		}
	
		int cntr2 = 0;
		ce = clusters[best_j].first;
		while (ce != NULL) {
			AB_ids[cntr1+cntr2] = ce->seqid;
			B_ids[cntr2++] = ce->seqid;
			ce = ce->next;
		}
	}	


	int alignlen = do_alignment( &alignments[i_first], cluster_sizes[best_i], lengths[i_first], A_ids, 
					&alignments[j_first], cluster_sizes[best_j], lengths[j_first], B_ids,
					&new_alignment, NULL,AA_TRUE);

//	alignment_costs[i_first] = sp(new_alignment, alignlen, cluster_sizes[best_i]+cluster_sizes[best_j]);
//	alignment_costs[j_first] = 0;


		 
	tree_nodes[new_node_id].left_child = alignment_roots[i_first];
	tree_nodes[new_node_id].right_child = alignment_roots[j_first];	
	tree_nodes[alignment_roots[i_first]].parent = 
				tree_nodes[alignment_roots[j_first]].parent = new_node_id;	
	alignment_roots[i_first] = new_node_id;	
	alignment_roots[j_first] = -1;


	//free old alignments
	int n;
	for(n=0; n<cluster_sizes[best_i]; n++) {
		free(alignments[i_first][n]);
	}
	for(n=0; n<cluster_sizes[best_j]; n++) {
		free(alignments[j_first][n]);
	}	
	
	free(alignments[i_first]);
	free(alignments[j_first]);

	//put new alignment 
	alignments[i_first] = new_alignment;
	lengths[i_first] = alignlen;
	lengths[j_first] = 0;
	
	
	
	//update sizes array to reflect merge, and					
	//append entries from b into a's list in clusters
	cluster_sizes[best_i] += cluster_sizes[best_j];
	clusters[best_i].last->next = clusters[best_j].first;
	clusters[best_i].last = clusters[best_j].last;
	
	alignment2cluster_map[i_first] = best_i;
	alignment2cluster_map[j_first] = -1;
	
	cluster_count--; 

				
	//compact clusters list. At the very least, empty out clusters[best_b]
	//if necessary, shift cluster_count-th entry into b slot.
	if (best_j != cluster_count) { //remove list from b
		clusters[best_j].first = clusters[cluster_count].first;
		clusters[best_j].last = clusters[cluster_count].last;
		cluster_sizes[best_j] = cluster_sizes[cluster_count];
		
		alignment2cluster_map[clusters[cluster_count].first->seqid] = best_j;
	}
	clusters[cluster_count].first = NULL;
	clusters[cluster_count].last = NULL;
	cluster_sizes[cluster_count] = 0;


	//on-the-fly polishing:
	if ( AA_TRUE == polish_internal) {
		polishInternal(best_i);
	}


/**********************************************************
 * 
 * The distances from best_i to all others must be recalculated, and
 * heaps must reflect new scores
 *  
 **********************************************************/

	//For the AVG_* methods, the heap must have entries removed for the two 
	// alignments being merged ... then one inserted (below) for the new alignment
	int i, tmp_i, tmp_j;
	IDPair *pair;
	if ( AVG_INTERNAL == guide_tree_method || 
		MIN_INTERNAL == guide_tree_method ||
		TOTAL_INTERNAL == guide_tree_method ) { // int_ext method doesn't have these extra nodes in the heap
		for (i=0; i<cluster_count; i++) {
			tmp_i = clusters[i].first->seqid;
			if (tmp_i != i_first && tmp_i != j_first) {
				HeapDelete ( distance_nodes[tmp_i][i_first], distance_heap );
				free(  distance_nodes[tmp_i][i_first]->Value); // pair		
	//			free(  distance_nodes[tmp_i][i_first]->Key); // int		

				HeapDelete ( distance_nodes[tmp_i][j_first], distance_heap );			
				free(  distance_nodes[tmp_i][j_first]->Value); // pair			
//				free(  distance_nodes[tmp_i][j_first]->Key); // int
			} 
		}
	} else if ( MST == guide_tree_method ) {
		//For the MIN_* methods, just keep the smaller of the two heap entries, and make sure it's "pair" points to the correct entries.

		for (i=0; i<cluster_count; i++) {
			tmp_i = clusters[i].first->seqid;
			if (tmp_i != i_first && tmp_i != j_first) {
				int x = (int)(distance_nodes[tmp_i][i_first]->Key);
				int y = (int)(distance_nodes[tmp_i][j_first]->Key);
				
				if (x<y) {
					HeapDelete ( distance_nodes[tmp_i][j_first], distance_heap );			
					free(  distance_nodes[tmp_i][j_first]->Value); // pair
					//free(  distance_nodes[tmp_i][j_first]->Key); // int	
				} else { //y<x				
					HeapDelete ( distance_nodes[tmp_i][i_first], distance_heap );
					free(  distance_nodes[tmp_i][i_first]->Value); // pair
					//free(  distance_nodes[tmp_i][i_first]->Key); // int
					//that leaves the j_first entry on the heap, which should now reference (and be pointed to by) the i_first entry
					distance_nodes[tmp_i][i_first] = distance_nodes[i_first][tmp_i] = distance_nodes[tmp_i][j_first];
					pair = (IDPair*)(distance_nodes[tmp_i][i_first]->Value);
					pair->i = i_first < tmp_i ? i_first : tmp_i;
					pair->j = i_first < tmp_i ? tmp_i : i_first;					
				}
			}
		}
		free (A_ids);
		free (B_ids);	
		free(AB_ids);
		return;
	} else if ( TJ == guide_tree_method || MST_UPGMA == guide_tree_method || UPGMA == guide_tree_method ) {
		/*
		 * UPGMA
		 * for each node k, the distance to the new merged alignment (x) of i and j is
		 *		dxk=(dik + djk) /2		
		 * 
		 * TJ
		 * from http://artedi.ebc.uu.se/course/sommar/njoin/  && http://bio.lundberg.gu.se/courses/ht05/bio1/molphylo2.pdf
		 * for each node k, the distance to the new merged alignment (x) of i and j is
		 *		dxk=(dik + djk - dij) /2	
		 * 
		 * 
		 * MST_UPGMA  - weighted mixture of Min spanning tree and UPGMA
		 * for each node k, the distance to the new merged alignment (x) of i and j is
		 *		dxk= .9*(min(dik + djk))  +  .1*((dik + djk)/2)	 * 
		 * 
		 */
		int dij = (int)(distance_nodes[i_first][j_first]->Key);
		for (i=0; i<cluster_count; i++) {
			tmp_i = clusters[i].first->seqid;
			if (tmp_i != i_first && tmp_i != j_first) {
				int dik = (int)(distance_nodes[tmp_i][i_first]->Key);
				int djk = (int)(distance_nodes[tmp_i][j_first]->Key);
				
				HeapDelete ( distance_nodes[tmp_i][i_first], distance_heap );
				free(  distance_nodes[tmp_i][i_first]->Value); // pair
				//free(  distance_nodes[tmp_i][i_first]->Key); // int
				HeapDelete ( distance_nodes[tmp_i][j_first], distance_heap );			
				free(  distance_nodes[tmp_i][j_first]->Value); // pair			
				//free(  distance_nodes[tmp_i][j_first]->Key); // int

				int dxk;
				if ( TJ == guide_tree_method )
					dxk = (dik + djk - dij) /2;
				if ( MST_UPGMA == guide_tree_method )
					dxk = 0.9*(dik<djk?dik:djk) + 0.1*((dik + djk)/2);
				else  //UPGMA
					dxk = (dik + djk) /2;
				
				
				pair = aa_malloc(sizeof(IDPair));
				//pair->i must be less than pair->j, because of later action
				pair->i = i_first < tmp_i ? i_first : tmp_i;
				pair->j = i_first < tmp_i ? tmp_i : i_first;
				
				distance_nodes[tmp_i][i_first] = distance_nodes[i_first][tmp_i] = HeapInsert((Pointer)dxk, (Pointer) pair, distance_heap);
			}
		}
		free (A_ids);
		free (B_ids);	
		free(AB_ids);
		return;
	} else if ( NJ == guide_tree_method ) {
		int k;
		nj_total_distances[i_first] = 0;

		for (k=0; k<cluster_count; k++) {
			int tmp_k = clusters[k].first->seqid;
			if ( tmp_k==i_first )
				continue;
			int tmp = (nj_optimal_distances[tmp_k][i_first] + nj_optimal_distances[tmp_k][j_first] - nj_optimal_distances[i_first][j_first]) / 2;
			nj_total_distances[tmp_k] -= nj_optimal_distances[tmp_k][i_first] + nj_optimal_distances[tmp_k][j_first];
			nj_total_distances[tmp_k] += tmp;
			nj_total_distances[i_first] += tmp;
			nj_optimal_distances[i_first][tmp_k] = nj_optimal_distances[tmp_k][i_first] = tmp;
		}
				
		int min_cost = AA_INF;
		int j, min_i, min_j;
		
		if (cluster_count == 1) {
			min_i = min_j = -1;
		} else if (cluster_count == 2) {
			min_i = clusters[0].first->seqid;
			min_j = clusters[1].first->seqid;			
		} else {
			// need to pick the min over all i,j of  dist[i][j] - (tot[i]+tot[j])/(K-2)
			for (i=0; i<cluster_count-1; i++){
				tmp_i = clusters[i].first->seqid;
				for (j=i+1; j<cluster_count; j++){
					tmp_j = clusters[j].first->seqid;
					int reduction = (nj_total_distances[tmp_i] + nj_total_distances[tmp_j])/(cluster_count-2);
					int M = nj_optimal_distances[tmp_i][tmp_j] - reduction;
					if (M < min_cost) {
						min_cost = M;
						min_i = tmp_i;
						min_j = tmp_j;
					}					
				}
			}
		}
		
		pair = aa_malloc(sizeof(IDPair));
		pair->i = min_i;
		pair->j = min_j;
		HeapInsert((Pointer)min_cost, (Pointer) pair, distance_heap);

		free (A_ids);
		free (B_ids);	
		free(AB_ids);
		return;

	}


	

	//Calc distance of new alignment to all others. 
	// Add an entry for the new alignment to the heap
	int j,k;
	int i_align, j_align;
	int aligncost = 0;
	
	unsigned int tmp_len;
	int tmp_cost1, tmp_cost2;

	int new_count,  k_first; // only used for AVG_INTERNAL_EXTERNAL case
	int min_cost = AA_INF;
	int min_i=-1, min_j=-1; //, number_pairs;// only used for AVG_INTERNAL_EXTERNAL case

	int *C_ids = aa_malloc( K * sizeof(int));
	int *D_ids = aa_malloc( K * sizeof(int));

	int dist;
	char **res_align;			

    ClusterEntry *ce = NULL;
	for (j=0; j<cluster_count; j++) {
		j_align = clusters[j].first->seqid;
	
		if (i_first != j_align) {

			if (NULL != weights) {
				ce = clusters[j].first;
				cntr1=0;
				while (ce != NULL) {
					C_ids[cntr1++] = ce->seqid;
					ce = ce->next;
				}
			}

			if ( AVG_INTERNAL == guide_tree_method || 
				MIN_INTERNAL == guide_tree_method ||
				TOTAL_INTERNAL == guide_tree_method ) {		

				int cost_type = (MIN_INTERNAL == guide_tree_method) ? MIN_COST : alignment_cost_type ;
				int get_avg = (AVG_INTERNAL == guide_tree_method) ? AA_TRUE : AA_FALSE ;
				aligncost = calcBestAlignCost(space_optimized, use_weights_in_merge, get_avg, &alignments[i_first], cluster_sizes[best_i], lengths[i_first], AB_ids,
															&alignments[j_align], cluster_sizes[j], lengths[j_align], C_ids, &res_align, &tmp_len,NULL,
															cost_type);		
															
				//dist is used only for the initial merge, and might be based on pct_id
				dist = calc_dist(aligncost, i_first, j, res_align, tmp_len);

				free(res_align[0]);
				free(res_align[1]);			
				free(res_align);			
		
//				printf("merging in %d would have distance %d\n\n=====================\n", j_align, dist);
				
				pair = aa_malloc(sizeof(IDPair));
				//pair->i must be less than pair->j, because of later action
				pair->i = i_first < j_align? i_first : j_align;
				pair->j = i_first < j_align? j_align : i_first;

				distance_nodes[j_align][i_first] = distance_nodes[i_first][j_align] = HeapInsert((Pointer)dist, (Pointer) pair, distance_heap);
				
			} else if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) {
				if (cluster_count == 2) {
					min_i = i_first;
					min_j = j_align;
					break;	
				}

				// optimally align i_first and j_first (calcBestAlignCost_preservingOrig) ... tmp_align
				// reuse new_alignment here, it's a free variable
				tmp_cost1 = calcBestAlignCost_preservingOrig(space_optimized, use_weights_in_merge, AA_TRUE, 
												alignments[i_first], cluster_sizes[best_i], lengths[i_first], AB_ids,
												alignments[j_align], cluster_sizes[j], lengths[j_align], C_ids,
												&new_alignment, &tmp_len, NULL, ALIGN_COST);				
/*				tmp_cost1 = calcBestAlignCost_preservingOrig(space_optimized,alignments[i_first], cluster_sizes[best_i], lengths[i_first], AB_ids,
												alignments[j_align],cluster_sizes[j], lengths[j_align], C_ids,
												&new_alignment, &tmp_len, NULL);					
*/
				new_count = cluster_sizes[best_i]+cluster_sizes[j];
				//tmp_cost1 = sp(new_alignment, tmp_len, new_count);
				aligncost = 0;
//				number_pairs = 0;

				if (NULL != weights) {
					int cntr;
					for (cntr=0; cntr<cluster_sizes[j]; cntr++) { 
						AB_ids[cntr+cluster_sizes[best_i]] = C_ids[cntr];
					} // really, this is ABC_ids
				}
				for (k=0; k<cluster_count; k++) {					
					k_first = clusters[k].first->seqid;

					if (NULL != weights) {
						ce = clusters[k].first;
						cntr1=0;
						while (ce != NULL) {
							D_ids[cntr1++] = ce->seqid;
							ce = ce->next;
						}
					}
					if (k_first != i_first && k_first != j_align) {
						//calc alignmerge cost of tmp_align and k (calcBestAlignCost_preservingOrig) 
						tmp_cost2 = calcBestAlignCost(space_optimized, use_weights_in_merge, AA_TRUE,
																&new_alignment, new_count, tmp_len, AB_ids,
																&alignments[k_first], cluster_sizes[k], lengths[k_first], D_ids,
																NULL, NULL, NULL, ALIGN_COST );
							
/*						tmp_cost2 = calcAlignMergeCost(	space_optimized, &new_alignment, new_count, tmp_len, tmp_cost1, AB_ids,
																&alignments[k_first], cluster_sizes[k], lengths[k_first], alignment_costs[k_first], D_ids );
*/						
						internal_external_distances[i_first][j_align][k_first] = 
									internal_external_distances[j_align][i_first][k_first] = tmp_cost2;				

//						number_pairs += cluster_sizes[k] * new_count ;
						aligncost += tmp_cost2;			
					}
				}
				
				aligncost = (int)(.5 + (aligncost/cluster_count));
				aligncost += tmp_cost1 ;

//				int num_new_pairs = cluster_sizes[best_i] * cluster_sizes[j];
				//add in the cost of aligning A & B
//				aligncost += tmp_cost1 - (alignment_costs[i_first] + alignment_costs[j_align]) ;/// num_new_pairs; // cost of A aligned to B
				//...and the # pair in that alignment
//				number_pairs += num_new_pairs;
				
				internal_external_total_distances[j_align][i_first] = 
						internal_external_total_distances[i_first][j_align] = aligncost;				
/*				internal_external_numpairs[j_align][i_first]		= 
						internal_external_numpairs[i_first][j_align]		= number_pairs;				
*/
				//track lowest cost - add just the lowest cost entry to the heap, at the end of the outer loop
 				if (  (aligncost /*/number_pairs*/) < min_cost) {
					min_cost = aligncost; //number_pairs;
					min_i = i_first;
					min_j = j_align;
				}								
			}				
		}
	}
	
	//if internal_external method, we need to revise all other pairs to account for the loss of i and j,
	// and the creation of a new i. 
	if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) { 
		int x,y; 
//		int number_pairs;
		if (cluster_count > 2) {
		
			if (NULL != weights) {
				free(C_ids);
				C_ids = aa_malloc( cluster_sizes[best_i] * sizeof(int));	
				ce = clusters[best_i].first;
				cntr1=0;
				while (ce != NULL) {
					C_ids[cntr1++] = ce->seqid;
					ce = ce->next;
				}
			}			
			for (i=0; i<cluster_count; i++){
				i_align = clusters[i].first->seqid;
				if (i_align != i_first) {
					for (j=i+1; j<cluster_count; j++){
						j_align = clusters[j].first->seqid;
						if (j_align != i_first) {					
							//need these old distances removed before adding the new distances
							x = internal_external_distances[j_align][i_align][i_first];
							y = internal_external_distances[j_align][i_align][j_first];
			
							if (NULL != weights) {
								free(A_ids);
								A_ids = aa_malloc( cluster_sizes[i] * sizeof(int));
								free(AB_ids);
								AB_ids = aa_malloc( (cluster_sizes[i]+cluster_sizes[j]) * sizeof(int));	
								cntr1 = 0;
								ClusterEntry *ce = clusters[i].first;
								while (ce != NULL) {
									AB_ids[cntr1] = ce->seqid;
									A_ids[cntr1++] = ce->seqid;
									ce = ce->next;
								}
								cntr2 = 0;
								free(B_ids);
								B_ids = aa_malloc( cluster_sizes[j] * sizeof(int));	
								ce = clusters[j].first;
								while (ce != NULL) {
									AB_ids[cntr1+cntr2] = ce->seqid;
									B_ids[cntr2++] = ce->seqid;
									ce = ce->next;
								}
							}	
			
			
							//optimally align i and j, then align result with new i_first, z = cost
							tmp_cost1 = calcBestAlignCost_preservingOrig(space_optimized, use_weights_in_merge, AA_TRUE, 
													alignments[i_align], cluster_sizes[i], lengths[i_align], A_ids,
													alignments[j_align], cluster_sizes[j], lengths[j_align], B_ids,
													&new_alignment, &tmp_len, NULL, ALIGN_COST);				
/*							tmp_cost1 = calcBestAlignCost_preservingOrig(space_optimized,alignments[i_align], cluster_sizes[i], lengths[i_align], A_ids,
															alignments[j_align],cluster_sizes[j], lengths[j_align], B_ids,
															&new_alignment, &tmp_len, NULL);					
*/
							new_count = cluster_sizes[i]+cluster_sizes[j];
							//tmp_cost1 = sp(new_alignment, tmp_len, new_count);
							tmp_cost2 = calcBestAlignCost( space_optimized, use_weights_in_merge, AA_TRUE,
																&new_alignment, new_count, tmp_len, AB_ids,
																&alignments[i_first], cluster_sizes[best_i], lengths[i_first], C_ids, 
																NULL, NULL, NULL, ALIGN_COST );
			
/*							tmp_cost2 = calcAlignMergeCost(	space_optimized, &new_alignment, new_count, tmp_len, tmp_cost1, AB_ids, 
																			&alignments[i_first], cluster_sizes[best_i], lengths[i_first], alignment_costs[i_first], C_ids );
*/			
							internal_external_distances[j_align][i_align][i_first] = 
									internal_external_distances[i_align][j_align][i_first] = tmp_cost2;
							internal_external_distances[j_align][i_align][j_first] = 
									internal_external_distances[i_align][j_align][j_first] = 0;				
							
							aligncost = internal_external_total_distances[i_align][j_align];
//							number_pairs = internal_external_numpairs[i_align][j_align];
							
							aligncost -= (x+y);
							aligncost *= (cluster_count-1);
							aligncost += tmp_cost2;
							aligncost = (int)(.5 + (aligncost/cluster_count));
							
							internal_external_total_distances[j_align][i_align] =
									internal_external_total_distances[i_align][j_align] = aligncost;
											
	/*						number_pairs -= ( new_count*old_cluster_size_i + new_count*old_cluster_size_j);
							number_pairs += ( new_count*cluster_sizes[i_first] );
	*/		
/*							internal_external_numpairs[j_align][i_align] = 
									internal_external_numpairs[i_align][j_align] = number_pairs;
*/							
							if (aligncost /*/number_pairs*/ < min_cost) {
								min_cost = aligncost; //number_pairs;
								min_i = i_align;
								min_j = j_align;
							}												
						}
					}
				}
			}
		}

		
		//add just the lowest cost entry to the heap.
		// This is the only entry in the heap if the method id internal_external ... just using the infrastructure in place for other methods
		pair = aa_malloc(sizeof(IDPair));
		pair->i = min_i < min_j ? min_i : min_j;
		pair->j = min_i < min_j ? min_j : min_i;

		HeapInsert((Pointer)min_cost, (Pointer) pair, distance_heap);
		
	}

	free (A_ids);
	free (B_ids);	
	free(AB_ids);
	free(C_ids);
	free(D_ids);	
}

/*int calcAlignMergeCost ( int space_optimized, int use_weights, char ***A, int K, int M, int aligncostA, int *A_ids, char ***B, int L, int N, int aligncostB, int *B_ids, int cost_type ALL_PAIRS,ALIGN_COST,MIN_COST) {
//	return 	calcAlignMergeCostHelper ( space_optimized, use_weights, A, K, M, aligncostA, A_ids, B, L, N, aligncostB, B_ids, AA_TRUE);
	return calcBestAlignCostHelper(space_optimized, use_weights, A, K, M, A_ids, B, L, N, B_ids, NULL, NULL, NULL, cost_type);
}
*/
/*int calcAlignMergeCostHelper ( int space_optimized, int use_weights, char ***A, int K, int M, int aligncostA, int *A_ids, char ***B, int L, int N, int aligncostB, int *B_ids, int allPairs) {
	int total_weight;	
	int aligncost = calcBestAlignCostHelper(space_optimized, use_weights, A, K, M, A_ids, B, L, N, B_ids, NULL, NULL, NULL, &total_weight, allPairs);
//	aligncost -= (aligncostA + aligncostB);
	return aligncost;
}
*/

int calcBestAlignCost(int space_optimized, int use_weights, int get_avg, char ***A, int K, int M, int *A_ids, char ***B, int L, int N, int *B_ids, char ***result_alignment, 
																unsigned int *result_alignlen, unsigned char **script, int cost_type /*ALL_PAIRS,ALIGN_COST,MIN_COST*/)
{
/*	return calcBestAlignCostHelper(space_optimized, use_weights, A, K, M, A_ids, B, L, N, B_ids, result_alignment, result_alignlen, script, cost_type );
}


int calcBestAlignCostHelper(int space_optimized, int use_weights, char ***A, int K, int M, int *A_ids, char ***B, int L, int N, int *B_ids, char ***result_alignment, 
																unsigned int *result_alignlen, unsigned char **script, int cost_type ALL_PAIRS,ALIGN_COST,MIN_COST)
{
*/
	char **tmp_alignment;
	int total_weight;
	int align_weight;	
	
	unsigned int alignlen = do_alignment_get_weights( A, K, M, A_ids, B, L, N, B_ids, 
								&tmp_alignment, script, use_weights, &total_weight, &align_weight);
	int aligncost;
	
	// the "do_alignment" call sets the weights as needed, so no need to reset them here
	if (MIN_COST == cost_type ) {
		aa_setUnweighted();
		aligncost = min_pair(tmp_alignment, alignlen, K,L);  		
	} else if (ALL_PAIRS == cost_type ) {
		aligncost = sp(tmp_alignment, alignlen, K+L);  
		if (get_avg) {
			aligncost = aligncost/total_weight;
		}
	} else if ( ALIGN_COST == cost_type) {
		aligncost = alignmentCost(tmp_alignment,tmp_alignment+K,K,L,alignlen,alignlen);
		if (get_avg) {
			aligncost = aligncost/align_weight;
		}
	}



	if (result_alignment != NULL && result_alignlen != NULL) {
		*result_alignment = tmp_alignment;
		*result_alignlen = alignlen;
	} else {
		int n;
		for(n=0; n<K+L; n++) {
			free(tmp_alignment[n]);
		}
		free(tmp_alignment);
	}
	return aligncost;
}

int calcBestAlignCost_preservingOrig (int space_optimized, int use_weights, int get_avg, char **A, int K, int M, int *A_ids, 
 																	char **B, int L, int N, int *B_ids,
 																	char ***result_alignment, unsigned int *result_alignlen,
 																	unsigned char **script, int cost_type /*ALL_PAIRS,ALIGN_COST,MIN_COST*/) 
{
	// need to preserve original (by copying to another var)
	//   because spaces get removed during alignment

	int i;
	char **tmpA = aa_malloc(K*sizeof(char*));
	for (i=0; i<K; i++) {
		tmpA[i] =aa_malloc(M * sizeof(char));
		memcpy(tmpA[i],A[i],M); 
	}
				
	char **tmpB = aa_malloc(L*sizeof(char*));
	for (i=0; i<L; i++) {
		tmpB[i] =aa_malloc(N * sizeof(char));
		memcpy(tmpB[i],B[i],N);					
	}	
	
	int result = calcBestAlignCost(space_optimized, use_weights, get_avg, &tmpA, K, M, A_ids, &tmpB, L, N, B_ids, result_alignment, result_alignlen, script, cost_type);
	
	for (i=0; i<K; i++) {
		free(tmpA[i]);
	}
	free(tmpA);
	
	for (i=0; i<L; i++) {
		free(tmpB[i]);
	} 
	free(tmpB);

	return result;
}

//returns true if at least one polishing step was accepted.
int polishInternal(int cluster_id) {
	
	int total_leaves = cluster_sizes[cluster_id];				

    if (total_leaves < 3) {//at least 3 leaves under node_id
		return AA_FALSE;
    }
	
	int *seq_ids = aa_malloc(total_leaves * sizeof(int));
	
	int first_seq_id = clusters[cluster_id].first->seqid;
	int subtree_root = alignment_roots[first_seq_id];
	char **best_alignment = alignments[first_seq_id];
	
	int best_length = lengths[first_seq_id];

	set_weight_matrix_for_cluster ( cluster_id );
						
	int best_cost =  sp(best_alignment, best_length, total_leaves);// alignment_costs[first_seq_id];


	int new_cost;	

	if (verbose) { printf ("performing internal polish on subtree containing seqs: ");}	

	int i=0;
	ClusterEntry *ce = clusters[cluster_id].first;
	while ( NULL != ce ) {
		seq_ids[i++] = ce->seqid;
		if (verbose) {printf ("%d, ", ce->seqid);}
		ce = ce->next;
	}
	if (verbose) { printf  ("\n\tStarts with cost %d\n",best_cost);}	

		
	//if (verbose) printf ("internal polish best_cost: %d\n",best_cost);	
	
    int edges[4];
    int leaves_before[4];
    int leaves_under[4];
    int left_subtree_leafcount;
	int tmp_leafcount;
    	    
	int edge_count;
	int skip_edge = -1;
	
	//for each grandchild of root, try a two way split/merge
	int left_kid = tree_nodes[subtree_root].left_child;
	leaves_before[0] = 0;
	if (tree_nodes[left_kid].left_child == -1 )	{
		//just one kid on the left (not a full subtree)
		edges[0] = left_kid; 
	    leaves_before[1] = 1;	    
	    leaves_under[0] = 1;
	    edge_count = 1;
		left_subtree_leafcount = 1;
		skip_edge = 0;
	} else {
		edges[0] = tree_nodes[left_kid].left_child;
		tmp_leafcount = 0;
	    get_leaves (edges[0], NULL, &tmp_leafcount, -1);
	    leaves_under[0] = tmp_leafcount;
	    leaves_before[1] = tmp_leafcount;
		

		edges[1] = tree_nodes[left_kid].right_child;			
		tmp_leafcount = 0;
		get_leaves (edges[1], NULL, &tmp_leafcount, -1);
	    leaves_under[1] = tmp_leafcount;
	    leaves_before[2] = leaves_before[1] + tmp_leafcount; //aka get_leaves(left_kid)
		edge_count = 2;	

		left_subtree_leafcount = leaves_under[0] + leaves_under[1];
	}

	int right_kid = tree_nodes[subtree_root].right_child;
	if (tree_nodes[right_kid].left_child == -1 )	{			
		//just one kid on the left (not a full subtree)
		edges[edge_count] = right_kid;
		leaves_under[edge_count] = 1;			
		skip_edge = 2; // the first two edges must be ok, or there wouldn't be >=3 edges under current node.
	} else {
		edges[edge_count] = tree_nodes[right_kid].left_child;
		tmp_leafcount = 0;
		get_leaves (edges[edge_count], NULL, &tmp_leafcount, -1);
		leaves_under[edge_count] = tmp_leafcount;
		
		edge_count++;	
		leaves_before[edge_count] = leaves_before[edge_count-1] + tmp_leafcount;
		edges[edge_count] = tree_nodes[right_kid].right_child;
		
		tmp_leafcount = 0;			
		get_leaves (edges[edge_count], NULL, &tmp_leafcount, -1);
		leaves_under[edge_count] = tmp_leafcount;	
	}

	int j;
	int first_pass = AA_TRUE;
	int change_accepted_ever = AA_FALSE;
	int change_accepted = AA_FALSE;


	int *A_ids;
	int *B_ids;
	while (first_pass || change_accepted) {	
		change_accepted = AA_FALSE;

		if ( !first_pass )  {
			first_pass = AA_FALSE;	 
			//do a cut across the halves of the subtree
			int A_size = left_subtree_leafcount;
			int B_size = total_leaves - left_subtree_leafcount;
			char **tmpA = aa_malloc(A_size*sizeof(char*));
			char **tmpB = aa_malloc(B_size*sizeof(char*));


			int a_pos=0;
			int b_pos=0;


			A_ids = aa_malloc(A_size * sizeof(int));
			B_ids = aa_malloc(B_size * sizeof(int));

			for (j=0; j<A_size; j++) {
				tmpA[a_pos] = aa_malloc(best_length * sizeof(char));
				memcpy(tmpA[a_pos],best_alignment[j],best_length);
				if (NULL != weights) { A_ids[a_pos] = seq_ids[j];}
				a_pos++;
			}
			for (j=A_size; j<total_leaves; j++) {
				tmpB[b_pos] = aa_malloc(best_length * sizeof(char));
				memcpy(tmpB[b_pos],best_alignment[j],best_length);					
				if (NULL != weights) { B_ids[b_pos] = seq_ids[j];}
				b_pos++;
			}
			
			char **tmp_alignment;
/*			int new_length = do_alignment( &tmpA, A_size, best_length, A_ids, 
											&tmpB, B_size, best_length, B_ids,
											&tmp_alignment, NULL, AA_TRUE);		
			new_cost = sp(tmp_alignment, new_length, A_size+B_size); 
*/
			unsigned int new_length; 
			new_cost = calcBestAlignCost(space_optimized, AA_TRUE, AA_FALSE, 
								&tmpA, A_size, best_length, A_ids, 
								&tmpB, B_size, best_length, B_ids,
								&tmp_alignment, &new_length, NULL, ALL_PAIRS);
			
			if (NULL != weights) {			
				free(A_ids);
				free(B_ids);
			}
				
			if (new_cost < best_cost) {
				for (j=0; j<total_leaves; j++) 
					free (best_alignment[j]);
				free (best_alignment);
				best_alignment = alignments[first_seq_id] = tmp_alignment;
				best_length = new_length;
				best_cost = new_cost;
				change_accepted = AA_TRUE;
				change_accepted_ever = AA_TRUE;
				if (verbose) {
					printf  ("\timprove to cost %d (at top edge)\n", best_cost);
				}
			} else {
				for (j=0; j<total_leaves; j++) 
					free (tmp_alignment[j]);
				free (tmp_alignment);				
			}
			 
			 
			for (j=0; j<A_size; j++) 
				free(tmpA[j]);
			free(tmpA);
			
			for (j=0; j<B_size; j++) 
				free(tmpB[j]);
			free(tmpB);	
			 
		}

		
		for (i=0; i<=edge_count; i++) {
			//make two alignments:
			//	1st consists of leafcount(edges[i]) seqs, starting at the leaves_before[i]'th sequence
			//	2nd consists of the first leaves_before[i] seqs and the last (subtree_leafcount-leaves_before[i]-leafcount(edges[i]), starting at leaves_before[i].
			
			if (i==skip_edge) 
				continue; // this is the edge that was just used for merging the left and right groups ... no need to do it again.	

			int A_size = leaves_under[i];
			int B_size = total_leaves - leaves_under[i];
				
			char **tmpA = aa_malloc(A_size*sizeof(char*));
			char **tmpB = aa_malloc(B_size*sizeof(char*));
			int a_pos=0;
			int b_pos=0;

			A_ids = aa_malloc(A_size * sizeof(int));
			B_ids = aa_malloc(B_size * sizeof(int));

			//All leaves not under edge i go into tmpB, while those under i go into tmpA
			for (j=0; j<leaves_before[i]; j++) {
				tmpB[b_pos] = aa_malloc(best_length * sizeof(char));
				memcpy(tmpB[b_pos],best_alignment[j],best_length);					
				B_ids[b_pos++] = seq_ids[j];
			}
			for (j=leaves_before[i]; j<leaves_before[i]+leaves_under[i]; j++) {
				tmpA[a_pos] = aa_malloc(best_length * sizeof(char));
				memcpy(tmpA[a_pos],best_alignment[j],best_length);					
				A_ids[a_pos++] = seq_ids[j];
			}
			for (j=leaves_before[i]+leaves_under[i]; j<total_leaves; j++) {
				tmpB[b_pos] = aa_malloc(best_length * sizeof(char));
				memcpy(tmpB[b_pos],best_alignment[j],best_length);
				B_ids[b_pos++] = seq_ids[j];					
			}

			char **tmp_alignment;
			unsigned int new_length;
			new_cost = calcBestAlignCost(space_optimized, AA_TRUE, AA_FALSE, 
								&tmpA, A_size, best_length, A_ids, 
								&tmpB, B_size, best_length, B_ids,
								&tmp_alignment, &new_length, NULL, ALL_PAIRS);			
			free(A_ids);
			free(B_ids);
			
			if (new_cost < best_cost) {
				best_length = new_length;
				best_cost = new_cost;

				for (j=0; j<total_leaves; j++) 
					free (best_alignment[j]);

				//copy seqs back into the alignment ... but in the correct order
				for (j=0; j<leaves_under[i]; j++) {
					alignments[first_seq_id][j+leaves_before[i]] = tmp_alignment[j];
				}
				for (j=leaves_under[i]; j<leaves_under[i]+leaves_before[i]; j++) {
					alignments[first_seq_id][j-leaves_under[i]] = tmp_alignment[j];
				}			
				for (j=leaves_under[i]+leaves_before[i]; j<total_leaves; j++) {
					alignments[first_seq_id][j] = tmp_alignment[j];
				}
				free (tmp_alignment);
				change_accepted = AA_TRUE;
				change_accepted_ever = AA_TRUE;
				if (verbose) {
					printf  ("\timprove to cost %d (at grandchild edge)\n", best_cost);
				}				
			} else {
				for (j=0; j<total_leaves; j++) 
					free (tmp_alignment[j]);
				free (tmp_alignment);			
			}			
			
			for (j=0; j<A_size; j++) 
				free(tmpA[j]);
			free(tmpA);
			
			for (j=0; j<B_size; j++) 
				free(tmpB[j]);
			free(tmpB);	
		}    
		
		first_pass = AA_FALSE;
    }
    
	if (change_accepted_ever) {
	    //alignment_costs[first_seq_id] = best_cost; fix this
	    lengths[first_seq_id] = best_length;
	}

	free (seq_ids);
	
	return change_accepted_ever;
}

unsigned int polishAlignment (int num_cuts, int K, int space_optimized, char** alignment, int alignlen, char ***polished_alignment, int randomized_polish_reps) {

	int i,j,k,m;
	int best_cost;
	unsigned int best_alignlen;	
	
		
	time (&start);	
	
	if ( do_local_search ) {
		if ( TREE_CUT == local_search_method || THREE_CUT_TREE == local_search_method) {
		  /*int **tmp = aa_malloc(2*sizeof(int*));
			tmp[0] = aa_malloc(2*sizeof(int));
			tmp[1] = aa_malloc(2*sizeof(int));
		  */
			induced_ij = aa_malloc(K*sizeof(int*));
			for (i=0; i<K; i++)
				induced_ij[i] = aa_malloc(K*sizeof(int));
						
			int x;
			aa_setUnweighted();
			for (i=0; i<K-1; i++) {
				for (j=i+1; j<K; j++) {	
					x = alignmentCost(alignment+i, alignment+j, 1, 1, alignlen, alignlen);
					induced_ij[i][j] = induced_ij[j][i] = x;
				} 
			} 				
		}	
		best_alignlen = localSearch (num_cuts, K, space_optimized, alignment, alignlen, polished_alignment, &best_cost);
		if (verbose) printf ("localsearch1 cost: %d\n",best_cost);	
	} else {
		best_alignlen = alignlen;
		aa_setWeights(K,weights);
		best_cost = sp(alignment, alignlen, K);  
		if (verbose) printf ("no localsearch cost: %d\n",best_cost);	
		*polished_alignment = aa_malloc(K*sizeof(char*));
		for (i=0; i<K; i++) {
			(*polished_alignment)[i] = aa_malloc(alignlen * sizeof(char));
			memcpy((*polished_alignment)[i],alignment[i],alignlen); 
		}
	}
	
	
	if (randomized_polish_reps == 0 && timelimit <= 0) {
	  if ( do_local_search ) {
		if ( TREE_CUT == local_search_method || THREE_CUT_TREE == local_search_method) {
		  for (i=0; i<K; i++)
			free(induced_ij[i]);
		  free(induced_ij);
		}
	  }				
	  return best_alignlen;
	}	

	char ***splits = aa_malloc(num_cuts * sizeof(char**));
	for(i=0; i<num_cuts; i++) 
		splits[i] = aa_malloc(K * sizeof(char*));
	
	int **maps = aa_malloc(num_cuts * sizeof(int*));
	for(i=0; i<num_cuts; i++) 
		maps[i] = aa_malloc(K * sizeof(int));

	int *split_sizes = aa_malloc(num_cuts * sizeof(int));

	int new_cost;
	unsigned int new_alignlen1;
	unsigned int new_alignlen2;	
	char **tmp_alignment2;	
 	int which_merge; // 3-way only

	if (timelimit > 0  && randomized_polish_reps <= 0) {
		//rather than running a fixed number of iterations, run until the clock runs out.
		//  so set number of iterations to something huge, 
		//  then check at the beginning of each iteration if timer has expired
		randomized_polish_reps = 100000;
	}	

	for (i=1; i<=randomized_polish_reps; i++) {

		if (timelimit > -1) {
			time (&now);
			if (difftime (now,start) > timelimit) {
				fprintf (stderr, "time limit of %d exceeded. Ending execution\n",timelimit);
				break;	
			}
		}

		//	if (verbose) 
		printf("random repetition %d (of %d)\n", i, randomized_polish_reps);
		//pick a random split and align the cuts

		randomlySplit(num_cuts, K, polished_alignment, best_alignlen, &splits,
						 &split_sizes, &maps);
													 
		if ( num_cuts == 2) { 
			which_merge = 1;
			new_cost = calcBestAlignCost(space_optimized, AA_TRUE, AA_FALSE, 
								splits+0, split_sizes[0], best_alignlen, maps[0],
								splits+1, split_sizes[1], best_alignlen, maps[1],								
								&tmp_alignment2, &new_alignlen2, NULL, ALL_PAIRS);
/*			
			new_alignlen2 = do_alignment( splits+0, split_sizes[0], best_alignlen, maps[0], 
							splits+1, split_sizes[1], best_alignlen, maps[1],
							&tmp_alignment2, NULL, AA_TRUE);					
			new_cost = sp(tmp_alignment2, new_alignlen2, K); 
*/
		} else { // num_cuts == 3

			// need to try three possible alignments, pick the best, 
			// then reassign the maps so the new order is correct
			bestThreeWayMerge(K, space_optimized, splits, maps, split_sizes, 
					best_alignlen, &new_alignlen1, &new_alignlen2, 
					&tmp_alignment2,  NULL, NULL,
					&new_cost, &which_merge);
			if (verbose) printf ("3waymerge cost: %d\n",new_cost); 
			int *tmp_map;
			if (which_merge == 2 || which_merge == -2) {
				tmp_map = maps[1];
				maps[1] = maps[2];
				maps[2] = tmp_map;				
			} else if (which_merge == 3 || which_merge == -3) {
				tmp_map = maps[0];
				maps[0] = maps[1];
				maps[1] = maps[2];				
				maps[2] = tmp_map;				
			}			
		}		

		for (k=0; k<num_cuts ; k++) {
			for(j=0; j<split_sizes[k]; j++){
				free(splits[k][j]);
			}
		}



		// reorder based on a map
		j=0;
		new_alignment = aa_malloc(K*sizeof(char*));
		while (maps[0][j] != -1) {
			new_alignment[maps[0][j]] = tmp_alignment2[j];
			j++;
		}
		k=0;
		while (maps[1][k] != -1) {
			new_alignment[maps[1][k]] = tmp_alignment2[j+k];
			k++;
		}
		if (num_cuts == 3) {
			m=0;
			while (maps[2][m] != -1) {
				new_alignment[maps[2][m]] = tmp_alignment2[j+k+m];
				m++;
			}	
		}



		if ( do_local_search ) {
			free(tmp_alignment2);
			// then iteratively pick "best" cut of this new alignment, until improvement stops
			new_alignlen2 = localSearch (num_cuts, K, space_optimized, new_alignment, new_alignlen2, &tmp_alignment2, &new_cost);
			if (verbose) printf ("localsearch2 cost: %d\n",new_cost);
		} else {
			for (j=0; j<K; j++) {
				tmp_alignment2[j] = new_alignment[j];
			}			
		}
		
										
		if (new_cost < best_cost) {
			for(j=0;j<K;j++) {
				free((*polished_alignment)[j]);	
				(*polished_alignment)[j] = tmp_alignment2[j];
			}			
			best_alignlen = new_alignlen2;
			best_cost = new_cost;
		} else {
			for(j=0;j<K;j++) {
				free(tmp_alignment2[j]);	
			}
		}
		free(tmp_alignment2);

	}


	for (i=0; i<num_cuts ; i++) {
		free(splits[i]);
		free (maps[i]);
	}
	free (splits);	
	free (split_sizes);
	free (maps);	
	

	if ( do_local_search ) {
		if ( TREE_CUT == local_search_method || THREE_CUT_TREE == local_search_method) {
			for (i=0; i<K; i++)
				free(induced_ij[i]);
			free(induced_ij);
		}
	}				
	
	return best_alignlen;
}

void randomlySplit(int num_groups, int K, char ***alignment, int alignlen, char ****splits, 
						int **split_sizes, int ***maps) 
{
	if ( TREE_CUT == random_cut_method) {
		randomlySplit_tree(K, alignment, alignlen, splits, split_sizes, maps);		
	} else {
		randomlySplit_pure(num_groups, K, alignment, alignlen, splits, split_sizes, maps);	
	}
}


void randomlySplit_tree(int K, char ***alignment, int alignlen, char ****splits, int **split_sizes, int ***maps) {

	int *leaves_under_i = aa_malloc(K*sizeof(int));
	int *leaves_under_j = aa_malloc(K*sizeof(int));
	
	int i;

	assign_order_vals(tree_root); // for the "subtree" and "sibling" tests below 


	int subtree_i;
	int subtree_j;
	int trials_with_node_i = 10;
	int trials_so_far = 0;
	while (1) {		
		if (trials_so_far == trials_with_node_i) 
			trials_so_far = 0;		

		if (trials_so_far == 0) {
			//pick first subtree - can't be root
			while (1) {		
		        subtree_i = (int)( (float)rand()*tree_node_count/RAND_MAX);
//		        if (verbose) printf ("trying subtree i= %d\n", subtree_i);
				if (subtree_i != tree_root)
					break;
			}		
		}
	
		trials_so_far++;
		
		//pick second subtree - several things it can't be ...
        subtree_j = (int) ((float)rand()*tree_node_count/RAND_MAX);
//        if (verbose) printf ("trying subtree j= %d\n", subtree_j);
		//j is root
		if (subtree_j == tree_root)
			continue;
		//i is ancestor of j
		if (tree_nodes[subtree_j].preorder > tree_nodes[subtree_i].preorder && tree_nodes[subtree_j].postorder < tree_nodes[subtree_i].postorder)
			continue;
		//j is ancestor of i
		if (tree_nodes[subtree_i].preorder > tree_nodes[subtree_j].preorder && tree_nodes[subtree_i].postorder < tree_nodes[subtree_j].postorder)
			continue;			
		//j is i's sibling
		if (tree_nodes[subtree_j].parent == tree_nodes[subtree_i].parent ) 
	    	continue;	
		   
		// passed that gauntlet? Then it's an ok node:
		break;
	}		    	    
    
	//that's the random cut.  Set up the arrays:
	int not_under_i_or_j[K];
	for (i=0; i<K; i++)
		not_under_i_or_j[i]=AA_TRUE;
	
	int leaf_count_i = 0;
	int leaf_count_j = 0;						
	get_leaves ( subtree_i , &leaves_under_i , &leaf_count_i, -1);	
	get_leaves ( subtree_j , &leaves_under_j , &leaf_count_j, -1);

	(*split_sizes)[0] = leaf_count_i;
	(*split_sizes)[1] = leaf_count_j;	
	(*split_sizes)[2] = K - leaf_count_i - leaf_count_j;	
	int j;
    for (i = 0; i < 3; i++) {
	  //		(*splits)[i]= aa_malloc((*split_sizes)[i] * sizeof(char*));
	    for (j = 0; j < K; j++) {
	    	(*maps)[i][j] = -1;
	    }
    }
	
	int m=0;
	for (i=0; i<leaf_count_i; i++) {
		not_under_i_or_j[leaves_under_i[i]] = AA_FALSE;
       	(*splits)[0][m] = aa_malloc(alignlen*sizeof(char));
		memcpy( (*splits)[0][m],(*alignment)[leaves_under_i[i]], alignlen);
		(*maps)[0][m] = leaves_under_i[i];
		m++;
	}

	
	m=0;
	for (i=0; i<leaf_count_j; i++) {					
		not_under_i_or_j[leaves_under_j[i]] = AA_FALSE;
       	(*splits)[1][m] = aa_malloc(alignlen*sizeof(char));
		memcpy( (*splits)[1][m],(*alignment)[leaves_under_j[i]], alignlen);
		(*maps)[1][m] = leaves_under_j[i];		
		m++;
	}

	m=0;
	for (i=0; i<K; i++) {					
		if (not_under_i_or_j[i]) {
	       	(*splits)[2][m] = aa_malloc(alignlen*sizeof(char));
			memcpy( (*splits)[2][m],(*alignment)[i], alignlen);
			(*maps)[2][m] = i;		
			m++;
		}
	}
	
	free(leaves_under_i);
	free(leaves_under_j);

}


void randomlySplit_pure(int num_groups, int K, char ***alignment, int alignlen, char ****splits, 
						int **split_sizes, int ***maps) {

	int keep_going = AA_TRUE;
	int i, j;
	float x;
		
	int subset_membership[K];
	

	while(keep_going) {
		for (i=0; i<num_groups; i++)
			(*split_sizes)[i] = 0;
			
	    for (i = 0; i < K; i++) {
	        x = (float) rand()/RAND_MAX;
	        //j = num_groups*x; 	   // even distribution
	        
	        //<trying to make smaller groups>
	        if (x < rand_cut_group1_pct) 
	        	j=0;
	        else {
	        	if (num_groups == 2)
	        		j = 1;
        		else 
     				if ( x < rand_cut_group2_pct)
     					j = 1;
     				else
     					j = 2;   		
	        }
	        //</trying to make smaller groups>	        

 	    	subset_membership[i] = j;
	   		(*split_sizes)[j]++;
	    } 

	    keep_going = AA_FALSE;	    
	    int total=0;
        for (i=0; i<num_groups; i++) {
        	total += (*split_sizes)[i];
		    if ((*split_sizes)[i] == 0)
			    keep_going = AA_TRUE;
        }
	}    
	
	if (verbose) printf("\nrandom split: A=%d, B=%d, C=%d\n", (*split_sizes)[0], (*split_sizes)[1], (*split_sizes)[2]);
	
    for (i=0; i<num_groups; i++) {	
		(*splits)[i]= aa_malloc((*split_sizes)[i] * sizeof(char*));
	    for (j = 0; j < K; j++) {
	    	(*maps)[i][j] = -1;
	    }
    }
    
    
    for (j=0; j<num_groups; j++) {
	    int m=0;
	    for (i = 0; i < K; i++) {
	        if ( subset_membership[i] == j ) {
	        	(*splits)[j][m] = aa_malloc(alignlen*sizeof(char));
				memcpy( (*splits)[j][m],(*alignment)[i], alignlen);
				(*maps)[j][m] = i;
				m++;
	        }
        }
    }    
}


int realign_across_edge_cut(int K, char **best_alignment, int best_len, int edge,
									char ***result_alignment, int *new_cost)
{
	int *leaves_under_i = aa_malloc(K*sizeof(int));
	int *leaves_above_i = aa_malloc(K*sizeof(int));	
	
	int i;
	int not_under_i[K];
    for (i=0; i<K; i++)
		not_under_i[i]=AA_TRUE;
	
	int leaf_count_i = 0;
	get_leaves ( edge , &leaves_under_i , &leaf_count_i, -1);	

	//split the seqs apart
	char **A = aa_malloc(leaf_count_i * sizeof (char*));
	char **B = aa_malloc((K-leaf_count_i) * sizeof (char*));	

	for (i=0; i<leaf_count_i; i++) {
		not_under_i[leaves_under_i[i]] = AA_FALSE;
       	A[i] = aa_malloc(best_len*sizeof(char));
		memcpy( A[i], best_alignment[leaves_under_i[i]], best_len);

	}
	int m=0;
	for (i=0; i<K; i++) {					
		if (not_under_i[i]) {
	       	B[m] = aa_malloc(best_len*sizeof(char));
			memcpy( B[m], best_alignment[i], best_len);
			leaves_above_i[m] = i;
			m++;
		}
	}
	
	//realign the seqs
	char **merged_alignment;
	unsigned int tmp_len;
	*new_cost = calcBestAlignCost(space_optimized, use_weights_in_merge, AA_FALSE, 
								&A, leaf_count_i, best_len, leaves_under_i,
								&B, K - leaf_count_i, best_len, leaves_above_i,
								&merged_alignment, &tmp_len, NULL, ALL_PAIRS);


	//map back to correct order for return seq
	*result_alignment = aa_malloc(K * sizeof(char*));
	for (i=0; i<leaf_count_i; i++) 
		(*result_alignment)[leaves_under_i[i]] = merged_alignment[i];
	for (i=0; i<K-leaf_count_i; i++) 
		(*result_alignment)[leaves_above_i[i]] = merged_alignment[leaf_count_i+i];
	free(merged_alignment);


	free(leaves_under_i);
	free(leaves_above_i);
	
	for (i=0; i<leaf_count_i; i++)
		free(A[i]);
	for (i=0; i<K-leaf_count_i; i++)
		free(B[i]);	
	free(A);		
	free(B);
	
	return tmp_len;
}
																


unsigned int localSearch (int num_cuts, int K, int space_optimized, char** alignment, 
								int alignlen, char ***polished_alignment, int *cost) 
{
	*polished_alignment = NULL;
	int i, j;

	if ( ANY_CUT == local_search_method) {
		
		return localSearch_purecut (num_cuts, K, space_optimized, alignment, alignlen, polished_alignment, cost);				
	
	}else {//if (TREE_CUT == local_search_method || THREE_CUT_TREE == local_search_method) {	
		
		int num_times_no_change = 0;

		aa_setWeights(K,weights);
		int best_cost = sp(alignment, alignlen, K);
		unsigned int best_len = alignlen;
		int new_cost;
		unsigned int new_len;

		char **best_alignment = aa_malloc(K*sizeof(char*));
		for (i=0; i<K; i++){
			best_alignment[i] = aa_malloc(alignlen*sizeof(char));
			memcpy(best_alignment[i], alignment[i], alignlen);
		}		

		if (verbose) printf ("going into new local search loop. Current cost is: %d\n", best_cost);


		if ( ALL == two_cut_type) {
			int num_edges = 2*K-2;
			int *edges = aa_malloc(num_edges*sizeof(int));
			int keep_going = AA_TRUE;
			
			while (keep_going) {
				get_single_edge_treecut_list ( K, &edges, num_edges );
				printf ("order of edges is:  ");
				for (i=0; i<num_edges; i++){
					printf ("%d, ", edges[i]);		
				}
				printf ("\n");
				int accepted_changes = 0;
				for (i=0; i<num_edges; i++){
					char ** result_alignment;
//					printf ("trying to align across edge %d  ... ", edges[i]);
					
					new_len = realign_across_edge_cut(K, best_alignment, best_len, edges[i],
																&result_alignment, &new_cost);
			
//					printf ( "with new cost %d\n", new_cost);
					if (new_cost < best_cost) {
						printf ("\tedge %d improved cost from %d to %d\n", i, best_cost, new_cost);
						keep_going = AA_TRUE;
						accepted_changes++;
						best_cost = new_cost;
						best_len = new_len;
						for (j=0; j<K; j++){
							best_alignment[j] = aa_realloc(best_alignment[j], best_len*sizeof(char));
							memcpy(best_alignment[j], result_alignment[j], best_len);
						}		
						int ii, jj;			
						aa_setUnweighted();
						for (ii=0; ii<K-1; ii++) {
							for (jj=ii+1; jj<K; jj++) {	
								int x = alignmentCost(best_alignment+ii, best_alignment+jj, 1, 1, best_len, best_len);
								induced_ij[ii][jj] = induced_ij[jj][ii] = x;
							} 
						} 																
					}
					for (j=0;j<K;j++)
						free(result_alignment[j]);	
					free(result_alignment);															

					if (accepted_changes == two_cut_reorder_threshold) {
						printf ("** 5 accepted changes. start over\n");
						break; // start loop over	
					}
				}
				if (accepted_changes == 0)
					keep_going = AA_FALSE;
			}
			free(edges);
			
		} else { //either not a two-cut, or two_cut_type == BEST

			while (num_times_no_change<2) {
	
				char **result_alignment;
				
				if (2 == num_cuts) { // must be two_cut_type == BEST
					int *edge = aa_malloc (sizeof(int));
					get_single_edge_treecut_list ( K, &edge, 1 );

					new_len = realign_across_edge_cut(K, best_alignment, best_len, edge[0],
															&result_alignment, &new_cost);
					
				} else { //3 == num_cuts
					pick_twoedge_treecut(K, &i, &j);
	
					if (TREE_CUT == local_search_method ) 
						new_len = treeCutMerge_complex(K, space_optimized, best_alignment, best_len, 
													i, j, &result_alignment, &new_cost) ;				
					else //THREE_CUT_TREE == local_search_method
						new_len = treeCutMerge_simple(K, space_optimized, best_alignment, best_len, 
													i, j, &result_alignment, &new_cost) ;
				}
				 
				if (verbose) printf ("end of local search loop. Current cost is: %d\n", new_cost);
				
				if ( new_cost > best_cost ) {
					for (i=0;i<K;i++)
						free(result_alignment[i]);	
					free(result_alignment);
					break;
				} else {
	
					if (new_cost < best_cost) {
						num_times_no_change = 0;
					} else {
						num_times_no_change++;
					}
	
					best_cost = new_cost;
					best_len = new_len;
					for (i=0; i<K; i++){
						best_alignment[i] = aa_realloc(best_alignment[i], best_len*sizeof(char));
						memcpy(best_alignment[i], result_alignment[i], best_len);
					}		
					
					for (i=0;i<K;i++)
						free(result_alignment[i]);	
					free(result_alignment);				
	
				}
			}
		}

		*polished_alignment = best_alignment;	
		*cost = best_cost;
		return best_len;

	}
}

void get_single_edge_treecut_list (int K, int **edges, int num_edges ) 
{
	int i, j, k, m, n;

	for (i=0; i<num_edges; i++) 
		(*edges)[i] = -1;
		
	BestEdgeEntry *head;
	BestEdgeEntry *tail;
	head = tail = NULL;
	
	assign_order_vals(tree_root); // for the "subtree" and "sibling" tests below 
		

	int best_diff = -1;
	int *leaves_under_i = aa_malloc(K * sizeof(int));
	int not_under_i[K];
	int leaves_above_i[K];	
	
	int filled_slots = 0;
	for(i=0; i<tree_node_count; i++) {
		if (i == tree_root) {
			continue;
		}
    	// calc best benefit of cut, add it to the list (sorted)
		// first find list of leaves under i or not-under i
		for (k=0; k<K; k++)
			not_under_i[k]=AA_TRUE;
		
    	int leaf_count_i = 0;
		get_leaves ( i , &leaves_under_i , &leaf_count_i, -1);	
		for (k=0; k<leaf_count_i; k++) {
			not_under_i[leaves_under_i[k]] = AA_FALSE;
		}

		j=0;
		for (k=0; k<K; k++) {
			if ( not_under_i[k]) {
				leaves_above_i[j] = k;
				j++;
			}
		}

	
		int diff = 0.0;
		// then consider weighted avg (induced pairwise cost vs optimal p/w cost), keep the biggest difference
		int weight_sum = 0.0;
		for (m=0; m<leaf_count_i; m++) {
			for (n=0; n<K-leaf_count_i; n++) {					
				int x = leaves_under_i[m];
				int y = leaves_above_i[n];
				int weight = (NULL==weights) ? 1 : weights[x][y];
				weight_sum += weight;
				diff += weight * (induced_ij[x][y] - tree_cut_optimal_distances[x][y]); 
			}
		}
		diff = (int)(0.5+(diff/weight_sum));

//		printf ("considering edge %d. Has diff = %d\n", i, diff);

		if ( 1 == num_edges) { // special case, to avoid overhead
			if (diff > best_diff) {
				best_diff = diff;
				(*edges)[0] = i;
			}
		} else {
		
			if (i==0) {
				head = tail = aa_malloc( sizeof(BestEdgeEntry) );	
				head->edge_id = i;
				head->diff = diff;
				head->next = head->prev = NULL;
				filled_slots = 1;
			} else {
				BestEdgeEntry *curr = head;
				
				while (curr != NULL || i<num_edges ) {
					if ( curr == NULL || diff > curr->diff ) {				
						//insert here
						BestEdgeEntry *entry = aa_malloc( sizeof(BestEdgeEntry) );	
						entry->edge_id = i;
						entry->diff = diff;
						if (curr == NULL) { // ran off the end of the list, but there are still slots to fill
//							printf ("\tinsert at end of list\n");
							tail->next = entry;
							entry->prev = tail;
							entry->next = NULL;
							tail = entry;
						} else { // diff > curr->diff,  in middle of list, insert just after curr entry
//							printf ("\tinsert before entry with diff = %d\n", curr->diff);
							if (curr == head) {
								head = entry;
							} else {
								curr->prev->next = entry;
								entry->prev = curr->prev;
							}							
							entry->next = curr;
							curr->prev = entry;
							if (filled_slots > num_edges) { // drop the tail
//									printf ("\tremove entry with diff = %d\n", tail->diff);
									curr = tail->prev;
									curr->next = NULL;
									free (tail);
									tail = curr; 
							} else {
								filled_slots++;
							}
						}
						break;
					}
					curr = curr-> next;
				}
			}
		}
	}	

	i=0;	
//	printf ("done with edges\n"); 
	if ( num_edges >  1 ) { // run through linked list to return array
		BestEdgeEntry *curr = head;
		BestEdgeEntry *prev ;
		while (curr != NULL ) {
			(*edges)[i] = curr->edge_id;
			prev = curr;
			curr = curr->next;
			free(prev);
			i++;
		}
	}
	free(leaves_under_i);
}



void pick_twoedge_treecut (int K, int *edge_i, int *edge_j ) 
{
	
	int i, j, k, m, n;
	
	assign_order_vals(tree_root); // for the "subtree" and "sibling" tests below 
		
	//pick best pair of edges: 1 to remove, the other specifies where to put the first
	float best_diff = -1;
	int *leaves_under_i = aa_malloc(K*sizeof(int));
	int *leaves_under_j = aa_malloc(K*sizeof(int));
	int not_under_i_or_j[K];	
	
	for(i=0; i<tree_node_count; i++) {
		for(j=i+1; j<tree_node_count; j++) {		

			//i or j are root
			if (tree_nodes[j].parent == -1 || tree_nodes[i].parent == -1) 
				continue;
			//i is ancestor of j
 			if (tree_nodes[j].preorder > tree_nodes[i].preorder && tree_nodes[j].postorder < tree_nodes[i].postorder)
 				continue;
			//j is ancestor of i
 			if (tree_nodes[i].preorder > tree_nodes[j].preorder && tree_nodes[i].postorder < tree_nodes[j].postorder)
 				continue;			
			//j is i's sibling
			if (tree_nodes[j].parent == tree_nodes[i].parent ) 
		    	continue;	
		    	    
	    	// calc best benefit of pair, retain if it's the best so far.	
			// first find list of leaves under i, j, or neither
			for (k=0; k<K; k++)
				not_under_i_or_j[k]=AA_TRUE;
			
	    	int leaf_count_i = 0;
			get_leaves ( i , &leaves_under_i , &leaf_count_i, -1);	
			for (k=0; k<leaf_count_i; k++) {
				not_under_i_or_j[leaves_under_i[k]] = AA_FALSE;
			}

	    	int leaf_count_j = 0;					
			get_leaves ( j , &leaves_under_j , &leaf_count_j, i);
			for (k=0; k<leaf_count_j; k++) {					
				not_under_i_or_j[leaves_under_j[k]] = AA_FALSE;
			}

		
			float diff1 = 0.0, diff2 = 0.0, diff3 = 0.0;
			// then consider weighted avg (induced pairwise cost vs optimal p/w cost), keep the biggest difference
			int weight_sum = 0, weight_sum2 = 0, weight_sum3 = 0;
			for (m=0; m<leaf_count_i; m++) {
				for (n=0; n<leaf_count_j; n++) {					
					int x = leaves_under_i[m];
					int y = leaves_under_j[n];
					int weight = (NULL == weights) ? 1 : weights[x][y];
					weight_sum += weight;
					diff1 += weight * (induced_ij[x][y] - tree_cut_optimal_distances[x][y]); 
				}
			}
			diff1 /= weight_sum;
			

//			if (num_cuts == 3){
			if (CUT_MULTI == three_tree_cut_type) {
				for (k=0; k<K; k++) {
					if (not_under_i_or_j[k]) {
						for (m=0; m<leaf_count_i; m++) {			
							int x = leaves_under_i[m];
							int weight = (NULL == weights) ? 1 : weights[x][k];
							weight_sum2 += weight;
							diff2 += weight * (induced_ij[x][k] - tree_cut_optimal_distances[x][k]); 
						}
	
						for (m=0; m<leaf_count_j; m++) {
							int y = leaves_under_j[m];
							int weight = (NULL == weights) ? 1 : weights[y][k];			
							weight_sum3 += weight;							
							diff3 +=  weight * (induced_ij[y][k] - tree_cut_optimal_distances[y][k]); 
						}
					}
				}
				diff2 /= weight_sum2;
				diff3 /= weight_sum3;

			}			
			float diff = diff1 + diff2 + diff3;
			
			if (diff > best_diff) {
				best_diff = diff;
				*edge_i = i;
				*edge_j = j;	
			}
		}
	}
	free(leaves_under_i);
	free(leaves_under_j);	
}

unsigned int treeCutMerge_simple(int K, int space_optimized, char **alignment, int alignlen, 
					int i, int j, char ***merged_alignment, int *cost) {

	int *leaves_under_i = aa_malloc(K * sizeof(int));
	int *leaves_under_j = aa_malloc(K * sizeof(int)); 

	int leaf_count_i = 0;
	get_leaves ( i , &leaves_under_i , &leaf_count_i, -1);	
	int leaf_count_j = 0;					
	get_leaves ( j , &leaves_under_j , &leaf_count_j, -1);



	// do the actual sequence merge
	char ***splits = aa_malloc(3 * sizeof(char**));
	splits[0] = aa_malloc( leaf_count_i * sizeof(char*));
	splits[1] = aa_malloc( leaf_count_j * sizeof(char*));
	splits[2] = aa_malloc( (K-leaf_count_i-leaf_count_j) * sizeof(char*));
	
	int *split_sizes = aa_malloc( 3 * sizeof(int));
	split_sizes[0] = leaf_count_i;
	split_sizes[1] = leaf_count_j;	
	split_sizes[2] = K-leaf_count_i-leaf_count_j;
	
	int **maps = aa_malloc(3 * sizeof(int*));
	maps[0] = aa_malloc(leaf_count_i * sizeof(int));
	maps[1] = aa_malloc(leaf_count_j * sizeof(int));
	maps[2] = aa_malloc( (K-leaf_count_i-leaf_count_j) * sizeof(int));		

	int k;
	int not_under_i_or_j[K];
	for (k=0; k<K; k++)
		not_under_i_or_j[k]=AA_TRUE;

	for (k=0; k<leaf_count_i; k++) {
		maps[0][k] = leaves_under_i[k];
		splits[0][k] = aa_malloc(alignlen * sizeof(char));
		memcpy(splits[0][k],alignment[leaves_under_i[k]],alignlen);		
		not_under_i_or_j[leaves_under_i[k]]=AA_FALSE;
	} 
	for (k=0; k<leaf_count_j; k++){
		maps[1][k] = leaves_under_j[k];
		splits[1][k] = aa_malloc(alignlen * sizeof(char));
		memcpy(splits[1][k],alignment[leaves_under_j[k]],alignlen);		
		not_under_i_or_j[leaves_under_j[k]]=AA_FALSE;
	}
	// leaves not under i or j
	int x=0;
	for (k=0; k<K; k++){
		if (not_under_i_or_j[k]) { 
			//need to alloc alignments[x]?
			maps[2][x] = k;
			splits[2][x] = aa_malloc(alignlen * sizeof(char));
			memcpy(splits[2][x],alignment[k],alignlen);		
			x++;			
		}
	}

	unsigned int new_alignlen1, new_alignlen2;
	int which_merge;
	char **final_alignment;	
	bestThreeWayMerge(K, space_optimized, splits, maps, split_sizes, 
					alignlen, &new_alignlen1, &new_alignlen2, 
					&final_alignment, NULL, NULL,
					cost, &which_merge);
	
	int spots_before_i=0;
	int spots_before_j=leaf_count_i;
	int spots_before_k=leaf_count_i+leaf_count_j;
	
	if (which_merge == 2 || which_merge == -2) {
		//(0,2),1
		spots_before_j = K-leaf_count_j;
		spots_before_k = leaf_count_i;
	} else if (which_merge == 3 || which_merge == -3) {
		// (1,2),0						
		spots_before_j=0;
		spots_before_k=leaf_count_j;
		spots_before_i=K-leaf_count_i;
	}			 

	*merged_alignment = aa_malloc(K*sizeof(char**));
	for (k=0; k<leaf_count_i; k++) {
		(*merged_alignment)[maps[0][k]] = final_alignment[spots_before_i+k];
	} 
	for (k=0; k<leaf_count_j; k++){
		(*merged_alignment)[maps[1][k]] = final_alignment[spots_before_j+k];
	}
	// leaves not under i or j
	for (k=0; k<K-leaf_count_i-leaf_count_j; k++){
		(*merged_alignment)[maps[2][k]] = final_alignment[spots_before_k+k];
	} 

	free(final_alignment);


	// Alter the tree accordingly

	// * merge ij as one child of new root, with other child = the remaining tree

	//pull out subtree under i
	int i_parent = tree_nodes[i].parent;
	int i_sibling = (tree_nodes[i_parent].left_child == i) ? tree_nodes[i_parent].right_child : tree_nodes[i_parent].left_child ;
	if (i_parent == tree_root) {
		//just re-root the tree at the other sibling
		tree_root = i_sibling;
		tree_nodes[i_sibling].parent = -1;
	} else {
		int i_grandparent = tree_nodes[i_parent].parent;
		//replace i_parent with i_sibling
		tree_nodes[i_sibling].parent = i_grandparent;
		if (tree_nodes[i_grandparent].left_child == i_parent) {
			tree_nodes[i_grandparent].left_child = i_sibling;
		} else {
			tree_nodes[i_grandparent].right_child = i_sibling;	
		} 
	}


	//pull out subtree under j
	int j_parent = tree_nodes[j].parent;
	int j_sibling = (tree_nodes[j_parent].left_child == j) ? tree_nodes[j_parent].right_child : tree_nodes[j_parent].left_child ;
	if (j_parent == tree_root) {// parent is the root
		//just re-root the tree at the other sibling
		tree_root = j_sibling;
		tree_nodes[j_sibling].parent = -1;
	} else {
		int j_grandparent = tree_nodes[j_parent].parent;
		//replace j_parent with j_sibling
		tree_nodes[j_sibling].parent = j_grandparent;
		if (tree_nodes[j_grandparent].left_child == j_parent) {
			tree_nodes[j_grandparent].left_child = j_sibling;
		} else {
			tree_nodes[j_grandparent].right_child = j_sibling;	
		} 
	}

	if (which_merge == 1 || which_merge == -1) {
		//merge i and j, re-using i_parent
		tree_nodes[j].parent = i_parent;
		if (tree_nodes[i_parent].left_child == i) {
			tree_nodes[i_parent].right_child = j;
		} else {
			tree_nodes[i_parent].left_child = j;	
		}
		
		//place i and j as outgroup on tree, reusing j_parent as the new root to the tree
		tree_nodes[j_parent].right_child = i_parent; // right child is the new cluster formed by i and j
		tree_nodes[i_parent].parent = j_parent;

		tree_nodes[j_parent].left_child = tree_root; // left child = old root
		tree_nodes[tree_root].parent = j_parent;
	 	tree_root = j_parent;
	 				 			
	} else if (which_merge == 2 || which_merge == -2) {
		//merge i and root, re-using i_parent
		tree_nodes[tree_root].parent = i_parent;
		if (tree_nodes[i_parent].left_child == i) {
			tree_nodes[i_parent].right_child = tree_root;
		} else {
			tree_nodes[i_parent].left_child = tree_root;	
		}

		//place j as outgroup on tree, reusing j_parent as the new root to the tree
		tree_nodes[i_parent].parent = j_parent;
		if (tree_nodes[j_parent].left_child == j) {
			tree_nodes[j_parent].right_child = i_parent; 
		} else {
			tree_nodes[j_parent].left_child = i_parent; 
		}

	 	tree_root = j_parent;

	} else { // if (which_merge == 3 || which_merge == -3) {
		//merge j and root, re-using j_parent
		tree_nodes[tree_root].parent = j_parent;
		if (tree_nodes[j_parent].left_child == j) {
			tree_nodes[j_parent].right_child = tree_root;
		} else {
			tree_nodes[j_parent].left_child = tree_root;	
		}

		//place i as outgroup on tree, reusing i_parent as the new root to the tree
		tree_nodes[j_parent].parent = i_parent;
		if (tree_nodes[i_parent].left_child == i) {
			tree_nodes[i_parent].right_child = j_parent; 
		} else {
			tree_nodes[i_parent].left_child = j_parent; 
		}

	 	tree_root = i_parent;
	}
	
	tree_nodes[tree_root].parent = -1; // new root	

	free(leaves_under_i);
	free(leaves_under_j);

	return new_alignlen2;	
}




unsigned int treeCutMerge_complex (int K, int space_optimized, char **alignment, int alignlen, 
					int i, int j, char ***merged_alignment, int *cost) {


	//prepare the structures used to hold cluster information
	int k,m;
/*	
 * this breaks, even though all "alignments[0][k]" entries are valid. Why?
 * for (k=0; k<K; k++) 
		free (alignments[0][k]);
	free (alignments[0]);
*/

    ClusterEntry *ce = NULL;
	for (k=0; k<K; k++) {
		alignment_roots[k] = -1;	
	}	
	 

	int *A_ids = aa_malloc( K * sizeof(int));	
	int *AB_ids = aa_malloc( K * sizeof(int));	
	int *B_ids = aa_malloc( K * sizeof(int));	
	int *C_ids = aa_malloc( K * sizeof(int));

	
	/* ****************************************
	 * delete all ancestor nodes of i and j, placing remaining 
	 * subtrees onto cluster array
	 * *******************************************/
	 
	int mrca=-1; // most recent common ancestor
	int next = tree_nodes[i].parent;
	while (next != -1) {
		if (tree_nodes[j].preorder > tree_nodes[next].preorder &&
	 		tree_nodes[j].postorder < tree_nodes[next].postorder ){ //j is in next 's subtree
				mrca = next;
				break;
 		}
		next = tree_nodes[next].parent;
 	}
 	if (next == -1) {
 		fprintf (stderr, "couldn't find mrca. That's bad\n");
 		exit (0);
 	}


	int free_node = -1; // no free nodes yet
	cluster_count = 0;
	int chosen_subtrees[2]= {i,j};

	for (k=0; k<2; k++) {
		int to_remove = tree_nodes[chosen_subtrees[k]].parent;
		int sibling;
		if (tree_nodes[to_remove].left_child == chosen_subtrees[k]) {
		 	sibling = tree_nodes[to_remove].right_child;
		} else {
		 	sibling = tree_nodes[to_remove].left_child;			
		}
		while (  (k==0 && to_remove!=mrca) || (k==1 && to_remove != -1) ) {
			//fill data structures with this new cluster
			//sibling is a subtree that should be kept as a cluster
			if (! (k==1 && to_remove==mrca) ) { // if the internal test is true (i.e. "if" is false), sibling would be a node that's already deleted during pass 0, so don't add it
				tree_nodes[sibling].parent = -1;

				int leaf_count=0;
				int *leaves_under_sibling = aa_malloc(K*sizeof(int));			
				get_leaves ( sibling , &leaves_under_sibling , &leaf_count, -1);				

				int first = leaves_under_sibling[0];						    
				alignment_roots[first] = sibling;

				alignments[first] = aa_malloc(leaf_count * sizeof(char*));
				for (m=0; m<leaf_count; m++) {
					alignments[first][m] = aa_malloc(alignlen * sizeof(char));
					memcpy(alignments[first][m], alignment[leaves_under_sibling[m]], alignlen);
	
				  	ce = aa_malloc(sizeof(ClusterEntry));
					ce->seqid = leaves_under_sibling[m];
					ce->next = NULL;  
	
					if ( m == 0 ) {
						clusters[cluster_count].first = ce ;
					} else {
						clusters[cluster_count].last->next = ce ;					
					}
					clusters[cluster_count].last = ce;
	
				}
				alignment2cluster_map[first] = cluster_count;
				cluster_sizes[cluster_count] = leaf_count;
				lengths[first] = RemoveNullColumns (alignments[first], leaf_count, alignlen);
				
				set_weight_matrix_for_cluster ( cluster_count );
//				alignment_costs[first] = sp(alignments[first], lengths[first], leaf_count );
				
				cluster_count++;
			}	

 
			// prepare to move up the tree
			int parent_of_to_remove = tree_nodes[to_remove].parent ;
			if (tree_nodes[parent_of_to_remove].left_child == to_remove) {
				sibling = tree_nodes[parent_of_to_remove].right_child;
			} else {
				sibling = tree_nodes[parent_of_to_remove].left_child;				
			}

			//remove ancestor node from tree, put on "free nodes" list
			tree_nodes[to_remove].left_child = tree_nodes[to_remove].right_child =
					tree_nodes[to_remove].parent = tree_nodes[to_remove].preorder = tree_nodes[to_remove].postorder = -1;

			if (free_node != -1){
				tree_nodes[to_remove].preorder = free_node; // re-using "preorder" as "pointer" to next free node
			}		
			free_node = to_remove;
			
			to_remove = parent_of_to_remove;			
		}
	}
	//reuse root as new parent of i and j
	tree_nodes[tree_root].left_child  = i;
	tree_nodes[tree_root].right_child = j;
	tree_nodes[i].parent = tree_root;
	tree_nodes[j].parent = tree_root;	
	
	if (free_node == tree_root) 
		free_node = tree_nodes[tree_root].preorder; //shift list over
	
	
	
	//merge i and j, then add them to data structures, too
	int *leaves_under_i = aa_malloc(K * sizeof(int));
	int *leaves_under_j = aa_malloc(K * sizeof(int)); 	
	int leaf_count_i = 0;
	get_leaves ( i , &leaves_under_i , &leaf_count_i, -1);	
	int leaf_count_j = 0;					
	get_leaves ( j , &leaves_under_j , &leaf_count_j, -1);
		

	char **tmp_alignment1 = aa_malloc( leaf_count_i * sizeof(char*));
	char **tmp_alignment2 = aa_malloc( leaf_count_j * sizeof(char*));
	
	for (k=0; k<leaf_count_i; k++) { 
		tmp_alignment1[k] = alignment[leaves_under_i[k]];
		A_ids[k] = leaves_under_i[k];
	}
	for (k=0; k<leaf_count_j; k++) {
		tmp_alignment2[k] = alignment[leaves_under_j[k]];
		B_ids[k] = leaves_under_j[k];	
	}
	int first = leaves_under_i[0]; 
	unsigned int len;
	
/*	alignment_costs[first] = calcBestAlignCost_preservingOrig(space_optimized, use_weights_in_merge, AA_FALSE, 
														tmp_alignment1, leaf_count_i, alignlen, A_ids, 
 														tmp_alignment2, leaf_count_j, alignlen, B_ids,
 														alignments+first, &len, NULL, ALL_PAIRS);	
*/
	calcBestAlignCost_preservingOrig(space_optimized, use_weights_in_merge, AA_FALSE, 
														tmp_alignment1, leaf_count_i, alignlen, A_ids, 
 														tmp_alignment2, leaf_count_j, alignlen, B_ids,
 														alignments+first, &len, NULL, ALL_PAIRS);													
	lengths[first] = len;
	alignment_roots[first] = tree_root;
	
	for (m=0; m<leaf_count_i+leaf_count_j; m++) {
	  	ce = aa_malloc(sizeof(ClusterEntry));
		ce->next = NULL;  
		if ( m < leaf_count_i)
			ce->seqid = leaves_under_i[m];
		else 
			ce->seqid = leaves_under_j[m-leaf_count_i];

		if ( m == 0 ) 
			clusters[cluster_count].first = ce ;
		 else 
			clusters[cluster_count].last->next = ce ;					

		clusters[cluster_count].last = ce;
	}
	alignment2cluster_map[first] = cluster_count;
	cluster_sizes[cluster_count] = leaf_count_i + leaf_count_j;			
	cluster_count++;


	/* ****************************************
	 * Calculate distances between the clusters
	 *  easy, except for NJ, TJ & UPGMA (UPGMA not implemented correctly)
	 * Not extensively tested, yet.
	 * *******************************************/

	distance_heap = CreateHeap(Compare);	
	distance_nodes = aa_malloc(K * sizeof(HeapNode**));
	for (i=0; i<K; i++){
		distance_nodes[i] = aa_malloc(K * sizeof(HeapNode*));
	}
		
	if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) {
		internal_external_total_distances 	= aa_malloc(K * sizeof(int*));
		internal_external_numpairs 			= aa_malloc(K * sizeof(int*));
		internal_external_distances 		= aa_malloc(K * sizeof(int**));
		for (i=0; i<K; i++) {
			internal_external_total_distances[i]= aa_malloc(K * sizeof(int));
			internal_external_numpairs[i]		= aa_malloc(K * sizeof(int));
			internal_external_distances[i] 		= aa_malloc(K * sizeof(int*));
			for (j=0; j<K; j++) {
				internal_external_distances[i][j] = aa_malloc( K * sizeof(int)); // the last element holds the sum
			}
		}
	}


	/* need to pre-calculate a new distance matrix containing distances for all the
	 * alignments (clusters) created by the above node removal */
	int **precomputed_distances;
	int *active_indices;
	if ( MST_UPGMA == guide_tree_method || UPGMA == guide_tree_method || TJ == guide_tree_method ||	NJ == guide_tree_method ) {
		precomputed_distances = aa_malloc(K*sizeof(int*));
		active_indices = aa_malloc(K*sizeof(int));
		for (i=0; i<K; i++) {	
			precomputed_distances[i] = aa_malloc(K*sizeof(int));
			memcpy(precomputed_distances[i],tree_cut_optimal_distances[i],K*sizeof(int) );
			active_indices[i] = AA_TRUE;
		}

		//calls a recursive function that, in the end, assigns the right distances to all
		//current alignments, as they should be according to the subtrees already merged
		for (i=0; i<cluster_count; i++) {
			int tmp_i = clusters[i].first->seqid;			
			alterDistanceTable(alignment_roots[tmp_i], K, &precomputed_distances, &active_indices);
		}
		free(active_indices);
	}


	IDPair *pair;
	if ( NJ == guide_tree_method ) {

		for (i=0; i<K; i++) {
			for (j=0; j<K; j++) {			
				nj_optimal_distances[i][j] = precomputed_distances[i][j];
			}
		}
	
		int tmp_i, tmp_j;		
		for (i=0; i<cluster_count; i++){
			tmp_i = clusters[i].first->seqid; //index to alignment  (i is index to clusters)
			nj_total_distances[tmp_i] = 0;
			for (j=0; j<cluster_count; j++){	
				if (j==i) continue;
				tmp_j = clusters[j].first->seqid;			
				nj_total_distances[tmp_i] += nj_optimal_distances[tmp_i][tmp_j];
			}
		}
		// need to pick the min over all i,j of  dist[i][j] - (tot[i]+tot[j])/(K-2)
		int min_cost = AA_INF;
		int min_i, min_j;
		for (i=0; i<cluster_count-1; i++){
			tmp_i = clusters[i].first->seqid;
			for (j=i+1; j<cluster_count; j++){
				tmp_j = clusters[j].first->seqid;
				int reduction = (nj_total_distances[tmp_i] + nj_total_distances[tmp_j])/(cluster_count-2);
				int M = nj_optimal_distances[tmp_i][tmp_j] - reduction;
				if (M < min_cost) {
					min_cost = M;
					min_i = tmp_i;
					min_j = tmp_j;
				}					
			}
		}	
		pair = aa_malloc(sizeof(IDPair));
		pair->i = min_i;
		pair->j = min_j;
		HeapInsert((Pointer)min_cost, (Pointer) pair, distance_heap);		

	} else {
				
		unsigned int tmp_len;
		int aligncost;
		int tmp_cost1, tmp_cost2; // only used for AVG_INTERNAL_EXTERNAL case
		int min_cost = AA_INF;
		int min_i, min_j, number_pairs;// only used for AVG_INTERNAL_EXTERNAL case


		for (i=0; i<cluster_count-1; i++){
			for (j=i+1; j<cluster_count; j++){
				int ii = clusters[i].first->seqid; //index to alignment  (i is index to clusters)
				int jj = clusters[j].first->seqid;			

				if (NULL != weights) {
					ce = clusters[i].first;
					int cntr1 = 0;
					while (ce != NULL) {
						AB_ids[cntr1] = ce->seqid;
						A_ids[cntr1++] = ce->seqid;
						ce = ce->next;
					}
					ce = clusters[j].first;
					int cntr2 = 0;
					while (ce != NULL) {
						AB_ids[cntr1+cntr2] = ce->seqid;
						B_ids[cntr2++] = ce->seqid;
						ce = ce->next;
					}
				}			
	
				if ( AVG_INTERNAL == guide_tree_method || 
					TOTAL_INTERNAL == guide_tree_method ||
					MIN_INTERNAL == guide_tree_method ||
					MST == guide_tree_method ||
					UPGMA == guide_tree_method ||
					MST_UPGMA == guide_tree_method ||					
					TJ == guide_tree_method )
				{
	
					if ( UPGMA == guide_tree_method ||
						 MST_UPGMA == guide_tree_method ||					
							TJ == guide_tree_method ) 
					{
						aligncost = precomputed_distances[ii][jj];
					} else	if ( MST == guide_tree_method ) {
						int x,y;
						aligncost = AA_INF;
						for (x=0; x<cluster_sizes[i]; x++) {
							for (y=0; y<cluster_sizes[j]; y++) {
								int c = tree_cut_optimal_distances[x][y];	
								if (c < aligncost) 
									aligncost = c;
							}
						}
					} else { // AVG_INTERNAL , TOTAL_INTERNAL or MIN_INTERNAL 
	
						int cost_type = (MIN_INTERNAL == guide_tree_method) ? MIN_COST : alignment_cost_type ;
						int get_avg = (AVG_INTERNAL == guide_tree_method) ? AA_TRUE : AA_FALSE ;

						aligncost = calcBestAlignCost(space_optimized, use_weights_in_merge, get_avg, alignments+ii, cluster_sizes[i], lengths[ii], A_ids,
																	alignments+jj, cluster_sizes[j], lengths[jj], B_ids, NULL,NULL,NULL,
																	cost_type);			
					}
					
					pair = aa_malloc(sizeof(IDPair));
					pair->i = ii;
					pair->j = jj;
					distance_nodes[jj][ii] = distance_nodes[ii][jj] = HeapInsert((Pointer)min_cost, (Pointer) pair, distance_heap);
	
				} else if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) {
					
					// optimally align i and j (calcBestAlignCost_preservingOrig) ... tmp_align			
					//use initial_alignment temporarily here since it's available.
					tmp_cost1 = calcBestAlignCost_preservingOrig(space_optimized, use_weights_in_merge, AA_TRUE, 
													alignments[ii], cluster_sizes[i], lengths[ii], A_ids,
													alignments[jj], cluster_sizes[j], lengths[jj], B_ids,
													merged_alignment, &tmp_len, NULL, ALIGN_COST);
						
					aligncost = 0;
					number_pairs = 0;
					for (k=0; k<cluster_count; k++) {
						int kk = clusters[k].first->seqid; //index to alignment  (i is index to clusters)
						if (kk != ii && kk != jj){
							//calc alignmerge cost of tmp_align and k  
							if (NULL != weights) {
								ce = clusters[k].first;
								int cntr = 0;
								while (ce != NULL) {
									C_ids[cntr++] = ce->seqid;
									ce = ce->next;
								}
							}		 
							tmp_cost2 =  calcBestAlignCost(space_optimized, use_weights_in_merge, AA_TRUE, 
																merged_alignment, cluster_sizes[i]+cluster_sizes[j], tmp_len, AB_ids,
																alignments+kk, cluster_sizes[k], lengths[kk], C_ids, 
																NULL, NULL, NULL, ALIGN_COST );																	/*
							tmp_cost2 = calcAlignMergeCost(	space_optimized, use_weights_in_merge, AA_TRUE, 
																merged_alignment, cluster_sizes[i]+cluster_sizes[j], tmp_len, tmp_cost1, AB_ids,
																alignments+kk, cluster_sizes[k], lengths[kk], C_ids, NULL, NULL, NULL, ALIGN_COST );
																*/
		
							internal_external_distances[ii][jj][kk] = 
									internal_external_distances[jj][ii][kk] = tmp_cost2;				
							aligncost += tmp_cost2 ;
							//number_pairs += cluster_sizes[k] * (cluster_sizes[i]+cluster_sizes[j]);
						}
					}

					aligncost = (int)(.5+ (aligncost/cluster_count));
					//add in the cost of aligning A & B
					aligncost += tmp_cost1 ;
//					number_pairs += cluster_sizes[i]*cluster_sizes[j];
					
					internal_external_total_distances[jj][ii] = 
							internal_external_total_distances[ii][jj] = aligncost;				
//					internal_external_numpairs[jj][ii]		= 
//							internal_external_numpairs[ii][jj]		= number_pairs;				
	
					//track lowest cost - add just the lowest cost entry to the heap, at the end of the outer loop
					if (aligncost/*/number_pairs*/ < min_cost) {
						min_cost = aligncost; //number_pairs;
						min_i = ii;
						min_j = jj;
					}				
				}
			}
		}
		if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) {
			//add just the lowest cost entry to the heap.
			// This is the only entry in the heap if the method id internal_external ... just using the infrastructure in place for other methods
			pair = aa_malloc(sizeof(IDPair));
			pair->i = min_i;
			pair->j = min_j;
			HeapInsert((Pointer)min_cost, (Pointer) pair, distance_heap);
		}

	}

	// begin clustering sequences
	int best_i = -1;	
	int best_j = -1;

		 
    while (cluster_count > 1) {  	
    	pair = (IDPair*)HeapDeleteMin(distance_heap);
		best_i = pair->i;
		best_j = pair->j;
   		free(pair);
   		 	
    	//these are the ids of the alignments. 
    	//What I really need are the cluster ids, so use map:
    	int my_i = alignment2cluster_map[best_i];
    	int my_j = alignment2cluster_map[best_j];

		if (my_j < my_i) {
			int tmp = my_i;
			my_i = my_j;
			my_j = tmp;	 
			best_i = best_j; // necessary after the final loop, fpor the cleanup step below
		}
		 	 
		mergeAlignments (K, my_i, my_j, space_optimized, free_node, AA_TRUE);
				
		tree_root = free_node; //when the loop exits, this will indeed be the root.
		free_node = tree_nodes[free_node].preorder;		
    }
 
	//	need to shuffle the alignment around to match the original input of strings
	//something like this:
	ce = clusters[0].first;
	i = 0;
	*merged_alignment = aa_malloc(K * sizeof(char *));
	alignlen = lengths[best_i];
						
	while (ce != NULL) {
		(*merged_alignment)[ce->seqid] = aa_malloc(alignlen * sizeof(char));
		memcpy((*merged_alignment)[ce->seqid], alignments[best_i][i], alignlen);
		free(alignments[best_i][i]);
		ce = ce->next;	
		i++;
	}

	aa_setWeights(K,weights);
	*cost = sp(*merged_alignment, alignlen, K);


	for (i=0; i<K; i++) 
		free(distance_nodes[i]);
	free(distance_nodes);
	
	if ( MST_UPGMA == guide_tree_method || UPGMA == guide_tree_method ||	TJ == guide_tree_method ||	NJ == guide_tree_method ) {
		for (i=0; i<K; i++) 
			free(precomputed_distances[i]);
		free(precomputed_distances);
	}



	if ( AVG_INTERNAL_EXTERNAL == guide_tree_method ) {
		for (i=0; i<cluster_count; i++) {
			for (j=0; j<cluster_count; j++) {
				free(internal_external_distances[i][j]);
			}
			free(internal_external_total_distances[i]);
			free(internal_external_numpairs[i]);
			free(internal_external_distances[i]);
		}
		free(internal_external_total_distances);
		free(internal_external_numpairs );
		free(internal_external_distances );
	}

	free(leaves_under_i);
	free(leaves_under_j);

	free (A_ids);
	free (AB_ids);
	free (B_ids);				
	free (C_ids);				

	
	return 	alignlen;	
}

void set_weight_matrix_for_cluster (int id) {

	if (NULL == weights) 
		return;

	int *ids = aa_malloc( cluster_sizes[id] * sizeof(int));
	
	ClusterEntry *ce = clusters[id].first;
	int i=0, j=0;
	while (ce != NULL) {
		ids[i++] = ce->seqid;
		ce = ce->next;
	}

	int **tmp_weights = aa_malloc( cluster_sizes[id]*sizeof(int*));

	for (i=0; i<cluster_sizes[id]; i++) {
		tmp_weights[i] = aa_malloc(cluster_sizes[id]*sizeof(int));
		for (j=0; j<cluster_sizes[id]; j++) {
			tmp_weights[i][j] = weights[ids[i]][ids[j]];
		}
	}
	
	aa_setWeights(cluster_sizes[id],tmp_weights);

	free(ids);
	for (i=0; i<cluster_sizes[id]; i++)
		free(tmp_weights[i]);
	free(tmp_weights);
	
}


void get_leaves (int id, int **leaf_array, int *counter, int subtree_excluded) {

	if ( tree_nodes[id].left_child == -1 ) { //leaf
		if (leaf_array != NULL) {
			(*leaf_array)[*counter] = id;
		}
		(*counter)++;
	} else {
		if (subtree_excluded != tree_nodes[id].left_child) 
			get_leaves( tree_nodes[id].left_child, leaf_array, counter, subtree_excluded );
		if (subtree_excluded != tree_nodes[id].right_child) 
			get_leaves( tree_nodes[id].right_child, leaf_array, counter, subtree_excluded );				
	}
} 

int alterDistanceTable(int id, int K, int ***distances, int **active) {
	
	if (tree_nodes[id].left_child == -1) { // leaf
		return id;	
	}
	
	// working with an internal node
	int i = alterDistanceTable(tree_nodes[id].left_child, K, distances, active);
	int j = alterDistanceTable(tree_nodes[id].right_child, K, distances, active);

	int k;
	for (k=0; k<K; k++) {
		if (!(*active)[k])
			continue;
		if (k == i || k == j)
			continue;
	
		int tmp ;
		if ( 	TJ == guide_tree_method ) 
			tmp =  ( (*distances)[i][k] + (*distances)[j][k] - (*distances)[i][j] ) /2 ;
		 else if ( MST_UPGMA == guide_tree_method ) 
			tmp = 0.9*((*distances)[i][k]<(*distances)[j][k]?(*distances)[i][k]:(*distances)[j][k] ) + 0.1*(( (*distances)[i][k] + (*distances)[j][k] ) /2 );
		 else //if ( UPGMA == guide_tree_method 
			tmp = ( (*distances)[i][k] + (*distances)[j][k] ) /2;

		(*distances)[i][k] = (*distances)[k][i] = tmp;
	}
	(*active)[j] = AA_FALSE;
	
	return i;
}

void checkIsLetter(char c) {
	if ( (c<'A' || c > 'Z') && c != '-' ) {
		printf ("\n**last letter (%c)  was bad\n", c);
		fflush (stdout);
		fflush (stderr);	
		exit(1);
	}
}

unsigned int localSearch_purecut (int num_cuts, int K, int space_optimized, char** alignment, 
								int alignlen, char ***polished_alignment, int *cost) 
{
	int i, j;
	char ***splits = aa_malloc(num_cuts * sizeof(char**));
	int *split_sizes = aa_malloc( num_cuts * sizeof(int));
	
	char **final_alignment = aa_malloc(K * sizeof(char*));
	for (i=0; i< K; i++) {
		final_alignment[i] = aa_malloc(alignlen * sizeof(char));
		memcpy(final_alignment[i], alignment[i], alignlen);
	}

	aa_setWeights(K,weights);
	int prev_cost = sp(final_alignment, alignlen, K);
	int new_cost ;  
	int final_alignlen = alignlen;

	int num_times_no_cost_change = 0;
	
	int keep_going = AA_TRUE; 
	unsigned int old_alignlen1;
	unsigned int new_alignlen1;
	unsigned int new_alignlen2;
	unsigned char *old_script1; 
	unsigned char *old_script2;	
	unsigned char *new_script1; 
	unsigned char *new_script2;	

		
	int **maps = aa_malloc(num_cuts * sizeof(int*));
	for (i=0;i<num_cuts;i++)
		maps[i] = aa_malloc(K * sizeof(int));

	char **tmp_alignment;

	char **new_alignment;			
	int which_merge;
	int align1_size;
	if (verbose) printf ("going into new local search loop. Current cost is: %d\n", prev_cost);	
	while (keep_going) {
		
/*		int xx,yy;
		printf("A\n");
		for (xx=0; xx<K; xx++){
			for (yy=0; yy<final_alignlen; yy++) {
				printf ("%c", final_alignment[xx][yy]);
				checkIsLetter(final_alignment[xx][yy]);				
			}	
			printf ("\n");
		}
		printf("\n");				
*/
		nWaySplit(num_cuts, K, space_optimized, final_alignment, final_alignlen, 
						&splits, &split_sizes, &maps); 
				
		if (num_cuts == 2) {
			//in splitA and splitB, check for each column if all positions are gap. If so, label it a gap column
			// if both splitA and splitB have gap column on same column, remove it.

/*printf("after nwaysplit\n");
		for (xx=0; xx<K; xx++){
			for (yy=0; yy<final_alignlen; yy++) {
				printf ("%c", final_alignment[xx][yy]);
				checkIsLetter(final_alignment[xx][yy]);				
			}	
			printf ("\n");
		}
		printf("\n");		
*/		
			getInducedScript(splits[0], split_sizes[0], splits[1], split_sizes[1], final_alignlen, &old_script2);


/*		printf("after get induced\n");
		for (xx=0; xx<K; xx++){
			for (yy=0; yy<final_alignlen; yy++) {
				printf ("%c", final_alignment[xx][yy]);
				checkIsLetter(final_alignment[xx][yy]);
			}	
			printf ("\n");
		}
		printf("\n");	*/	
			new_alignlen2 = do_alignment( &splits[0], split_sizes[0], final_alignlen, maps[0],
							&splits[1], split_sizes[1], final_alignlen, maps[1],
							&new_alignment, &new_script2, AA_TRUE);
/*		printf("after newalign\n");
		for (xx=0; xx<K; xx++){
			for (yy=0; yy<final_alignlen; yy++) {
				printf ("%c", final_alignment[xx][yy]);
				checkIsLetter(final_alignment[xx][yy]);
			}	
			printf ("\n");
		}
		printf("\n");	*/								
			new_cost = sp(new_alignment, new_alignlen2, split_sizes[0]+split_sizes[1]);  

/*		printf("after cost\n");
		for (xx=0; xx<K; xx++){
			for (yy=0; yy<final_alignlen; yy++) {
				printf ("%c", final_alignment[xx][yy]);
				checkIsLetter(final_alignment[xx][yy]);
			}	
			printf ("\n");
		}
		printf("\n");		
			*/
		} else { //num_cuts == 3

			// need to decide which way of merging works best:
			//          ((A,B),C)  ,  ((A,C),B)  ,  ((B,C),A)
			bestThreeWayMerge(K, space_optimized, splits, maps, split_sizes, final_alignlen, 
						&new_alignlen1, &new_alignlen2, &new_alignment, 
						&new_script1, &new_script2, &new_cost, &which_merge);

			int *tmp_map;
			if (which_merge == 1 || which_merge == -1) {
				align1_size = split_sizes[0] + split_sizes[1];
				tmp_alignment = aa_malloc(align1_size * sizeof(char*));
				for (i=0; i<split_sizes[0]; i++)
					tmp_alignment[i] = splits[0][i];
				for (j=0; j<split_sizes[1]; j++)
					tmp_alignment[i+j] = splits[1][j];				
				getInducedScript(tmp_alignment, align1_size, splits[2], split_sizes[2], alignlen, &old_script2);
				if (which_merge == 1) 
					old_alignlen1 = getInducedScript(splits[0], split_sizes[0], splits[1], split_sizes[1], alignlen, &old_script1);
			} else if (which_merge == 2 || which_merge == -2) {
				tmp_map = maps[1];
				maps[1] = maps[2];
				maps[2] = tmp_map;
				align1_size = split_sizes[0] + split_sizes[2];
				tmp_alignment = aa_malloc(align1_size * sizeof(char*));
				for (i=0; i<split_sizes[0]; i++)
					tmp_alignment[i] = splits[0][i];
				for (j=0; j<split_sizes[2]; j++)
					tmp_alignment[i+j] = splits[2][j];				
				getInducedScript(tmp_alignment, align1_size, splits[1], split_sizes[1], alignlen, &old_script2);
				if (which_merge == 2)
					old_alignlen1 = getInducedScript(splits[0], split_sizes[0], splits[2], split_sizes[2], alignlen, &old_script1);
			} else if (which_merge == 3 || which_merge == -3) {
				tmp_map = maps[0];
				maps[0] = maps[1];
				maps[1] = maps[2];
				maps[2] = tmp_map;
				align1_size = split_sizes[1] + split_sizes[2];
				tmp_alignment = aa_malloc(align1_size * sizeof(char*));
				for (i=0; i<split_sizes[1]; i++)
					tmp_alignment[i] = splits[1][i];
				for (j=0; j<split_sizes[2]; j++)
					tmp_alignment[i+j] = splits[2][j];				
				getInducedScript(tmp_alignment, align1_size, splits[0], split_sizes[0], alignlen, &old_script2);
				if (which_merge == 3)
					old_alignlen1 = getInducedScript(splits[1], split_sizes[1], splits[2], split_sizes[2], alignlen, &old_script1);
			}			
			free(tmp_alignment);
			
		}

/*		printf("B1\n");
		for (xx=0; xx<K; xx++){
			for (yy=0; yy<new_alignlen2; yy++) {
				printf ("%c", new_alignment[xx][yy]);
				checkIsLetter(new_alignment[xx][yy]);
			}	
			printf ("\n");
		}
		printf("\n");		
		printf("B2\n");
		for (xx=0; xx<K; xx++){
			for (yy=0; yy<final_alignlen; yy++) {
				printf ("%c", final_alignment[xx][yy]);
				checkIsLetter(final_alignment[xx][yy]);
			}	
			printf ("\n");
		}
		printf("\n");		*/


		for(i=0; i<num_cuts; i++) {
			for(j=0; j<split_sizes[i]; j++) 
				free(splits[i][j]);	
			free (splits[i]);
		}

		if (verbose) printf ("end of local search loop. Current cost is: %d\n", new_cost);	

		if (new_cost > prev_cost)
			break;
		
//printf ("1A\n");
		if (new_cost == prev_cost)
			num_times_no_cost_change++;

		if (new_cost < prev_cost) {
//printf ("1B\n");
				num_times_no_cost_change = 0;
				prev_cost = new_cost; //so "keep_going = TRUE"
		} else if ( num_times_no_cost_change == 2) {
//printf ("1C\n");
			keep_going = AA_FALSE; // avoiding loops
		} else if (new_alignlen2 != final_alignlen) {
			//a change has been made, so "keep_going = TRUE"
		} else {
			// a change may still have been made. compare gap positions.
//printf ("1D\n");			
			keep_going = AA_FALSE;
			for (i=0; (!keep_going && i<new_alignlen2); i++) {
				if (old_script2[i] != new_script2[i] ) {
					// a change has been made	
					keep_going = AA_TRUE; 
				}	
			}			
			if (num_cuts == 3 && which_merge > 0) {  // only need to check this if it either exists (num_cuts ==3) or could possibly have changed (which_merge>0)
				for (i=0; (!keep_going && i<new_alignlen1); i++) {
					if (old_script1[i] != new_script1[i] ) {
						// a change has been made	
						keep_going = AA_TRUE; 
					}	
				}
			}
		}

		if (num_cuts == 3 && which_merge>0 ){
			free (old_script1) ;
			free (new_script1) ;			
		}
		free (old_script2) ;
		free (new_script2) ;			


/*		printf("C1\n");
		for (xx=0; xx<K; xx++){
			for (yy=0; yy<new_alignlen2; yy++) {
				printf ("%c", new_alignment[xx][yy]);
				checkIsLetter(new_alignment[xx][yy]);
			}	
			printf ("\n");
		}
		printf("\n");		
		printf("C2\n");
		for (xx=0; xx<K; xx++){
			for (yy=0; yy<final_alignlen; yy++) {
				printf ("%c", final_alignment[xx][yy]);
				checkIsLetter(final_alignment[xx][yy]);
			}	
			printf ("\n");
		}
		printf("\n");		*/

		
		if (keep_going) {
			for(i=0; i<K; i++) {
				free(final_alignment[i]);
			}

			//reorder the seqs to the original ...
			i=0;
			while (maps[0][i] != -1) {
	//			printf("A mapping new alignment %d to final_alignment %d\n", i, maps[0][i]);
				final_alignment[maps[0][i]] = new_alignment[i];
				i++;
			}
			int j=0;
			while (maps[1][j] != -1) {
//				printf("B mapping new alignment %d to final_alignment %d\n", j+i, maps[1][j]);				
				final_alignment[maps[1][j]] = new_alignment[j+i];
				j++;
			}				
			if (num_cuts == 3) {
				int k=0;
				while (maps[2][k] != -1) {
					//printf("C mapping new alignment %d to final_alignment %d\n", k+j+i, maps[2][k]);
					final_alignment[maps[2][k]] = new_alignment[k+j+i];
					k++;
				}	
			}						


//			printf("E alignlen was %d, is %d\n", final_alignlen, new_alignlen2);
			final_alignlen = new_alignlen2;
			
/*			for (xx=0; xx<K; xx++){
				for (yy=0; yy<final_alignlen; yy++) {
					printf ("%c", final_alignment[xx][yy]);
					checkIsLetter(final_alignment[xx][yy]);
				}	
				printf ("\n");
			}
			printf("\n");		*/
			

		} else {
			for(i=0; i<K; i++) {
				free(new_alignment[i]);
			}			
		}		

		free(new_alignment);
		
	}

	for (i=0; i<num_cuts; i++)
		free (maps[i]);
	free (maps);	

	free (splits);
	free (split_sizes);
	
	*polished_alignment = final_alignment;
	*cost = prev_cost;
		
	return final_alignlen;	

}

//n must be either 2 or 3
// designed to be called only with a complete alignment of all K seqs, already reordered to mach original order 
void nWaySplit (int n, int K, int space_optimized, char** alignment, int alignlen, 
						char ****splits, int **split_sizes, int ***maps) {

	if (n!=2 && n!= 3) {
		fprintf(stderr, "nWaySplit should only be used with 2- or 3-way splits\n");
		exit(1);
	}

	int i,j,k; 

	for (i=0; i<n; i++)
		(*splits)[i] = aa_malloc(K* sizeof(char*));
	
	for (i=0; i<n; i++) {
		for (j=0; j<K; j++)
			(*maps)[i][j] = -1;
	}

	
	int largest_diff = -1;
	int largest_diff_i=-1, largest_diff_j=-1, largest_diff_k=-1;
	float diff_ij, diff_ik, diff_jk;
	float diff;
	
	int **pairwise_optimal;		
	int **induced_pairwise;
/*
int xx, yy;
	printf("in Nwaysplit\n");
	for (xx=0; xx<K; xx++){
		for (yy=0; yy<alignlen; yy++) {
			printf ("%c", alignment[xx][yy]);
			checkIsLetter(alignment[xx][yy]);
		}	
		printf ("\n");
	}
	printf("\n");		
	*/
	
	
	calculatePairwiseDistances(K, alignment, alignlen, &pairwise_optimal, &induced_pairwise, space_optimized) ;
										
	if (n == 2) {
					
		for (i=0; i<K-1; i++){
			for (j=i+1; j<K; j++) {
				diff = induced_pairwise[i][j] - pairwise_optimal[i][j]; 
				if (  diff > largest_diff) {		
					largest_diff = diff;
					largest_diff_i = i;
					largest_diff_j = j;
				}
			}
		} 		
	} else { //n == 3				

		// pick the triple with the largest sum of diffs between optimal and induced
		for (i=0; i<K-2; i++){
			for (j=i+1; j<K-1; j++) {
				diff_ij = induced_pairwise[i][j] - pairwise_optimal[i][j]; 
				for (k=j+1; k<K; k++) {
					diff_ik = induced_pairwise[i][k] - pairwise_optimal[i][k]; 
					diff_jk = induced_pairwise[j][k] - pairwise_optimal[j][k]; 
				
					diff = diff_ij + diff_ik + diff_jk;
					
					if (  diff > largest_diff) {		
						largest_diff = diff;
						largest_diff_i = i;
						largest_diff_j = j;
						largest_diff_k = k;					
					}
				}
			}
		} 
	}


	(*splits)[0][0] = aa_malloc(alignlen * sizeof(char));
	memcpy((*splits)[0][0],alignment[largest_diff_i],alignlen);
	(*maps)[0][0] = largest_diff_i;
	
	(*splits)[1][0] = aa_malloc(alignlen * sizeof(char));
	memcpy((*splits)[1][0],alignment[largest_diff_j],alignlen);
	(*maps)[1][0] = largest_diff_j;

	if (n == 3) {
		(*splits)[2][0] = aa_malloc(alignlen * sizeof(char));
		memcpy((*splits)[2][0],alignment[largest_diff_k],alignlen);
		(*maps)[2][0] = largest_diff_k;
	}
	
	int unassigned_count = K;
	for (i=0; i<n; i++) {
		(*split_sizes)[i] = 1;
		unassigned_count--;
	}

	int *unassigned = aa_malloc(K*sizeof(int));
	for(i=0; i<K; i++)
		unassigned[i] = (i==largest_diff_i || i==largest_diff_j || i==largest_diff_k )? AA_FALSE : AA_TRUE;


	int largest_diff_which_group;
//	int induced[n][K];
	int **induced;
	induced = aa_malloc(n * sizeof(int*));
	for (i=0; i<n; i++)
		induced[i] = aa_malloc(K * sizeof(int));	
		
	for (i=0; i<K; i++) {
		induced[0][i] = induced_pairwise[i][largest_diff_i]; //alignmentCost(alignment+i, alignment+largest_diff_i, 1, 1, alignlen, alignlen);	
		induced[1][i] = induced_pairwise[i][largest_diff_j]; //alignmentCost(alignment+i, alignment+largest_diff_j, 1, 1, alignlen, alignlen);
		if (n == 3) 
			induced[2][i] = induced_pairwise[i][largest_diff_k];//alignmentCost(alignment+i, alignment+largest_diff_k, 1, 1, alignlen, alignlen);	
	}
	float normalized_diff1, normalized_diff2, normalized_diff3 ;
	 

//	int seq2cluster_cost[n][K];
//	int seq2cluster_weight[n][K];
	int **seq2cluster_cost;
	int **seq2cluster_weight;	
	seq2cluster_cost = aa_malloc(n * sizeof(int*));
	seq2cluster_weight = aa_malloc(n * sizeof(int*));	
	for (i=0; i<n; i++){
		seq2cluster_cost[i] = aa_malloc(K * sizeof(int));	
		seq2cluster_weight[i] = aa_malloc(K * sizeof(int));			
	}
		

	for (i=0; i<n; i++){
		for (j=0; j<K; j++) {
			seq2cluster_cost[i][j] = pairwise_optimal[j][(*maps)[i][0]];
			if (NULL != weights) {
				seq2cluster_weight[i][j] = weights[j][(*maps)[i][0]];
			} else {
				seq2cluster_weight[i][j] = 1; 
			}
		}
	}
					//alternate "best" calc would be: 
	/*					for each group, 
							append seq to end of group; (e.g. A)
							try both ways of adding that group progressively to the other two groups
									(e.g. ((A to B) to C),  ((A to C) to B))
							pick the way with the lowest score. (e.g. that's bestA)
					Requires recalc for each seq at each step ... much slower		
	*/
	
	while (unassigned_count>0) {
		int largest_diff = -1;
		int diff, diff_which_group;

		for (i=0;i<K;i++){
			//Pick which group this sequence should be added to in order to acheive the best 
			//difference between optimal pairwise and induced pairwise.
			if (unassigned[i]) {

				if (n == 2) {				

					//induced pairwise cost with this sequence added would be 
					normalized_diff1 = seq2cluster_weight[1][i] == 0 ? 0 : (float)(induced[1][i] - seq2cluster_cost[1][i]) / seq2cluster_weight[1][i] ;
					normalized_diff2 = seq2cluster_weight[0][i] == 0 ? 0 : (float)(induced[0][i] - seq2cluster_cost[0][i]) / seq2cluster_weight[0][i] ;


//					printf ("seq %d: ndiff1 = %.1f, ndiff2 = %.1f\n", i, normalized_diff1, normalized_diff2);
					
					if (normalized_diff1 > normalized_diff2 ) {
							diff = normalized_diff1;
							diff_which_group = 1;
					} else {
							diff = normalized_diff2;
							diff_which_group = 2;
					}						
				} else { // n==3
					
					float diff1 = seq2cluster_weight[0][i] == 0 ? 0 : (float)(induced[0][i] - seq2cluster_cost[0][i]) / seq2cluster_weight[0][i] ;
					float diff2 = seq2cluster_weight[1][i] == 0 ? 0 : (float)(induced[1][i] - seq2cluster_cost[1][i]) / seq2cluster_weight[1][i] ;
					float diff3 = seq2cluster_weight[2][i] == 0 ? 0 : (float)(induced[2][i] - seq2cluster_cost[2][i]) / seq2cluster_weight[2][i] ;										
					
					//induced pairwise cost with this sequence added would be 
					//  the cost without the seq + the cost of aligning it to group1 + cost against group2
					normalized_diff1 =   diff2 + diff3;
					normalized_diff2 =   diff1 + diff3;
					normalized_diff3 =   diff1 + diff2;

	
					if (normalized_diff1 > normalized_diff2 && normalized_diff1 > normalized_diff3) {
							diff = normalized_diff1;
							diff_which_group = 1;
					} else if ( normalized_diff2 > normalized_diff1 && normalized_diff2 > normalized_diff3 ) {
							diff = normalized_diff2;
							diff_which_group = 2;
					} else { //( normalized_diff3 > normalized_diff1 && normalized_diff3 > normalized_diff2 ) {
							diff = normalized_diff3;
							diff_which_group = 3;						
					}
				}	
				//Then identify the unassigned sequence who's assignment give the *biggest* difference					
				if ( diff > largest_diff) {
					largest_diff = diff;
					largest_diff_i = i;
					largest_diff_which_group = diff_which_group;
				}				
			}
		}
		
		unassigned[largest_diff_i] = AA_FALSE;
		if (largest_diff_which_group == 1) {
			(*splits)[0][(*split_sizes)[0]] = aa_malloc(alignlen * sizeof(char));
			memcpy((*splits)[0][(*split_sizes)[0]],alignment[largest_diff_i],alignlen);
			
			(*maps)[0][(*split_sizes)[0]] = largest_diff_i;
			(*split_sizes)[0]++;
			
/*			printf ("just added to split[0], which now contains: ");
			int kk;
			for (kk=0; kk<(*split_sizes)[0]; kk++) {
				printf("%d,",(*maps)[0][kk]);	
			}
			printf ("\n");
*/			
		} else if (largest_diff_which_group == 2) {
			(*splits)[1][(*split_sizes)[1]] = aa_malloc(alignlen * sizeof(char));
			memcpy((*splits)[1][(*split_sizes)[1]],alignment[largest_diff_i],alignlen);
			
			(*maps)[1][(*split_sizes)[1]] = largest_diff_i;
			(*split_sizes)[1]++;
			
/*			printf ("just added to split[1], which now contains: ");
			int kk;
			for (kk=0; kk<(*split_sizes)[1]; kk++) {
				printf("%d,",(*maps)[1][kk]);	
			}
			printf ("\n");
*/			
		} else if (n == 3) { 
			(*splits)[2][(*split_sizes)[2]] = aa_malloc(alignlen * sizeof(char));
			memcpy((*splits)[2][(*split_sizes)[2]],alignment[largest_diff_i],alignlen);
			
			(*maps)[2][(*split_sizes)[2]] = largest_diff_i;
			(*split_sizes)[2]++;
		} else {
			fprintf(stderr, "nWaySplit : shouldn't be here ...\n");
			exit(1);		
		}
		unassigned_count--;	  
		
		int **tmp = aa_malloc(2*sizeof(int*));
		tmp[0] = aa_malloc(2*sizeof(int));
		tmp[1] = aa_malloc(2*sizeof(int));

		for (j=0; j<K; j++) {
			if (! unassigned[j])
				continue;
			int pw_weight = (NULL == weights) ? 1 : weights[largest_diff_i][j];	

			seq2cluster_weight[largest_diff_which_group-1][j] += pw_weight;			

			seq2cluster_cost[largest_diff_which_group-1][j] +=  
						pairwise_optimal[j][largest_diff_i] * pw_weight;

			induced[largest_diff_which_group-1][j] += 
			 			induced_pairwise[j][largest_diff_i] * pw_weight;




/*			tmp[0][0] = tmp[1][1] = 0;
			tmp[0][1] = tmp[1][0] = weights[largest_diff_i][j];
			aa_setWeights(2,tmp);
			
			int change = alignmentCost(alignment+j, alignment+largest_diff_i, 1, 1, alignlen, alignlen);

			if (largest_diff_which_group == 1)
				induced1[j] += change;
			else if (largest_diff_which_group == 2)
				induced2[j] += change;
			else if (largest_diff_which_group == 3)
				induced3[j] += change;
*/
		}
		free(tmp[0]);
		free(tmp[1]);			
		free(tmp);					
	}

	for (i=0; i<K; i++) {
		free(pairwise_optimal[i]);
		free(induced_pairwise[i]);
	}
	free (pairwise_optimal); 
	free(induced_pairwise);

	free(unassigned);
}


void bestThreeWayMerge(int K, int space_optimized, char ***splits, int **maps, int *split_sizes, 
					int alignlen, unsigned int *new_alignlen1, unsigned int *new_alignlen2, 
					char ***return_alignment2, 
					unsigned char **return_script1, unsigned char **return_script2, 
					int *cost, int *which_merge)
{
	int i, j, k;	
	char **tmp_alignment1;
	char **tmp_alignment2;
	char **tmp_alignment3;	
	unsigned int tmp_alignlen1;
	unsigned int tmp_alignlen2;	
	int tmp_cost;
	int align1_size;
	unsigned char *tmp_script1;
	unsigned char *tmp_script2;	
	
	*return_alignment2 = NULL;
	if (return_script1 != NULL)
		*return_script1 = NULL;
	if (return_script2 != NULL)
		*return_script2 = NULL;

	*new_alignlen1 = alignlen;
	*cost = AA_INF;

	int x,y,z;
	if (allow_both_2way_and_3way_cut) { // this first case is really a 2-way cut masquerading as a 3-way cut
		// before doing progressive merges, get costs of induced (A,B) aligned to C
		//												induced (A,C) aligned to B
		//												induced (B,C) aligned to A
		tmp_alignment1 = aa_malloc( K * sizeof(char*));
		tmp_alignment2 = aa_malloc( K * sizeof(char*));

		int *A_ids = aa_malloc( K * sizeof(int));	
		int *B_ids = aa_malloc( K * sizeof(int));	


		for (i=0; i<3; i++) {
			switch (i) {
				case 0: x=0; y=1; z=2; break;  // (0,1),2
				case 1: x=0; y=2; z=1; break;  // (0,2),1
				case 2: x=1; y=2; z=0; break;  // (1,2),0						
			}

			
			for (j=0; j<split_sizes[x]; j++) {
				tmp_alignment1[j] = splits[x][j];
				A_ids[j] = maps[x][j];
			}
			for (k=0; k<split_sizes[y]; k++){
				tmp_alignment1[j+k] = splits[y][k];
				A_ids[j+k] = maps[y][k];
			}
			for (j=0; j<split_sizes[z]; j++){
				tmp_alignment2[j] = splits[z][j];
				B_ids[j] = maps[z][j];
			}

			tmp_cost = calcBestAlignCost_preservingOrig(space_optimized, use_weights_in_merge, AA_FALSE, 
								tmp_alignment1, split_sizes[x] + split_sizes[y], alignlen, A_ids,
								tmp_alignment2, split_sizes[z], alignlen, B_ids,
								&tmp_alignment3, &tmp_alignlen2, &tmp_script2, ALL_PAIRS);
				
/*			tmp_cost = calcBestAlignCost_preservingOrig (space_optimized, tmp_alignment1, split_sizes[x] + split_sizes[y], alignlen, A_ids, 
		 														tmp_alignment2, split_sizes[z], alignlen, B_ids, &tmp_alignment3, &tmp_alignlen2, &tmp_script2);	*/
			if (tmp_cost < *cost) {
				*which_merge = -(i+1);
				*cost = tmp_cost;
	
				if (*return_alignment2 != NULL) {
					for(j=0; j<K; j++) 
					    free((*return_alignment2)[j]);
					free((*return_alignment2));	
				}
				
				*return_alignment2 = tmp_alignment3;
				*new_alignlen2 = tmp_alignlen2;
		
				align1_size = split_sizes[x] + split_sizes[y];	
			
				if (return_script2 != NULL) {
					if (i!=0) 
						free(*return_script2);				
					*return_script2 = tmp_script2;
				}
			}	else {
				for(j=0; j<K; j++) {
					free(tmp_alignment3[j]);
				}
				free(tmp_alignment3);			
				free(tmp_script2);
			}
			
		}	
		free(tmp_alignment1);
		free(tmp_alignment2);		
		free(A_ids);
		free(B_ids);
	}
	
	
	//these are used because I want to preserve the original splits.
	char **local_splits[3];
	for (i=0; i<3; i++) {
		local_splits[i] = aa_malloc(split_sizes[i] * sizeof(char*)); 
		for (j=0; j<split_sizes[i]; j++){
			local_splits[i][j] = aa_malloc(alignlen * sizeof(char)); 
			memcpy( local_splits[i][j], splits[i][j], alignlen);
		}
	}
	//Now we try aligning the first two optimally, then aligning the result with the third - for all three combinations
	int tmp_alignlens[3];
	tmp_alignlens[0] = RemoveNullColumns (local_splits[0], split_sizes[0], alignlen);
	tmp_alignlens[1] = RemoveNullColumns (local_splits[1], split_sizes[1], alignlen);
	tmp_alignlens[2] = RemoveNullColumns (local_splits[2], split_sizes[2], alignlen);				
	

	for (i=0; i<3; i++) {
		switch (i) {
			case 0: x=0; y=1; z=2; break;  // (0,1),2
			case 1: x=0; y=2; z=1; break;  // (0,2),1
			case 2: x=1; y=2; z=0; break;  // (1,2),0						
		}
	
		// ((A,B),C)
		int *tmp_map = aa_malloc( ( split_sizes[x] + split_sizes[y]) * sizeof(int));
		for (j=0; j<split_sizes[x]; j++) {
			tmp_map[j] = maps[x][j];
		}
		for (j=0; j<split_sizes[y]; j++) {
			tmp_map[j+split_sizes[x]] = maps[y][j];
		}				
		
		tmp_alignlen1 = do_alignment( &local_splits[x], split_sizes[x], tmp_alignlens[x], maps[x], 
						&local_splits[y], split_sizes[y], tmp_alignlens[y], maps[y],
						&tmp_alignment1, &tmp_script1, AA_TRUE);		
		tmp_alignlen2 = do_alignment( &tmp_alignment1, split_sizes[x] + split_sizes[y], tmp_alignlen1, tmp_map, 
						&local_splits[z], split_sizes[z], tmp_alignlens[z], maps[z],
						&tmp_alignment2, &tmp_script2, AA_TRUE); 
		tmp_cost = sp(tmp_alignment2, tmp_alignlen2, K);  
		free(tmp_map);
	
		if (tmp_cost < *cost) {		
						
			*new_alignlen1 = tmp_alignlen1;
			*new_alignlen2 = tmp_alignlen2;
			*cost = tmp_cost;
			align1_size = split_sizes[x] + split_sizes[y];	

			if (*return_alignment2 == NULL) {
				*return_alignment2 = aa_malloc( K * sizeof(char*));
			}

			for (j=0; j<K; j++) {
				if (i!=0  ||  allow_both_2way_and_3way_cut )
					free((*return_alignment2)[j]);
				(*return_alignment2)[j] = tmp_alignment2[j];
			}
			 
			if (return_script1 != NULL) {
				if (i!=0) 
					free(*return_script1);	
				*return_script1 = tmp_script1;
			} else {
			  free(tmp_script1);
			}
			if (return_script2 != NULL) {
				if (i!=0  ||  allow_both_2way_and_3way_cut )
					free(*return_script2);
				*return_script2 = tmp_script2;		
			} else {
			  free(tmp_script2);
			}
			*which_merge=i+1;			
		} else {
			for (j=0; j<K; j++) 
				free(tmp_alignment2[j]);		
			free(tmp_script1);	
			free(tmp_script2);
		}
		
		for (j=0; j<split_sizes[x] + split_sizes[y]; j++)	
			free(tmp_alignment1[j]);				

		free (tmp_alignment1); 
		free (tmp_alignment2);

	}

	for (i=0; i<3; i++) {
		for (j=0; j<split_sizes[i]; j++)
			free(local_splits[i][j]);
		free(local_splits[i]);
	}

	
/*	if (*which_merge > 0 ) { // if <0, keeping the induced alignment of two clusters worked better than optimally aligning them before aligning with the third
		free(tmp_alignment3);
	}		*/
} 


//expects to be called with the full alignment, already re-ordered to match the original input order
void calculatePairwiseDistances(int K, char **alignment, int alignlen, 
					int ***pairwise_optimal, int ***induced_pairwise, int space_optimized) 
{


	// distance = difference between the pairwise alignment cost imposed by 
	// the multiple alignment in excess of the optimal pairwise cost
	int i,j; 
	*pairwise_optimal =aa_malloc(K * sizeof(int*));
	*induced_pairwise =aa_malloc(K * sizeof(int*));
	for (i=0; i<K; i++) {
		(*pairwise_optimal)[i] = aa_malloc(alignlen * sizeof(int));
		(*induced_pairwise)[i] = aa_malloc(alignlen * sizeof(int));
	}

	aa_setUnweighted();	
	for (i=0; i<K-1; i++){
		for (j=i+1; j<K; j++) {
					
			(*pairwise_optimal)[i][j] = (*pairwise_optimal)[j][i] = 
									 calcBestAlignCost_preservingOrig(space_optimized, AA_FALSE, AA_FALSE, 
																		alignment+i, 1, alignlen, &i, 
																		alignment+j, 1, alignlen, &j, 
																		NULL, NULL, NULL, ALIGN_COST);

			// the previous statement will have set the weights as needed for this statement ... so I don't need to do anything
			(*induced_pairwise)[i][j] = (*induced_pairwise)[j][i] = alignmentCost(alignment+i, alignment+j, 1, 1, alignlen, alignlen);
		}
	} 
}

int getInducedScript(char **splitA, int splitA_size, char **splitB, int splitB_size, 
				int alignlen, unsigned char **script){

	int i, j, k=0, isGapA, isGapB;

	*script = aa_malloc(alignlen * sizeof(int)); // loose bound on biggest alignment possible
	
	for (i=0; i<alignlen; i++) {
		isGapA = AA_TRUE;
		for (j=0; j<splitA_size; j++) {
			if (splitA[j][i] != AA_EPS) {
				isGapA = AA_FALSE;
				break;
			}
		}
		isGapB = AA_TRUE;
		for (j=0; j<splitB_size; j++) {
			if (splitB[j][i] != AA_EPS) {
				isGapB = AA_FALSE;
				break;
			}
		}		
		
		if ( !isGapA || !isGapB ){
			if ( !isGapA  && !isGapB ) 
				(*script)[k] = AA_SUB;
			else if ( isGapA  )
				(*script)[k] = AA_INS; 
			else // if ( isGapB  )				
				(*script)[k] = AA_DEL;

			k++;
		}
	}
	
	return k;
}

void fill_compressed_alph_map() {
	compressed_alph_map	= aa_malloc(256*sizeof(char));
	
	/*Dayhoff - Amino acid groups according to MAFFT (sextet5)
	# 0 =  A G P S T
	# 1 =  I L M V
	# 2 =  N D Q E B Z
	# 3 =  R H K
	# 4 =  F W Y
	# 5 =  C
	# 6 =  X . - U
	*/
	compressed_alph_map['a'] = 	compressed_alph_map['A'] = 'A';
	compressed_alph_map['g'] = 	compressed_alph_map['G'] = 'A';
	compressed_alph_map['p'] = 	compressed_alph_map['P'] = 'A';		
	compressed_alph_map['s'] = 	compressed_alph_map['S'] = 'A';
	compressed_alph_map['t'] = 	compressed_alph_map['T'] = 'A';

	compressed_alph_map['i'] = 	compressed_alph_map['I'] = 'B';
	compressed_alph_map['l'] = 	compressed_alph_map['L'] = 'B';
	compressed_alph_map['m'] = 	compressed_alph_map['M'] = 'B';
	compressed_alph_map['v'] = 	compressed_alph_map['V'] = 'B';			

	compressed_alph_map['n'] = 	compressed_alph_map['N'] = 'C';
	compressed_alph_map['d'] = 	compressed_alph_map['D'] = 'C';
	compressed_alph_map['q'] = 	compressed_alph_map['Q'] = 'C';
	compressed_alph_map['e'] = 	compressed_alph_map['E'] = 'C';
	compressed_alph_map['b'] = 	compressed_alph_map['B'] = 'C';
	compressed_alph_map['z'] = 	compressed_alph_map['Z'] = 'C';

	compressed_alph_map['r'] = 	compressed_alph_map['R'] = 'D';
	compressed_alph_map['h'] = 	compressed_alph_map['H'] = 'D';
	compressed_alph_map['k'] = 	compressed_alph_map['K'] = 'D';
	
	compressed_alph_map['f'] = 	compressed_alph_map['F'] = 'E';
	compressed_alph_map['w'] = 	compressed_alph_map['W'] = 'E';
	compressed_alph_map['y'] = 	compressed_alph_map['Y'] = 'E';
	
	compressed_alph_map['c'] = 	compressed_alph_map['C'] = 'F';
	
	compressed_alph_map['x'] = 	compressed_alph_map['X'] = 'G';
	compressed_alph_map['u'] = 	compressed_alph_map['U'] = 'G';

	compressed_alph_map['-'] = '-';	
}
