/*****************************************************************
 * FILE: Opal.h
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: @OPAL_RELEASE_DATE@
 * Version:  @OPAL_RELEASE_VERSION@
 * 
 ****************************************************************/

#include <stdio.h> /* fprintf */
#include "heap.h"

#define PROGNAME "Opal"
#define VERSION "@OPAL_RELEASE_VERSION@"

enum format {CLUSTAL_FORMAT, FASTA_FORMAT};
enum guide_tree_type {TOTAL_INTERNAL, AVG_INTERNAL, AVG_INTERNAL_EXTERNAL, MIN_INTERNAL, MST, MST_UPGMA, NJ, TJ, UPGMA};
enum local_search_type {TREE_CUT, THREE_CUT_TREE, ANY_CUT};
enum cut_choice_type {CUT_COST_AVG, CUT_COST_TOTAL};
enum cost_type {ALL_PAIRS,ALIGN_COST,MIN_COST};
enum three_tree_cut_type {CUT_PAIR, CUT_MULTI};
enum two_cut_type {NONE, ALL, BEST};
enum distance_type {EDIT, NORM_EDIT, PCT_ID, COMPR_PCT_ID};
enum overall_type {THREE_CUT, EXH_SEARCH};

typedef struct cluster_entry { /* id of one sequence in a group of seqs*/
  struct cluster_entry  *next;   /* next entry in cluster list               */
  unsigned int   seqid;      /* id          */
} ClusterEntry;

typedef struct best_edge_entry {
  struct best_edge_entry  *next;   
  struct best_edge_entry  *prev;   
  int   edge_id;
  int diff;
} BestEdgeEntry;

typedef struct  { 
  struct cluster_entry  *first;   
  struct cluster_entry  *last;   
} ClusterList;

typedef struct { 
  int i;   
  int j;   
} IDPair;

typedef struct {
	int preorder;
	int postorder;
	int parent;
	int left_child;
	int right_child;
} Node;

int ReadSequences(char *seqFile, char ****alignments, int **lengths, char ***names);
				
int  ReadCost(char *cost_file, int ***cost, char **alphabet, int **alphaconv);
int  ReadWeights(char *weight_file, int K, int ***weights);

void printAlignment_clustal(int K, char **alignment, int alignlen, FILE * out, 
			int outputwidth, char **names);
void printAlignment_fasta(int K, char **alignment, int alignlen, FILE * out, 
			int outputwidth, char **names);
void printcost(int space_optimized, int gamma, int lambda, char* costFile, char* seqFile, int K, 
			char **alignment, int alignlen, char* out_file, int **weights);
void printhelp(char *progname);
void Usage(char *progname); /* emit usage message */

void getWeights(char *weight_file, int K, int ***weights);


#define   OPAL_GAMMA      60
#define   OPAL_LAMBDA     38
#define   OPAL_GAMMA_T    15
#define   OPAL_LAMBDA_T     36
