/*
 * heap.h -- Pairing heap definitions
 */

/*
 * Copyright 1989, 1996 by John Kececioglu
 */


#ifndef HeapInclude
#define HeapInclude

#define Proto(arguments) arguments

typedef void *Pointer;

/*
 * Nil pointer
 */
#ifndef Nil
#define Nil 0
#endif

/*
 * Heap key and value
 */
typedef Pointer HeapKey;
typedef Pointer HeapValue;

/*
 * Heap node
 */
typedef struct HeapStruct {
    HeapKey   Key;
    HeapValue Value;
    struct HeapStruct *Up, *Left, *Right;
       /*
        * `Up' is reused for the pool of free nodes
        */
} HeapNode;

/*
 * Pairing heap
 */
typedef struct {
   int (*Compare) Proto(( HeapKey, HeapKey )); /* Key comparison-function */
   HeapNode *Root;
      /*
       * `Root' is reused for the pool of free heaps
       */
} Heap;



typedef struct HeapBlockStruct {
	Heap *Block;
    struct HeapBlockStruct *Next;
} HeapBlock;

typedef struct NodeBlockStruct {
	HeapNode *Block;
    struct NodeBlockStruct *Next;
} NodeBlock;

/*
 * Creation and destruction
 */
extern Heap *CreateHeap  Proto(( int (*Compare) Proto((HeapKey, HeapKey)) ));
extern void  DestroyHeap Proto(( Heap *H ));

/*
 * Insertion and deletion
 */
extern HeapNode *HeapInsert Proto(( HeapKey K, HeapValue V, Heap *H ));
extern void      HeapDelete Proto(( HeapNode *N, Heap *H ));

/*
 * Arbitrary access
 */
extern HeapKey   HeapNodeKey   Proto(( HeapNode *N ));
extern HeapValue HeapNodeValue Proto(( HeapNode *N ));

/*
 * Finding the minimum
 */
extern HeapKey   HeapMinKey    Proto(( Heap *H ));
extern HeapValue HeapMinValue  Proto(( Heap *H ));
extern HeapValue HeapDeleteMin Proto(( Heap *H ));

/*
 * Miscellaneous
 */
extern Heap *HeapUnion       Proto(( Heap *A, Heap *B ));
extern void  HeapDecreaseKey Proto(( HeapNode *N, HeapKey K, Heap *H ));
extern short HeapIsEmpty     Proto(( Heap *H ));


extern void freeHeapPool ();
extern void freeNodePool ();


#endif /* HeapInclude */
